#include <stdio.h>

#include "helpers.h"
#include "String.h"
#include "Matrix.h"
#include "InventorFile.h"
#include "MiscUtility.h"
#include "ManifoldPointStruct.3Space.h"
#include "GeneticFillOutManifoldGroup.3Space.h"
#include "TestRunDataStruct.h"

//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

//::: remove/fix these magic numbers in the future...
//MAGIC NUMBERS!!!!
// FORi(200) {	//should also store generated points for future use
// if (nCross>max(CON_likely_zoomin,nmanifolds_in_zoomin/5)) {	// arbitrary cut-off value here...
// if (loop>15) nFertile+=1;
// if (!((in_region[inx].size>=40) || ((in_region[inx].size>=10)&&(p[inx].size==1))))
// CON_likely_zoomin(8)
// CON_vote_dim(11), CON_bacteria_dim(22),
 
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

#define index_btaken(v) (v.x1*CON_sq_bacteria_dim+v.x2*CON_bacteria_dim+v.x3)
#define index_nvotes(v) (v.x1*CON_sq_vote_dim+v.x2*CON_vote_dim+v.x3)

//////////////////////////////////////////////////////////////////////
boole bInList(const int x, const LinkedList<int> L)
{
  FORLIST(L) if (L.Now()==x) return TRUE;
  return FALSE;
}

//////////////////////////////////////////////////////////////////////
void CreateBoxArray(
  Array<rPoint3>&		box,
  const VolumeStruct&		R
)
{
  box.SetSize(24);
  const rPoint3& min=R.min;
  const rPoint3& max=R.max;
  rPoint3 L00=min;
  rPoint3 L01=rPoint3(max.x,min.y,min.z);
  rPoint3 L10=rPoint3(min.x,max.y,min.z);
  rPoint3 L11=rPoint3(max.x,max.y,min.z);
  rPoint3 U00=max;
  rPoint3 U01=rPoint3(min.x,max.y,max.z);
  rPoint3 U10=rPoint3(max.x,min.y,max.z);
  rPoint3 U11=rPoint3(min.x,min.y,max.z);
  box[ 0]=L00;    box[ 1]=L01;
  box[ 2]=L01;    box[ 3]=L11;
  box[ 4]=L11;    box[ 5]=L10;
  box[ 6]=L10;    box[ 7]=L00;
  box[ 8]=U00;    box[ 9]=U01;
  box[10]=U01;    box[11]=U11;
  box[12]=U11;    box[13]=U10;
  box[14]=U10;    box[15]=U00;
  box[16]=L00;    box[17]=U11;
  box[18]=L01;    box[19]=U10;
  box[20]=L11;    box[21]=U00;
  box[22]=L10;    box[23]=U01;
}

//////////////////////////////////////////////////////////////////////
void OutputVoting(
  const String&			filename,
  word* const&			nVotes,
  const int			CON_vote_dim,
  const VolumeStruct&		region,
  const int			max_votes,
  const VolumeStruct*		zoomin_region=NULL,
  const AlphaCoord*		true_alpha=NULL
)
{
  static LinkedList<VolumeStruct> old_boxes;
  const int CON_sq_vote_dim=SQ(CON_vote_dim);
  uint32 count=0;
  FORi(CON_vote_dim) FORj(CON_vote_dim) FORk(CON_vote_dim) {
    int offset=index_nvotes(VoxelCoord3Space(i,j,k));
    if (nVotes[offset]>1) ++count;
  }
  Array<AlphaCoord> vc(count);
  Array<RGB3Byte> col(count);
  count=0;
  FORi(CON_vote_dim) FORj(CON_vote_dim) FORk(CON_vote_dim) {
    int offset=index_nvotes(VoxelCoord3Space(i,j,k));
    if (nVotes[offset]>1) {
      vc[count]=VoxelCoord3Space(i,j,k).FindCenterInSpace(region,CON_vote_dim);
      if (nVotes[offset]==max_votes) col[count]=RGB3Byte_green;
      else col[count]=RGB3Byte(255,255,255)*(real(nVotes[offset])/max_votes);
      ++count;
    }
  }

  InventorFile::Open(filename);
  InventorFile::OutputColoredPoints(vc,col,10);
  if (zoomin_region!=NULL) {
    Array<AlphaCoord> inbox, outbox;
    CreateBoxArray(inbox,*zoomin_region);
    InventorFile::OutputLines(inbox,RGB3Byte_yellow);
//    InventorFile::OutputLines(outbox,RGB3Byte_blue);
    old_boxes.Add(region.Dopple());
    FORLIST(old_boxes) {
      CreateBoxArray(outbox,old_boxes.Now());
      InventorFile::OutputLines(outbox,RGB3Byte_blue);
    }
  }
  if (true_alpha!=NULL) {
    Array_AlphaCoord extra(1); extra[0]=*true_alpha;
    InventorFile::OutputPoints(extra,RGB3Byte_green,5);
  }
  InventorFile::Close();
}

//////////////////////////////////////////////////////////////////////
void OutputManifolds(
  const String&					filename,
  const Array<LinkedList_ManifoldPointStruct3Space>&	p,
  const AlphaCoord&				true_alpha,
  const Array<boole>*				pbManifoldOn=NULL
)
{
  static const RGB3Byte CON_cols[6]={RGB3Byte_green, RGB3Byte_red, RGB3Byte_cyan, RGB3Byte_blue, RGB3Byte_white, RGB3Byte_yellow };
  InventorFile::Open(filename);
  Array_AlphaCoord manifold;
  FORk(p.size) if ((pbManifoldOn==NULL)||((*pbManifoldOn)[k])) {
    manifold.SetSize(p[k].size);
    int ind=0;
    FORLIST(p[k]) manifold[ind++]=p[k].Now().a;
    InventorFile::OutputPoints(manifold,CON_cols[k%6]);
  }
  Array_AlphaCoord extra(1); extra[0]=true_alpha;
  InventorFile::OutputPoints(extra,RGB3Byte_gray,10);
  InventorFile::Close();
}

//////////////////////////////////////////////////////////////////////
GeneticFillOutManifoldGroup3Space::GeneticFillOutManifoldGroup3Space(
  const Array<ViewPairInfo>&	_pairinfo,
  const Array<Transform>&	_T0i_proj,
  const Array<rPoint3>&		_E,
  const AlphaCoord&		_true_alpha,
  const Transform&		_Tinorm,
  Array2D<rPoint3>*		_GLOB_Kpts,
  Transform* const		_pInternal
) :
  pairinfo(_pairinfo), T0i_proj(_T0i_proj), E(_E), true_alpha(_true_alpha),
  normalizer(_Tinorm), GLOB_Kpts(_GLOB_Kpts), pInternal(_pInternal),
  
  CON_n_manifolds(pairinfo.size),
  CON_vote_dim(11), CON_bacteria_dim(22),
  CON_sq_vote_dim(SQ(CON_vote_dim)),
  CON_cube_vote_dim(CON_sq_vote_dim*CON_vote_dim),
  CON_sq_bacteria_dim(SQ(CON_bacteria_dim)),
  CON_cube_bacteria_dim(CON_sq_bacteria_dim*CON_bacteria_dim),
  CON_likely_zoomin(6)
{
  bRangeDitherOn=TRUE;
  bHelpLosersOn=TRUE;
}

//////////////////////////////////////////////////////////////////////
GeneticFillOutManifoldGroup3Space::~GeneticFillOutManifoldGroup3Space()
{
  FOREACHi(bTaken) delete[] bTaken[i];
  FOREACHi(bVotedOn) delete[] bVotedOn[i];
  delete[] nVotes;
}

//////////////////////////////////////////////////////////////////////
void GeneticFillOutManifoldGroup3Space::ClearTabulators()
{
  FOREACHi(bTaken) FORj(CON_cube_bacteria_dim) bTaken[i][j]=FALSE;
  FOREACHi(bVotedOn) FORj(CON_cube_vote_dim) bVotedOn[i][j]=FALSE;
  FORi(CON_cube_vote_dim) nVotes[i]=0;
}

// find a seed for each manifold in the given region
//////////////////////////////////////////////////////////////////////
void GeneticFillOutManifoldGroup3Space::SeedInitialRegion()
{
  ClearTabulators();

  int nStillOn=0;
  FORi(CON_n_manifolds) {
    if (!bManifoldOn[i]) continue;
    SCREENPUT printf("manif %d\n",i);
    ManifoldPointStruct3Space first;
    int32 tries=0;
    do {
      if (tries++>2000) { bManifoldOn[i]=FALSE; break; }
      first=ManifoldPointStruct3Space(rand01(),random_angle(),pairinfo[i],T0i_proj,E,region,CON_bacteria_dim,CON_vote_dim);
    } while ((!first.bLegal)||(!region.bContains(first.a)));
    if (!bManifoldOn[i]) continue;
    p[i].Add(first.Dopple());
    in_region[i].Add(first.Dopple());
    bTaken[i][index_btaken(first.vb)]=TRUE;
    int offset=index_nvotes(first.vv);
    if (!bVotedOn[i][offset]) {
      bVotedOn[i][offset]=TRUE;
      ++nVotes[offset];
    }
    ++nStillOn;
  }
  
  SCREENPUT printf("%d manifolds were seeded\n",nStillOn);
  if (nStillOn<CON_likely_zoomin) throw GeneticFillOutManifoldGroup3SpaceError("not enough to zoomin");




/*  original code...
  ClearTabulators();

  int inx=0;
  FORi(CON_ncams) FORj(CON_ncams) if ((inx<CON_n_manifolds)&&(i<j)) {
    if (!bManifoldOn[i]) continue;
    printf("manif %d\n",inx);
    ManifoldPointStruct3Space first;
    do {
      first=ManifoldPointStruct3Space(random_real(),random_angle(),pairinfo[i],
        T0i_proj,E,region,CON_bacteria_dim,CON_vote_dim);
    } while ((!first.bLegal)||(!region.bContains(first.a)));
    p[inx].Add(first.Dopple());
    in_region[inx].Add(first.Dopple());
    bTaken[inx][index_btaken(first.vb)]=TRUE;
    int offset=index_nvotes(first.vv);
    if (!bVotedOn[inx][offset]) {
      bVotedOn[inx][offset]=TRUE;
      ++nVotes[offset];
    }
    ++inx;
  }
*/
}

#ifdef NOTDEFINED
//////////////////////////////////////////////////////////////////////
boole GeneticFillOutManifoldGroup3Space::MAMTDP_Bisect(
  real&				mink,
  real&				minth,
  const AlphaCoord&		target,
  const VoxelCoord3Space&	target_vv,
  const real&			width0,
  const real&			CON_parts,
  const int			mode,			// 0: search k, 1: search th
  const int			inx
)
{
  real dist;
  Array<ManifoldCoord> p(CON_parts);
  real width=width0;
  INFINITELOOP {
    real k, th, dk, dth;
    if (mode==0) {
      k=mink-width/2;
      th=minth;
      dk=width/(CON_parts-1);
      dth=0;
    } else {
      k=mink;
      th=minth-width/2;;
      dk=0;
      dth=width/(CON_parts-1);
    }
    int i=0;
    real mindist;
    real mininx=-1;
    while (i<CON_parts) {
      p[i].k=k;
      p[i].th=th;
      ManifoldPointStruct3Space mp=ManifoldPointStruct3Space(k,th,pairinfo[inx],
        T0i_proj,E,region,CON_bacteria_dim,CON_vote_dim);
      if ((mp.vv).Equals(target_vv)) { mink=k; minth=th; return TRUE; }
      dist=(mp.a-target).Norm();
      if ((mininx==-1)||(dist<mindist)) { mininx=i; mindist=dist; }
      ++i;
      k+=dk;
      th+=dth;
    }
//printf("%20.15f\n",mindist);
    mink=p[mininx].k;
    minth=p[mininx].th;
    width=width/CON_parts*2;
    if (width<width0*0.001) break;
  }
  return FALSE;
}

//////////////////////////////////////////////////////////////////////
boole GeneticFillOutManifoldGroup3Space::MoveAlongManifoldTowardsDesiredPosition(
  ManifoldPointStruct3Space&	minpt,
  const AlphaCoord&		target,
  const VoxelCoord3Space&	vv,
  const int			inx
)
{
  //::: assume each param within 1 of global min  //change 1 to be width of voxel region
  real mink=minpt.m.k;
  real minth=minpt.m.th;
  real width=1;
  INFINITELOOP {
    if (MAMTDP_Bisect(mink, minth, target, vv, width, 10, 0, inx)) break;
    if (MAMTDP_Bisect(mink, minth, target, vv, width, 10, 1, inx)) break;
    width*=0.75;
    if (width<0.001) return FALSE;
  }
  minpt=ManifoldPointStruct3Space(mink,minth,pairinfo[inx],
    T0i_proj,E,region,CON_bacteria_dim,CON_vote_dim);
  return TRUE;
}

//////////////////////////////////////////////////////////////////////
// this version chooses random manifold points near the starting position,
// which is assumed to be near the target voxel; returns best guess in
// pos0, returns true if pos0 is in target voxel, otherwise returns false
//////////////////////////////////////////////////////////////////////
boole GeneticFillOutManifoldGroup3Space::MoveAlongManifoldTowardsDesiredPosition2_earliest(
  ManifoldPointStruct3Space&	pos0,
  const VoxelCoord3Space&	target_vv,
  const int			inx
)
{
  FORi(2000) {	//should also store generated points for future use
    ManifoldPointStruct3Space q=pos0.DitherSmall(pairinfo[inx],T0i_proj,E,region,CON_bacteria_dim,CON_vote_dim);
    if ((q.bLegal)&&(q.vv.Equals(target_vv))) { pos0=q; return TRUE; }
  }
  return FALSE;
}
#endif

// this version chooses random manifold points near the starting position,
// which is assumed to be near the target voxel; returns best guess in
// pos0, returns true if pos0 is in target voxel, otherwise returns false
//////////////////////////////////////////////////////////////////////
boole GeneticFillOutManifoldGroup3Space::MoveAlongManifoldTowardsDesiredPosition2(
  ManifoldPointStruct3Space&	pos0,		// sends in current positon, returns better position
  const VoxelCoord3Space&	target_vv,
  const AlphaCoord&		target,
  const int			inx
)
{
  ManifoldPointStruct3Space p0=pos0;
  real min_dist=(pos0.a-target).Norm();
  FORi(1000) {	//should also store generated points for future use
    ManifoldPointStruct3Space q=p0.DitherSmall(pairinfo[inx],T0i_proj,E,region,CON_bacteria_dim,CON_vote_dim);
    if (!q.bLegal) continue;
    if (q.vv.Equals(target_vv)) { pos0=q; return TRUE; }
    real d=(q.a-target).Norm();
    if (d<min_dist) {		// only retain new position if better/closer than previous one
      min_dist=d;
      p0=q;
    }
  }
  return FALSE;
}

//////////////////////////////////////////////////////////////////////
void GeneticFillOutManifoldGroup3Space::SeeHowManyManifoldsCrossCertainRegion(
  boole&				bLikelyZoomin,
  const VoxelCoord3Space&	vv
)
{
  bLikelyZoomin=FALSE;
  AlphaCoord target=vv.FindCenterInSpace(region,CON_vote_dim);
  int vote_offset=index_nvotes(vv);
  Array<boole> bCross(CON_n_manifolds);
  FORi(CON_n_manifolds) bCross[i]=FALSE;
  int nCross=0;
  
  //::: for each manifold, estimate if manifold crosses desired region;
  //::: if enough do, assume that this is the zoomin region and turn off
  //::: manifolds that don't work
  FORi(CON_n_manifolds) {
    if (!bManifoldOn[i]) continue;

    //::: manifold crosses region if it already has voted on region...
    if (bVotedOn[i][vote_offset]) { SCREENPUT printf("IN: %d\n",i); ++nCross; bCross[i]=TRUE; }

    //::: otherwise, quickly try to move along manifold to desired region
    else { SCREENPUT printf("SEARCH: %d\t",i); 
      //::: find nearest point on manifold as starting position
      real mindist;
      boole bMinSet=FALSE;
      ManifoldPointStruct3Space minpt;
      FORLIST(in_region[i]) {
        const ManifoldPointStruct3Space& mp=in_region[i].Now();
        real dist=(target-mp.a).Norm();
	if ((!bMinSet)||(dist<mindist)) { bMinSet=TRUE; mindist=dist; minpt=mp; }
      }
      
      //::: move along manifold from starting position towards target position
      if (MoveAlongManifoldTowardsDesiredPosition2(minpt,vv,target,i)) {
        ++nCross;
	bCross[i]=TRUE;
	bVotedOn[i][vote_offset]=TRUE;
	++nVotes[vote_offset];
	in_region[i].Add(minpt.Dopple());
	//minpt.vv.Dump();
	SCREENPUT printf("bingo\n");
      } else
        SCREENPUT printf("\n");
    }
  }

  SCREENPUT printf("nCross==%d\n",nCross);
  SCREENPUT printf("need %d to zoomin\n",max(CON_likely_zoomin,nmanifolds_in_zoomin/3));
  if (nCross>=max(CON_likely_zoomin,nmanifolds_in_zoomin/3)) {	// arbitrary cut-off value here...
    bLikelyZoomin=TRUE;
    //::: turn off manifolds that don't cross region...
    FORi(CON_n_manifolds) {
      if (!bManifoldOn[i]) continue;
      if (!bCross[i]) bManifoldOn[i]=FALSE;
    }
  }
  nmanifolds_in_zoomin=nCross;	//ERROR!  only do this if succeed in zoomin
}

// this function looks for a 3D region of alpha space that seems to be
// intersecting all or most (with noise) of the manifolds; along the way,
// useful info about manifolds is built up
//////////////////////////////////////////////////////////////////////
void GeneticFillOutManifoldGroup3Space::FindBestZoomInVoxel()
{
//OutputVoting("vote.at.loop.start.iv",nVotes,CON_vote_dim,region,CON_n_manifolds);
//OutputManifolds("manifolds.at.loop.start.iv",p,true_alpha);

/*
//bad to do the following here because non-collapsed manifolds need
//to be filled-out more so that the search process will work, preventing them from
//being turned-off uselesslessly
  //::: determine voxel with most votes (this is updated incrementally later)
  int maxvotes=-1;
  VoxelCoord3Space maxvc, vc;
  for (vc.x1=0; vc.x1<CON_vote_dim; ++vc.x1)
    for (vc.x2=0; vc.x2<CON_vote_dim; ++vc.x2)
      for (vc.x3=0; vc.x3<CON_vote_dim; ++vc.x3) {
        int offset=index_nvotes(vc);
        if (nVotes[offset]>maxvotes) {
          maxvotes=nVotes[offset];
	  maxvc=vc;
        }
      }
  if (maxvotes>=CON_likely_zoomin) {
    boole bLikelyZoomin;
    SeeHowManyManifoldsCrossCertainRegion(bLikelyZoomin,maxvc);
    if (bLikelyZoomin) {
      zoomin_voxel=maxvc;
      return;
    }
  }
*/

  //::: bacteria-fill each manifold; stop when a voxel gets max votes
  Array<LinkedList_ManifoldPointStruct3Space> children(CON_n_manifolds);
  Array<boole> bReevalDitherRangeYet(CON_n_manifolds);
  FORi(bReevalDitherRangeYet.size) bReevalDitherRangeYet[i]=FALSE;
  int loop=0;
  const int CON_loser_level_step=50;
  const int CON_initial_explore_threshold=800;
  int loser_level=CON_loser_level_step;
  int nFertile=10;
  boole bExitAllowed=FALSE;
  int last_loser_level=-1;
  int nSameLoserLevel=0;

  INFINITELOOP {

    LinkedList<int> losers;
    losers.FreeMem();
    boole bHelpLosers=FALSE;	//did this to deal with collapsed manifolds...(improve)
    if (bHelpLosersOn) {
//      FORi(CON_n_manifolds) if (bManifoldOn[i]) if (p[i].size<loser_level) losers.Add(new int(i));
      FORi(CON_n_manifolds) if (bManifoldOn[i]) if (in_region[i].size<loser_level) losers.Add(new int(i));
//      FORLIST(losers) printf("%d  ",losers.Now());
//      printf("\n");
      if (losers.size>0) bHelpLosers=TRUE;
      if (!bHelpLosers) loser_level+=CON_loser_level_step;
    }

    //::: make sure k in range [0,1); otherwise, range dither won't work right
    FORi(CON_n_manifolds) if (bManifoldOn[i]) {
      FORLIST(p[i]) p[i].Now().m.PutThetaInRange();
      FORLISTnow(p[i],ManifoldPointStruct3Space)
        if (now.m.k==1.0) { now.m.k=0; continue; }
        if (within(now.m.k,0,1)) continue;
	if (now.m.k>1) { now.m.k=now.m.k-int(now.m.k); continue; }
	now.m.k=-now.m.k-int(-now.m.k);
	now.m.k=1.0-now.m.k;
      }
    }

    //::: calc dither range
   if (bRangeDitherOn) {
    FORi(CON_n_manifolds) if (bManifoldOn[i])
      if ((!bReevalDitherRangeYet[i])&&(p[i].size>100)) {	// 30 here is arbitrary
//        FORLIST(p[i]) p[i].Now().m.PutThetaInRange();
        DitherRangeStruct& dr=dither_range[i];
        dr.kmax=dr.kmin=p[i][0].m.k;
        dr.thmax=dr.thmin=p[i][0].m.th;
	FORLIST(p[i]) {
	  ManifoldCoord& m=p[i].Now().m;
	  if (m.k<dr.kmin) dr.kmin=m.k;
	  if (m.k>dr.kmax) dr.kmax=m.k;
	  if (m.th<dr.thmin) dr.thmin=m.th;
	  if (m.th>dr.thmax) dr.thmax=m.th;
	}
	dr.CalcDitherRange();
	bReevalDitherRangeYet[i]=TRUE;
      }
   }

    boole bSomeManifoldsNotExploredWellEnough=FALSE;
    FORi(CON_n_manifolds) {
      if (!bManifoldOn[i]) continue;
      if ((!bHelpLosers)||(bInList(i,losers))) {
      children[i].FreeMem();  //DONT NEED TO INDEX CHILDREN!
      int loop2=0;
      FORLIST(p[i]) if (loop2++>p[i].size-nFertile) { //only allow most recent 20 bacteria to spread
        ManifoldPointStruct3Space q;
	if ((bRangeDitherOn)&&(dither_range[i].bSet))
	  q=p[i].Now().RangeDither(pairinfo[i],T0i_proj,E,region,CON_bacteria_dim,CON_vote_dim,dither_range[i]);
	else
	  q=p[i].Now().Dither(pairinfo[i],T0i_proj,E,region,CON_bacteria_dim,CON_vote_dim);
	if (q.bLegal) {
//	  if (maximum_region.bContains(q.a))	//...future use in backtracking
//	    every_pt.Add(q.Dopple());
          if (region.bContains(q.a)) {
	    in_region[i].Add(q.Dopple());
	    if (!bTaken[i][index_btaken(q.vb)]) {
              bTaken[i][index_btaken(q.vb)]=TRUE;
	      children[i].Add(q.Dopple());
	      int offset=index_nvotes(q.vv);
              if (!bVotedOn[i][offset]) {
                bVotedOn[i][offset]=TRUE;
                ++nVotes[offset];
/*		
		if (bExitAllowed && (nVotes[offset]>=CON_likely_zoomin)) {
		  boole bLikelyZoomin;
		  SeeHowManyManifoldsCrossCertainRegion(bLikelyZoomin,q.vv);
		  if (bLikelyZoomin) {
		    printf("search initiated without maximum guarantee\n");
	            zoomin_voxel=q.vv;
	            p[i].Add(children[i]);	//make sure this latest point can be used as seed bacteria
	            return;
		  }
		}
*/		
		/*
	        if (nVotes[offset]==CON_n_manifolds) {
	          zoomin_voxel=q.vv;
	          p[i].Add(children[i]);	//make sure this latest point can be used as seed bacteria
	          return;
	        }
		*/
		
              }
            }
	  }
	}
      }
      p[i].Add(children[i]);

//printf("%d\n",p[i].size);
      }
    }

    //::: determine if manifolds have been explored/filled-out well enough
    FORi(CON_n_manifolds) {
      if (!bManifoldOn[i]) continue;
//      if (!((in_region[i].size>=20) ||
//            ((in_region[i].size>=4)&&(p[i].size==1))))
//        bSomeManifoldsNotExploredWellEnough=TRUE;
      if (!((in_region[i].size>=CON_initial_explore_threshold) ||
            ((in_region[i].size>=40)&&(p[i].size==1)))) {
        bSomeManifoldsNotExploredWellEnough=TRUE;
//	printf("manifold %d not filled-out enough (size=%d)\n",i,in_region[i].size);
      }
    }
    if (loser_level>1500) throw GeneticFillOutManifoldGroup3SpaceError("too many votes without zoom in");
    SCREENPUT printf("loser level=%5d    losers=%3d   same level=%d  lll=%d\r",loser_level,losers.size,nSameLoserLevel,last_loser_level);
    fflush(stdout);
    if (last_loser_level==loser_level) ++nSameLoserLevel; else { nSameLoserLevel=0; last_loser_level=loser_level; }
    if ((nSameLoserLevel>200)&&(losers.size==1)) {
      FORLISTnow(losers,int)
        bManifoldOn[now]=FALSE;
      }
    }

    if (loop>15) nFertile+=1;
    ++loop;

    SCREENPUT if (loop%50==49) OutputManifolds(String("genetic.")+loop+".iv",p,true_alpha,&bManifoldOn);
    SCREENPUT if (loop%100==99) OutputVoting("vote.iv",nVotes,CON_vote_dim,region,CON_n_manifolds);

    //::: when each manifold has been sufficiently "fleshed out"...
//    if ((!bExitAllowed)&&(!bSomeManifoldsNotExploredWellEnough)) {
    if (!bSomeManifoldsNotExploredWellEnough) {
      if (!bExitAllowed) SCREENPUT BANNER("exit now allowed");

      if (!bExitAllowed) TryToFindLinearSolution();	// only want to try this once...

      bExitAllowed=TRUE;
      int maxvotes=-1;
      VoxelCoord3Space maxvc, vc;
      for (vc.x1=0; vc.x1<CON_vote_dim; ++vc.x1)
        for (vc.x2=0; vc.x2<CON_vote_dim; ++vc.x2)
          for (vc.x3=0; vc.x3<CON_vote_dim; ++vc.x3) {
            int offset=index_nvotes(vc);
            if (nVotes[offset]>maxvotes) {
              maxvotes=nVotes[offset];
	      maxvc=vc;
            }
          }
      if (maxvotes>=CON_likely_zoomin) {
        boole bLikelyZoomin;
        SeeHowManyManifoldsCrossCertainRegion(bLikelyZoomin,maxvc);
        if (bLikelyZoomin) {
          zoomin_voxel=maxvc;
          return;
        }
      }
    }
  }  //infinite loop
}

//////////////////////////////////////////////////////////////////////
// check if manifolds are linear yet; if so, can find intersection in one step
//////////////////////////////////////////////////////////////////////
void GeneticFillOutManifoldGroup3Space::TryToFindLinearSolution()
{
  if (nZoominsPerformed<=3) return;

  const real CON_sqquickzoomratio=5e3;
  int nLinearManifolds=0;
  LinkedList<rVector> norm;
  LinkedList<AlphaCoord> shifts;
  norm.FreeMem();
  shifts.FreeMem();
  FORi(CON_n_manifolds) {
    if (!bManifoldOn[i]) continue;

    //::: find center C of points (for stability...any point in set should work)
    const int n=p[i].size;
    AlphaCoord C(0,0,0);
    FORLIST(p[i]) C+=p[i].Now().a;
    C/=n;

    //::: let Qi=Pi-C (treat C as origin...); create matrix M=[Q0 ... Qn]^T; normalize Qi's of course
    rMatrix M(n,3);
    int k=0;
    FORLIST(p[i]) {
      AlphaCoord q=p[i].Now().a-C;
      q/=q.Norm();
      M[k][0]=q.x;	M[k][1]=q.y;	M[k][2]=q.z;
      ++k;
    }

    //::: M has 1 null eigenvector when points form linear subspace; use svd on (M^T * M)
    rMatrix US,V;
    rVector sqS;
    rMatrix Q=(M.TransposeOf())*M;
    Q/=MatrixNorm(Q);
    Q.partial_Svd(US,V,sqS);

    //::: if smallest eigenvalue is small enough, assume linearity
//	if ((sqS[1]<CON_sqthresh)&&(sqS[2]<CON_sqthresh)) {
//printf("if %e > %e\n",sqS[1]/sqS[2],CON_sqquickzoomratio);
    if (sqS[1]/sqS[2]>CON_sqquickzoomratio) {
      rVector* m1=new rVector; *m1=V.GetColumn(V.cols-1);
      norm.Add(m1);
      ++nLinearManifolds;
      shifts.Add(new AlphaCoord(C));
    }
  }

  //::: if enough manifolds are linear, find common intersection point; algorithm done!
//  if (nLinearManifolds>=min(6,CON_likely_zoomin)) {
  if (nLinearManifolds>=CON_likely_zoomin) {
//  if (FALSE) {
    SCREENPUT printf("%d linear manifolds\n",nLinearManifolds);
    //::: each manifold i has (at least) 1 normal now: m1i; also a displacement Ci
    //::: answer A is such that (A-Ci) perp to m1i (for all i)
    //::: form matrix MM=[m10 m11 ... ]^T, NN=[m10*C0 m11*C1 ... ], and M=[MM -NN]
    rMatrix M(norm.size,4);
    int inx=0;
    shifts.Reset();
    FORLIST(norm) {
      const AlphaCoord& c=shifts.Now();
      const rVector& m1=norm.Now();
      FORi(3) M[inx][i]=m1[i];
      M[inx][3]=-(c.x*m1[0]+c.y*m1[1]+c.z*m1[2]);
      ++inx;
      shifts.Bump();
    }

    //::: null eigenvector of (M^T * M) is [A 1]^T up to a scalar
    rMatrix Q=(M.TransposeOf())*M;
//rVector true_ans(4);
//true_ans[0]=true_alpha.x;  true_ans[1]=true_alpha.y;  true_ans[2]=true_alpha.z;
//true_ans[3]=1;
//SCREENPUT (M*true_ans).Dump("residue using true_alpha");
    rVector v;
    real res;
    Q.FindNullEigenvector(v,res);
    v/=v[3];
    AlphaCoord ans(v[0],v[1],v[2]);
    SCREENPUT ans.Dump("final answer");
    SCREENPUT true_alpha.Dump("true alpha");
    AlphaCoord diff=ans-true_alpha;



// make into function
////////////////////////////////
  //::: KK is important; true internal calibration matrix (can be found from
  //::: tab's) transforms affine reconstruction (W) into metric reconstruction
  Array<Transform> allTabs(E.size-1);
  FORi(E.size) if (i>0) {
    Transform T0i=(T0i_proj[i]+Transform::Outer(E[i],ans));
    allTabs[i-1]=(normalizer.Inverse()*T0i*normalizer);
    allTabs[i-1].Normalize();
  }
  Transform absconic;
  MiscUtility::FindAbsoluteConicFromRelativeCalibrations(absconic,allTabs);
  rMatrix rmKK(&absconic[0][0],3,3);
  rmKK/=(rmKK[0][0]);
  rmKK=rmKK.InverseOf();
  int bGoodCholesky=rmKK.Cholesky();
  if (!bGoodCholesky) { SCREENPUT ALERT("Cholesky decomposition failed\n"); return; }
  rmKK=rmKK.InverseOf();
  Transform baseKK(rmKK.data);
  Transform answer=baseKK;
////////////////////////////////


    real frob_err, frob_err2;
    if (pInternal!=NULL) {
      Transform trueK=(normalizer.Inverse()*(*pInternal));
      Transform diff=(answer/answer.Norm())-(trueK/trueK.Norm());
      frob_err=diff.Norm();
      diff[0][2]=diff[1][2]=0;	// use to just check error in main components
      frob_err2=diff.Norm();
    } else
      frob_err=frob_err2=-1;



    SCREENPUT diff.Dump("difference");
    SCREENPUT printf("distance = %f\n",frob_err);
    SCREENPUT printf("residue= %e\n",res);
    SCREENPUT printf("%d zoomins performed\n",nZoominsPerformed);
    ReconstructScene(ans);
    if (res<1e-6) {
      int nManifoldsStillOn=0;
      FORi(CON_n_manifolds) if (bManifoldOn[i]) ++nManifoldsStillOn;

      GLOBAL_testrun_data.error_frobenius_main=frob_err;
      GLOBAL_testrun_data.error_frobenius=frob_err2;
//      GLOBAL_testrun_data.error_per_entry=diff;	// meaning has changed to entries of K
      GLOBAL_testrun_data.n_zoomins=nZoominsPerformed;
      GLOBAL_testrun_data.n_manifolds_at_end=nManifoldsStillOn;
      GLOBAL_testrun_data.n_total_manifold_points_generated=ManifoldPointStruct3Space::nBorn;
      GLOBAL_testrun_data.n_votes_to_zoomin=CON_likely_zoomin;
      GLOBAL_testrun_data.n_vote_voxels_per_side=CON_vote_dim;
      GLOBAL_testrun_data.n_bacteria_voxels_per_side=CON_bacteria_dim;
//      GLOBAL_testrun_data.min_flesh_out_level=CON_initial_explore_threshold;

      throw GeneticFillOutManifoldGroup3SpaceError("SUCCESS!",0);
    }
  }
}

//////////////////////////////////////////////////////////////////////
void GeneticFillOutManifoldGroup3Space::ReconstructScene(
  const AlphaCoord&		a
)
{
  //::: KK is important; true internal calibration matrix (can be found from
  //::: tab's) transforms affine reconstruction (W) into metric reconstruction
  Array<Transform> allTabs(E.size-1);
  FORi(E.size) if (i>0) {
    Transform T0i=(T0i_proj[i]+Transform::Outer(E[i],a));
//    Transform T0i=(T0i_proj[i]+Transform::Outer(E[i],true_alpha));
    allTabs[i-1]=(normalizer.Inverse()*T0i*normalizer);
    allTabs[i-1].Normalize();
  }
  Transform absconic;
  MiscUtility::FindAbsoluteConicFromRelativeCalibrations(absconic,allTabs);
  rMatrix rmKK(&absconic[0][0],3,3);
  rmKK/=(rmKK[0][0]);
//  rmKK.Dump("trying to inv-Cholesky-inv factor this");  
  rmKK=rmKK.InverseOf();
  int bGoodCholesky=rmKK.Cholesky();
  if (!bGoodCholesky) { SCREENPUT ALERT("Cholesky decomposition failed\n"); return; }
  rmKK=rmKK.InverseOf();
  Transform baseKK(rmKK.data);
///  baseKK.DumpNormalized("final K");  
///  if (pInternal!=NULL) (normalizer.Inverse()*(*pInternal)).DumpNormalized("true K");
  SCREENPUT (baseKK/baseKK[2][2]).Dump("final K");  
  if (pInternal!=NULL) {
    Transform trueK=(normalizer.Inverse()*(*pInternal));
    SCREENPUT (trueK/trueK[2][2]).Dump("true K");
  }
  Transform T_norm0_met=(normalizer*baseKK).Inverse();
  Transform T_met_norm0=T_norm0_met.Inverse();

  //::: make reconstructions
  Array<rPoint3> W;
  FORi(E.size) if (i>0) {
    W.FreeMem();
    Transform Tab=T0i_proj[i]+Transform::Outer(E[i],a);
////    Tab.Dump((String("T0_")+String(i,"%02d")).Ptr());
//    MiscUtility::AffineTriangulate_WithInternalParams(W,Tab,T_met_norm0,E[i],(*GLOB_Kpts)[0],(*GLOB_Kpts)[i]);

{
Array<rMatrix> cams(E.size);
FORi(cams.size) {
  cams[i].SetSize(3,4);
  Transform Tab=T0i_proj[i]+Transform::Outer(E[i],a);
  FORj(3) FORk(3) cams[i][j][k]=Tab[j][k];
  cams[i][0][3]=E[i].x;
  cams[i][1][3]=E[i].y;
  cams[i][2][3]=E[i].z;
}
cams[0].FillWith(0);
cams[0][0][0]=cams[0][1][1]=cams[0][2][2]=1;
rMatrix KKK(4,4);
KKK.MakeIdentity();
FORi(3) FORj(3) KKK[i][j]=T_met_norm0[i][j];
FORi(cams.size) cams[i]=cams[i]*KKK;
MiscUtility::Triangulate(W,cams,(*GLOB_Kpts));
/*
//::: write metric camera matrices for future use
FILE* file=fopen((String("recon.cameras.")+String(nZoominsPerformed,"%02d")).Ptr(),"wb");
ASSERT(file,"err0934034");
fprintf("%d\n",cams.size);
FORi(cams.size) {
  fprintf(file,"\n");
  FORj(3) fprintf(file,"%.9e %.9e %.9e %.9e\n",cams[i][j][0],cams[i][j][1],cams[i][j][2],cams[i][j][3]);
}
fclose(file);
*/
}

////    FORj(W.size) W[j]=T_norm0_met*W[j];	// turn affine recon into metric recon
    SCREENPUT InventorFile::OutputPoints(String("recon")+String(i,"%02d")+".iv",W,RGB3Byte_green);
    
    //::: output 3D --> 2D correspondence for each view (for later texture mapping)
    if (i==1) {		//only doing first pair for now
      FILE* file=fopen("given.dense.matching","rb");
      if (file) {
        int nViews, nPts;
        fscanf(file,"%d\n",&nPts);
        fscanf(file,"%d\n",&nViews);
	Array2D<rPoint3> dense(nViews,nPts);
        Transform T_normi_i=normalizer.Inverse();
	FORj(nViews) {
	  FORk(nPts) {
	    fscanf(file,"%lf %lf\n",&dense[j][k].x,&dense[j][k].y);
	    dense[j][k].z=1.0;
	    dense[j][k]=normalizer*dense[j][k];
	    dense[j][k]/=dense[j][k].z;
	  }
	  fscanf(file,"\n");
	}
        fclose(file);
//	Transform T_met_0=normalizer.Inverse()*T_met_norm0;
        MiscUtility::AffineTriangulate_WithInternalParams(W,Tab,T_met_norm0,E[i],dense[0],dense[i]);
        file=fopen((String("mapping.")+String(i,"%02d")+".0").Ptr(),"w");
        FORk(W.size) {
          rPoint3 p=T_normi_i*dense[0][k];
	  p/=p.z;
          fprintf(file,"%8e  %8e  %8e      %8e  %8e\n",W[k].x,W[k].y,W[k].z,p.x,p.y);
        }
        fclose(file);
      } else {
        Transform T_normi_i=normalizer.Inverse();
        FILE* file2=fopen((String("mapping.")+String(i,"%02d")+".0").Ptr(),"w");
        FORk(W.size) {
          rPoint3 p=T_normi_i*((*GLOB_Kpts)[0][k]);
	  p/=p.z;
          fprintf(file2,"%8e  %8e  %8e      %8e  %8e\n",W[k].x,W[k].y,W[k].z,p.x,p.y);
        }
        fclose(file2);
      }
    }

    if (pInternal!=NULL) {
      //::: use known internal calibration (NOTE: assumes already multiplied by Tinorm)
//      MiscUtility::AffineTriangulate_WithInternalParams(W,Tab,normalizer*(*pInternal),E[i],(*GLOB_Kpts)[0],(*GLOB_Kpts)[i]);
      MiscUtility::AffineTriangulate_WithInternalParams(W,T0i_proj[i]+Transform::Outer(E[i],true_alpha),*pInternal,E[i],(*GLOB_Kpts)[0],(*GLOB_Kpts)[i]);

/*
printf("i=%d\n",i);
pInternal->Dump("@@ pInternal");
E[i].Dump("@@ E[i]");
T0i_proj[i].Dump("@@ T0i_proj");
(T0i_proj[i]+Transform::Outer(E[i],true_alpha)).Dump("@@ calc Tab");
true_alpha.Dump("@@ true_alpha");
FORj(5) (*GLOB_Kpts)[0][j].Dump();
printf("---------------\n");
FORj(5) (*GLOB_Kpts)[i][j].Dump();
*/

      SCREENPUT InventorFile::OutputPoints(String("true.recon")+String(i,"%02d")+".iv",W,RGB3Byte_green);
//throw GeneticFillOutManifoldGroup3SpaceError();
    }
  }
}

//////////////////////////////////////////////////////////////////////
void GeneticFillOutManifoldGroup3Space::ZoomIn()
{
  static int nZoomins=0;
//    region=zoomin_voxel.FindVolume(region,CON_vote_dim);
//    AlphaCoord new_cent=zoomin_voxel.a;	// .a isn't really the center, is it
//    AlphaCoord new_cent=zoomin_voxel.vv.FindCenterInSpace(region,CON_vote_dim);
  AlphaCoord new_cent(0,0,0);
  real weight=0;
/*  
  for (int i=zoomin_voxel.x1-1; i<=zoomin_voxel.x1+1; ++i)
    for (int j=zoomin_voxel.x2-1; j<=zoomin_voxel.x2+1; ++j)
      for (int k=zoomin_voxel.x3-1; k<=zoomin_voxel.x3+1; ++k)
        if (within(i,0,CON_vote_dim)&&within(j,0,CON_vote_dim)&&within(k,0,CON_vote_dim)) {
          int offset=index_nvotes(VoxelCoord3Space(i,j,k));
          if ((nVotes[offset]==nmanifolds_in_zoomin)||
	      (nVotes[offset]==nmanifolds_in_zoomin-1)) {
	    new_cent+=VoxelCoord3Space(i,j,k).FindCenterInSpace(region,CON_vote_dim)*nVotes[offset];
	    weight+=nVotes[offset];
	  }
        }
*/
  int max_votes=-1;
  for (int i=zoomin_voxel.x1-2; i<=zoomin_voxel.x1+2; ++i)
    for (int j=zoomin_voxel.x2-2; j<=zoomin_voxel.x2+2; ++j)
      for (int k=zoomin_voxel.x3-2; k<=zoomin_voxel.x3+2; ++k)
        if (within(i,0,CON_vote_dim)&&within(j,0,CON_vote_dim)&&within(k,0,CON_vote_dim)) {
          int offset=index_nvotes(VoxelCoord3Space(i,j,k));
          if (nVotes[offset]>max_votes) max_votes=nVotes[offset];
        }
//printf("max votes=%d\n",max_votes);
//zoomin_voxel.Dump();
  for (int i=zoomin_voxel.x1-2; i<=zoomin_voxel.x1+2; ++i)
    for (int j=zoomin_voxel.x2-2; j<=zoomin_voxel.x2+2; ++j)
      for (int k=zoomin_voxel.x3-2; k<=zoomin_voxel.x3+2; ++k)
        if (within(i,0,CON_vote_dim)&&within(j,0,CON_vote_dim)&&within(k,0,CON_vote_dim)) {
          int offset=index_nvotes(VoxelCoord3Space(i,j,k));
          if (nVotes[offset]>=max_votes/2) {
	    real w=SQ(real(nVotes[offset]));
	    new_cent+=VoxelCoord3Space(i,j,k).FindCenterInSpace(region,CON_vote_dim)*w;
	    weight+=w;
	  }
        }
  if (weight==real(0)) throw GeneticFillOutManifoldGroup3SpaceError("huge problem -- weight is 0");
  new_cent=new_cent*(1.0/weight);
  
  //::: calc new, smaller zoomin region  
  old_region=region;
  AlphaCoord half_new_width=(region.max-region.min)*(0.5)*(0.5);
  region.min=new_cent-half_new_width;
  region.max=new_cent+half_new_width;
  SCREENPUT printf("current zoomin region:\n");
  SCREENPUT region.Dump();
  rPoint3 center=(region.max+region.min)*(0.5);
  SCREENPUT (center).Dump("center");
  SCREENPUT true_alpha.Dump("true alpha");
  ++nZoominsPerformed;

  if (half_new_width.Norm()<1)		//if we've zoomed in enough...
    ReconstructScene(center);

  SCREENPUT OutputVoting(String("vote.at.zoomin.")+(nZoomins++)+".iv",nVotes,CON_vote_dim,old_region,CON_n_manifolds,&region,&true_alpha);
//  OutputManifolds("manifolds.at.zoomin.iv",p,true_alpha);
  ClearTabulators();
  SeedNewRegion();
}

// find seeds for new region from previous list of bacteria manifold points
// (this is original and simpler version)
//////////////////////////////////////////////////////////////////////
void GeneticFillOutManifoldGroup3Space::SeedNewRegionFromBacteria()
{
  LinkedList_ManifoldPointStruct3Space new_p;
  FORi(CON_n_manifolds) {
    if (!bManifoldOn[i]) continue;
    new_p.FreeMem();
    FORLIST(p[i]) {
      ManifoldPointStruct3Space& now=p[i].Now();
      if (region.bContains(now.a)) {
        now.RecalcVoxels(region,CON_bacteria_dim,CON_vote_dim);	//crucial bug--voxels were based on old larger region
        int offset=index_btaken(now.vb);
	if (!bTaken[i][offset]) {
          new_p.Add(now.Dopple());
          bTaken[i][offset]=TRUE;
          int offset=index_nvotes(now.vv);
          if (!bVotedOn[i][offset]) {
            bVotedOn[i][offset]=TRUE;
            ++nVotes[offset];
          }
	}
      }
    }
    p[i].FreeMem(); p[i].Add(new_p);
    if (p[i].size==0) {
      throw GeneticFillOutManifoldGroup3SpaceError("bug -- no seed in zoomin region for some manifold");
    }
  }
}

// find seeds for new region from list of every manifold point in previous region
//////////////////////////////////////////////////////////////////////
void GeneticFillOutManifoldGroup3Space::SeedNewRegion()
{
  LinkedList_ManifoldPointStruct3Space new_p;
  LinkedList_ManifoldPointStruct3Space in_new_region;
  SCREENPUT printf("seeding new region; listing pts in region for each manifold (before reseeding)\n");
  FORi(CON_n_manifolds) {
    if (!bManifoldOn[i]) continue;
    new_p.FreeMem();
    in_new_region.FreeMem();
    SCREENPUT printf("manif=%3d  pts=%6d\n",i,in_region[i].size);
    FORLIST(in_region[i]) {
      ManifoldPointStruct3Space& now=in_region[i].Now();
      if (region.bContains(now.a)) {
        in_new_region.Add(now.Dopple());
        now.RecalcVoxels(region,CON_bacteria_dim,CON_vote_dim);	//crucial bug--voxels were based on old larger region
        int offset=index_btaken(now.vb);
	if (!bTaken[i][offset]) {
          new_p.Add(now.Dopple());
          bTaken[i][offset]=TRUE;
          int offset=index_nvotes(now.vv);
          if (!bVotedOn[i][offset]) {
            bVotedOn[i][offset]=TRUE;
            ++nVotes[offset];
          }
	}
      }
    }
    in_region[i].FreeMem(); in_region[i].Add(in_new_region);	//can do faster by deleting outsiders as go
    p[i].FreeMem(); p[i].Add(new_p);
    if (p[i].size==0) {
      throw GeneticFillOutManifoldGroup3SpaceError("bug -- no seed in zoomin region for some manifold");
    }
  }
}

//////////////////////////////////////////////////////////////////////
void GeneticFillOutManifoldGroup3Space::ListCollapsedManifolds()
{
  SCREENPUT BANNER("listing collapsed manifolds");
  int nCollapsed=0;
  FORi(CON_n_manifolds) {
    if (!bManifoldOn[i]) {
//      printf("manifold %d is off\n",i);
      continue;
    }
    if ((in_region[i].size>20)&&(p[i].size==1)) {	//20 is arbitrary
      SCREENPUT printf("manifold %d is probably collapsed\n",i);
      ++nCollapsed;
    }
  }
  SCREENPUT printf("%d collapsed\n",nCollapsed);
}

//////////////////////////////////////////////////////////////////////
void PrintOnOff(const boole b)
{
  SCREENPUT if (b) printf("ON\n"); else printf("OFF\n");
}

//////////////////////////////////////////////////////////////////////
void GeneticFillOutManifoldGroup3Space::RunStateReport()
{
  SCREENPUT { printf("range dithering is: ");  PrintOnOff(bRangeDitherOn); }
  SCREENPUT { printf("help losers is:     ");  PrintOnOff(bHelpLosersOn); }
  SCREENPUT { printf("likely zoomin:      %d\n", CON_likely_zoomin); }
}

//////////////////////////////////////////////////////////////////////
void GeneticFillOutManifoldGroup3Space::Run(
  const Array2D<boole>&	bPairOn
)
{
  SCREENPUT BANNER("starting search algorithm");
  
  bManifoldOn.SetSize(CON_n_manifolds);
  FORi(CON_n_manifolds) bManifoldOn[i]=FALSE;
  int inx=0;
  FORi(bPairOn.rows) FORj(bPairOn.cols) if (i<j) {
    bManifoldOn[inx]=bPairOn[i][j];
    ++inx;
  }
  nZoominsPerformed=0;
  
  RunStateReport();

  SCREENPUT true_alpha.Dump("true alpha");

//  region=VolumeStruct(AlphaCoord(-20,-20,-20), AlphaCoord(20,20,20));
  region=VolumeStruct(AlphaCoord(-100,-100,-100), AlphaCoord(100,100,100));
region.min*=0.1; region.max*=0.1;
  nmanifolds_in_zoomin=CON_n_manifolds;

  p.SetSize(CON_n_manifolds);
  in_region.SetSize(CON_n_manifolds);
  every_pt.SetSize(CON_n_manifolds);
  dither_range.SetSize(CON_n_manifolds);
  FORi(dither_range.size) {
    dither_range[i].kmax=1e20; dither_range[i].kmin=-1e20;
    dither_range[i].thmax=PI;  dither_range[i].thmin=-PI;
    dither_range[i].CalcDitherRange(); dither_range[i].bSet=FALSE;
//    dither_range[i].Dump();
  }

  //::: allocate tabulators
  bTaken.SetSize(CON_n_manifolds);
  FOREACHi(bTaken) bTaken[i]=new boole[CON_cube_bacteria_dim];
  bVotedOn.SetSize(CON_n_manifolds);
  FOREACHi(bVotedOn) bVotedOn[i]=new boole[CON_cube_vote_dim];
  nVotes=new word[CON_cube_vote_dim];

  SCREENPUT printf("finding initial seeds\n");
  SeedInitialRegion();

  ManifoldPointStruct3Space::nBorn=0;	//many "born" during seeding, but could be done more efficiently so not considering them

  SCREENPUT printf("beginning zoom in process\n");
  INFINITELOOP {
    FindBestZoomInVoxel();
    ListCollapsedManifolds();
    ZoomIn();
  }
}


