#ifndef INCLUDED_GeneticFillOutManifoldGroup3Space
#define INCLUDED_GeneticFillOutManifoldGroup3Space

///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////

#include "stdmacros.h"
#include "Array.h"
#include "Array2D.h"
#include "LinkedList.h"
#include "ManifoldPointStruct.3Space.h"

///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////

typedef LinkedList<ManifoldPointStruct3Space> LinkedList_ManifoldPointStruct3Space;

// structure so a group of tightly-related functions can share common variables easily
//////////////////////////////////////////////////////////////////////
struct GeneticFillOutManifoldGroup3Space {
  //::: info that must be passed in to call the algorithm
  const Array<ViewPairInfo>&	pairinfo;		//holds epipolar geometry info for each pair in use
  const Array<Transform>&	T0i_proj;		//this combined with E form all camera matrices in common projective basis
  const Array<rPoint3>&		E;
  const AlphaCoord&		true_alpha;
  const Transform&		normalizer;
  Array2D<rPoint3>*		GLOB_Kpts;		//simple kludge used in outputting affine reconstructions
  Transform* const		pInternal;		//allows true internal params to be passed in for testing

  //::: local constants
  const int			CON_n_manifolds;	//just pairinfo.size, renamed to make code clearer
  const int			CON_vote_dim;		//number of vote voxels in region under consideration
  const int			CON_bacteria_dim;	//number of bacteria voxels in region under consideration
  const int			CON_sq_vote_dim;
  const int			CON_cube_vote_dim;
  const int			CON_sq_bacteria_dim;
  const int			CON_cube_bacteria_dim;
  const int			CON_likely_zoomin;

  //::: local variables
  Array<LinkedList_ManifoldPointStruct3Space>	p;		//bacteria manifold points (well-spaced)
  Array<LinkedList_ManifoldPointStruct3Space>	every_pt;	//every manifold pt randomly generated that lies in maximum search region
  Array<LinkedList_ManifoldPointStruct3Space>	in_region;	//every manifold pt randomly generated that lies in current search region
  Array<boole*>			bTaken, bVotedOn;
  word*				nVotes;
  VoxelCoord3Space		zoomin_voxel;
  VolumeStruct			region, old_region, maximum_region;
  Array<DitherRangeStruct>	dither_range;
  Array<boole>			bManifoldOn;
  int				nmanifolds_in_zoomin;
  boole				bRangeDitherOn;
  boole				bHelpLosersOn;
  int				nZoominsPerformed;

  //::: methods / local functions
  void ClearTabulators();
  void SeedInitialRegion();
  void FindBestZoomInVoxel();
  void ZoomIn();
  void SeedNewRegion();
  void SeedNewRegionFromBacteria();
  void Run(const Array2D<boole>&);
//  boole MAMTDP_Bisect(real&, real&, const rPoint3&, const VoxelCoord3Space&,
//    const real&, const real&, const int, const int);
//  boole MoveAlongManifoldTowardsDesiredPosition(ManifoldPointStruct3Space&,
//    const rPoint3&, const VoxelCoord3Space&, const int);
  boole MoveAlongManifoldTowardsDesiredPosition2(ManifoldPointStruct3Space&,
    const VoxelCoord3Space&, const AlphaCoord&, const int);
  void SeeHowManyManifoldsCrossCertainRegion(boole&, const VoxelCoord3Space&);
  void ListCollapsedManifolds();
  void RunStateReport();
  void FindCalibrationFromTabs();
  void ReconstructScene(const AlphaCoord& a);
  void TryToFindLinearSolution();

  GeneticFillOutManifoldGroup3Space(const Array<ViewPairInfo>&, const Array<Transform>&,
    const Array<rPoint3>&, const rPoint3&, const Transform&,
    Array2D<rPoint3>*, Transform* const _pInternal=NULL);
  ~GeneticFillOutManifoldGroup3Space();
};

///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////

struct GeneticFillOutManifoldGroup3SpaceError {
  String	msg;
  int		error_code;

  GeneticFillOutManifoldGroup3SpaceError(const String& _msg, const int code=-1) { msg=_msg; error_code=code; }
};

///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////

#endif
