//////////////////////////////////////////////////////////////////////
rPoint3 PositionGivenByDetailedState(
  const MCMC_State_Detailed&	X,
  const Array<Array2D_rPoint3>&	lattice,
  const int			abs_mani_inx,
  const int			local_mani_inx
)
{
  const int n_th=lattice[abs_mani_inx].rows;
  const int n_ka=lattice[abs_mani_inx].cols;
  const rPoint3& aa=lattice[abs_mani_inx][X.t[local_mani_inx]][X.k[local_mani_inx]];
  real tt=X.tt[local_mani_inx];
  real kk=X.kk[local_mani_inx];
  boole b_invert=FALSE;
  if (tt+kk>1.0) {
    tt=1.0-tt;
    kk=1.0-kk;
    b_invert=TRUE;
  }
  rPoint3 ath, aka;
  if (b_invert) {
    if (X.t[local_mani_inx]==0) ath=aa;
    else ath=lattice[abs_mani_inx][X.t[local_mani_inx]-1][X.k[local_mani_inx]];
    if (X.k[local_mani_inx]==0) aka=aa;
    else aka=lattice[abs_mani_inx][X.t[local_mani_inx]][X.k[local_mani_inx]-1];
  }
  else {
    if (X.t[local_mani_inx]+1==n_th) ath=aa;
    else ath=lattice[abs_mani_inx][X.t[local_mani_inx]+1][X.k[local_mani_inx]];
    if (X.k[local_mani_inx]+1==n_ka) aka=aa;
    else aka=lattice[abs_mani_inx][X.t[local_mani_inx]][X.k[local_mani_inx]+1];
  }
  return aa + (ath-aa)*tt + (aka-aa)*kk;
}

//////////////////////////////////////////////////////////////////////
void TriangleContainingDetailedState(
  ClusterTriangle&		TRI,
  boole&			b_tri_defined,
  const MCMC_State_Detailed&	X,
  const Array<Array2D_rPoint3>&	lattice,
  const int			abs_mani_inx,
  const int			local_mani_inx
)
{
  const int n_th=lattice[abs_mani_inx].rows;
  const int n_ka=lattice[abs_mani_inx].cols;
  real tt=X.tt[local_mani_inx];
  real kk=X.kk[local_mani_inx];
  boole b_invert=FALSE;
  if (tt+kk>1.0) {
    tt=1.0-tt;
    kk=1.0-kk;
    b_invert=TRUE;
  }

  rPoint3 A,B,C;
  if (b_invert) {
    if ( (X.t[local_mani_inx]==0)||(X.k[local_mani_inx]==0) ) goto tcds_not_defined_exit;
    B=lattice[abs_mani_inx][X.t[local_mani_inx]-1][X.k[local_mani_inx]];
    C=lattice[abs_mani_inx][X.t[local_mani_inx]][X.k[local_mani_inx]-1];
  }
  else {
    if ( (X.t[local_mani_inx]+1==n_th)||(X.k[local_mani_inx]+1==n_ka) ) goto tcds_not_defined_exit;
    B=lattice[abs_mani_inx][X.t[local_mani_inx]+1][X.k[local_mani_inx]];
    C=lattice[abs_mani_inx][X.t[local_mani_inx]][X.k[local_mani_inx]+1];
  }
  A=lattice[abs_mani_inx][X.t[local_mani_inx]][X.k[local_mani_inx]];
  TRI.Init(A,B,C);
  b_tri_defined=TRUE;
  return;
  
tcds_not_defined_exit:

  b_tri_defined=FALSE;
  return;
}

//////////////////////////////////////////////////////////////////////
Array<int> const* GLOB_draws;		// for speed, so don't need to pass this into Likelihood() each call

//////////////////////////////////////////////////////////////////////
real IntersectionByTriangles::Likelihood(
  const MCMC_State_Detailed&	X
)
{
//  Array<rPoint3*> a00(X.size);
//  FORi(X.size) a00[i]=&lattice[i][X.t[i]][X.k[i]];
  Array<rPoint3> a(X.size);
  FORi(X.size) a[i]=PositionGivenByDetailedState(X,lattice,(*GLOB_draws)[i],i);
  real dist=0, dx,dy,dz;
  FORi(X.size) for (int j=i+1; j<X.size; ++j) {
    const rPoint3& ai=a[i];
    const rPoint3& aj=a[j];
    dx=ai.x-aj.x;
    dy=ai.y-aj.y;
    dz=ai.z-aj.z;
    dist+=dx*dx+dy*dy+dz*dz;
  }
//  return 1.0/(dist*dist);
  return 1.0/dist;		// more exploration this way
//  return 1.0/(dist*dist*dist);
}

//////////////////////////////////////////////////////////////////////
void IntersectionByTriangles::PerformIntersectionByMcmc(
  LinkedList<rPoint3>&		ll_mutuals,
  const Array<int>&		draws,
  const boole			CON_b_show_stats
)
{
  //::: use MCMC to find mutual intersection point of grid/manifolds
  //::: (1) choose initial state X at random; a state is three <theta,kappa>
  //:::     tuples, lattice points on precalc'ed grid
  //::: (2) find likelihood of X
  //::: (3) LOOP
  //:::     (3.1) choose a new state X* from neighbors of X
  //:::     (3.2) find likelihood of X*
  //:::     (3.3) choose random number r in [0,1]
  //:::     (3.4) if r<L(X*)/L(X), set X equal to X* (accept new state)
  //:::
  //::: perform loop a set number of times; keep track of "best" state reached during looping
  //::: MCMC finds expected values, but its properties can be useful in finding global mins

  GLOB_draws=&draws;			// needed for the kludge
  const int CON_n_trials=1;
  for (int loop=0; loop<CON_n_trials; ++loop) {
    //::: (1) choose initial state X at random; a state is three <theta,kappa>
    //:::     tuples, lattice points on precalc'ed grid
    MCMC_State_Detailed X(draws.size), XX, best_X;
    FORi(X.size) {
      X.t[i]=random(1,Theta.cols)-1;
      X.k[i]=random(1,Kappa.cols)-1;
      X.itt[i]=random(0,99);
      X.ikk[i]=random(0,99);
      X.tt[i]=real(X.itt[i])/100;
      X.kk[i]=real(X.ikk[i])/100;
    }

    //::: (2) find likelihood of X
    real X_like=Likelihood(X);

    //::: (3) LOOP
    real best_like=-1;
    for (long kk=0; kk<100000; ++kk) {
//      if (kk%100000==0) printf("COUNT=%ld\n",kk);
      //::: keep track of "best" state
      if (X_like>best_like) {
        best_like=X_like; best_X=X;
//        printf("%e",best_like);
//        FORi(X.size) printf("  <%3d,%3d>",X.t[i],X.k[i]);
//        printf("\n");
  //      P[0][X.t[0]][X.k[0]].Dump();
  //      P[1][X.t[1]][X.k[1]].Dump();
  //      P[2][X.t[2]][X.k[2]].Dump();
      }
      //::: (3.1) choose a new state X* from neighbors of X
  
      //::: discrete sub-lattice transition
      XX=X;
      int Theta_change, Kappa_change;
      int mani_change=random(1,X.size)-1;
      do {
        Theta_change=random(-1,1);
        Kappa_change=random(-1,1);
      } while ((Theta_change==0)&&(Kappa_change==0));
      if (rand01()<0.50) {
        //::: note tt and kk stay same but in adjacent square
        XX.t[mani_change]+=Theta_change+Theta.cols;
        XX.k[mani_change]+=Kappa_change+Kappa.cols;
        XX.t[mani_change]%=Theta.cols;
        XX.k[mani_change]%=Kappa.cols;
      }
      else {
        //::: stay in same square
        XX.itt[mani_change]+=Theta_change+100;
        XX.ikk[mani_change]+=Kappa_change+100;
        XX.itt[mani_change]%=100;
        XX.ikk[mani_change]%=100;
        XX.tt[mani_change]=real(XX.itt[mani_change])/100;
        XX.kk[mani_change]=real(XX.ikk[mani_change])/100;
      }
  
      //::: (3.2) find likelihood of X*
      real XX_like=Likelihood(XX);
      //::: (3.3) choose random number r in [0,1]
      real r=rand01();
      //::: (3.4) if r<L(X*)/L(X), set X equal to X* (accept new state)
      if (r<XX_like/X_like) { X=XX; X_like=XX_like; }	// do this with ptr swapping...
    }
  
    //::: find triangle in each manifold that contains probes defined by best_X
    Array<ClusterTriangle> TRI(best_X.size);
    boole b_defined=TRUE;
    FORi(best_X.size) {
      TriangleContainingDetailedState(TRI[i],b_defined,best_X,lattice,draws[i],i);
      if (!b_defined) break;			// undefined triangle
    }
    if (!b_defined) continue;

    //::: intersect the planes containing the triangles
    boole b_intersect;
    rPoint3 P;
    IntersectionOf3Triangles(b_intersect,P,TRI[0],TRI[1],TRI[2]);

    //::: store answer as a potential mutual intersection pt
    ll_mutuals.Add(new rPoint3(P));
  }
}
