#include <stdio.h>

#include "helpers.h"
#include "Matrix.h"
#include "VrmlFile.h"
#include "IntersectionByTriangles.h"

//////////////////////////////////////////////////////////////////////
typedef LinkedList<rPoint3> LinkedList_rPoint3;

//////////////////////////////////////////////////////////////////////
real CalcModulusConstraintError(
  const rPoint3&		q,
  const int			inx,
  const Array<ViewPairInfo>&	pairinfo,
  const Array<Transform>&	T0i_proj,
  const Array<rPoint3>&		E
)
{
  int ii=pairinfo[inx].i;
  int jj=pairinfo[inx].j;
  Transform M1i, M1j, H1i, H1j;
  rPoint3 e1i, e1j;
  if (ii==0) {
    M1i=Transform_Id;    e1i=rPoint3(0,0,0);
    M1j=T0i_proj[jj];    e1j=E[jj];
    H1i=Transform_Id;
    H1j=M1j+Transform::Outer(e1j,q);
  } else {
    M1i=T0i_proj[ii];    e1i=E[ii];
    M1j=T0i_proj[jj];    e1j=E[jj];
    H1i=M1i+Transform::Outer(e1i,q);
    H1j=M1j+Transform::Outer(e1j,q);
  }
  
  //::: normalize?
  real det=H1i.Determinant();
  real w=pow(fabs(det),1.0/3)*sign(det);
  M1i/=w;
  e1i/=w;
  det=H1j.Determinant();
  w=pow(fabs(det),1.0/3)*sign(det);
  M1j/=w;
  e1j/=w;

  // M1i=[m00 m01 m02; m10 m11 m12; m20 m21 m22];
  // M1j=[M00 M01 M02; M10 M11 M12; M20 M21 M22];
  // e1i=[eix eiy eiz]';
  // e1j=[ejx ejy ejz]';
  
  const real& m00=M1i[0][0];  const real& m01=M1i[0][1];  const real& m02=M1i[0][2];
  const real& m10=M1i[1][0];  const real& m11=M1i[1][1];  const real& m12=M1i[1][2];
  const real& m20=M1i[2][0];  const real& m21=M1i[2][1];  const real& m22=M1i[2][2];

  const real& M00=M1j[0][0];  const real& M01=M1j[0][1];  const real& M02=M1j[0][2];
  const real& M10=M1j[1][0];  const real& M11=M1j[1][1];  const real& M12=M1j[1][2];
  const real& M20=M1j[2][0];  const real& M21=M1j[2][1];  const real& M22=M1j[2][2];
  
  const real& eix=e1i.x;  const real& eiy=e1i.y;  const real& eiz=e1i.z;
  const real& ejx=e1j.x;  const real& ejy=e1j.y;  const real& ejz=e1j.z;
  
  const real& X=q.x;  const real& Y=q.y;  const real& Z=q.z;

  real al=(-ejx*M12*M21+ejx*M11*M22+ejz*M01*M12-ejz*M02*M11+ejy*M02*M21-ejy*M01*M22)*X+(-M20*M02*ejy+M00*ejy*M22+M20*ejx*M12-M00*M12*ejz-M10*ejx*M22+M10*M02*ejz)*Y+(M10*ejx*M21+M00*M11*ejz+M20*M01*ejy-M20*ejx*M11-M00*ejy*M21-M10*M01*ejz)*Z+M20*M01*M12-M00*M12*M21+M00*M11*M22-M10*M01*M22-M20*M02*M11+M10*M02*M21;
  real be=(ejx*M12*m21-ejy*M02*m21-ejx*M11*m22-ejx*m11*M22+eix*M12*M21-eiy*M02*M21+eiy*M01*M22+ejx*m12*M21+ejz*M02*m11+ejz*m02*M11-ejz*M01*m12-ejz*m01*M12-eix*M11*M22+eiz*M02*M11-eiz*M01*M12-ejy*m02*M21+ejy*m01*M22+ejy*M01*m22)*X+M10*M01*m22+M10*m01*M22-M20*m01*M12-M20*M01*m12-m20*M01*M12+m20*M02*M11+M20*m02*M11+M20*M02*m11-m10*M02*M21-M10*m02*M21-M10*M02*m21+m10*M01*M22-M00*m11*M22+M00*m12*M21+M00*M12*m21+m00*M12*M21-m00*M11*M22-M00*M11*m22+(-M00*eiy*M22+M00*m12*ejz+m10*ejx*M22+M10*ejx*m22-M10*M02*eiz-m10*M02*ejz+m20*M02*ejy-m20*ejx*M12+m00*M12*ejz+M00*M12*eiz-M00*ejy*m22-M20*ejx*m12+M20*M02*eiy-m00*ejy*M22-M20*eix*M12+M10*eix*M22-M10*m02*ejz+M20*m02*ejy)*Y+(m10*M01*ejz-M10*eix*M21-M00*m11*ejz+m00*ejy*M21+M00*ejy*m21+M10*m01*ejz-m20*M01*ejy-m00*M11*ejz-M10*ejx*m21-m10*ejx*M21+M20*eix*M11+M20*ejx*m11-M20*m01*ejy-M20*M01*eiy+m20*ejx*M11+M10*M01*eiz-M00*M11*eiz+M00*eiy*M21)*Z;
  real ga=(-ejx*m12*m21+ejx*m11*m22+ejy*m02*m21-ejy*m01*m22-eix*m12*M21-eiy*M01*m22-ejz*m02*m11+ejz*m01*m12+eix*M11*m22+eiy*M02*m21+eiy*m02*M21-eiy*m01*M22-eiz*m02*M11+eiz*m01*M12-eiz*M02*m11+eiz*M01*m12+eix*m11*M22-eix*M12*m21)*X-M10*m01*m22+M20*m01*m12-M20*m02*m11-m20*m02*M11-m20*M02*m11+m20*m01*M12+m20*M01*m12+M10*m02*m21-M00*m12*m21-m10*M01*m22+m10*M02*m21-m10*m01*M22+M00*m11*m22+m10*m02*M21-m00*m12*M21-m00*M12*m21+m00*M11*m22+(-m00*m12*ejz+m10*M02*eiz-m00*M12*eiz+m00*ejy*m22-M10*eix*m22-M20*m02*eiy+m10*m02*ejz+M10*m02*eiz-m10*ejx*m22+m00*eiy*M22-m20*M02*eiy-m20*m02*ejy+m20*eix*M12+M00*eiy*m22-M00*m12*eiz+m20*ejx*m12+M20*eix*m12-m10*eix*M22)*Y+(-M10*m01*eiz-m10*M01*eiz+m10*eix*M21+m10*ejx*m21+M00*m11*eiz-m10*m01*ejz-m00*ejy*m21-m20*eix*M11-M20*eix*m11+M20*m01*eiy+m00*M11*eiz+M10*eix*m21+m00*m11*ejz-m00*eiy*M21+m20*M01*eiy+m20*m01*ejy-m20*ejx*m11-M00*eiy*m21)*Z+m00*m11*M22;
  real ka=(-eiy*m02*m21+eiy*m01*m22+eiz*m02*m11+eix*m12*m21-eix*m11*m22-eiz*m01*m12)*X+(m20*m02*eiy+m10*eix*m22-m20*eix*m12+m00*m12*eiz-m00*eiy*m22-m10*m02*eiz)*Y+(-m20*m01*eiy-m10*eix*m21+m00*eiy*m21+m20*eix*m11+m10*m01*eiz-m00*m11*eiz)*Z-m20*m01*m12+m00*m12*m21-m00*m11*m22+m10*m01*m22-m10*m02*m21+m20*m02*m11;

/*
M1i.Dump();
M1j.Dump();
e1i.Dump();
e1j.Dump();
printf("%e  %e  %e  %e\n",al,be,ga,ka);
EXIT();
*/
  
  real dum=(ga*ga*ga*al-be*be*be*ka);
//if (dum*dum>1e4) dum=1e2;
  return dum*dum;
}

//////////////////////////////////////////////////////////////////////
real CalcModulusConstraintError2(
  const rPoint3&		q,
  const int			inx,
  const Array<ViewPairInfo>&	pairinfo,
  const Array<Transform>&	T0i_proj,
  const Array<rPoint3>&		E
)
{
  int ii=pairinfo[inx].i;
  int jj=pairinfo[inx].j;
  Transform Tab;
  if (ii==0)
    Tab=T0i_proj[jj]+Transform::Outer(E[jj],q);
  else {
    Transform TTi=T0i_proj[ii]+Transform::Outer(E[ii],q);
    Transform TTj=T0i_proj[jj]+Transform::Outer(E[jj],q);
    Tab=TTj*(TTi.Inverse());
  }
///  Tab=(normalizer.Inverse()*TT*normalizer);		// should not affect modulus constraint calc
//  Tab.Normalize();

  const real& a=Tab[0][0];
  const real& b=Tab[0][1];
  const real& c=Tab[0][2];
  const real& d=Tab[1][0];
  const real& e=Tab[1][1];
  const real& f=Tab[1][2];
  const real& g=Tab[2][0];
  const real& h=Tab[2][1];
  const real& i=Tab[2][2];

  //::: al = -w^3, be = w^2(1+exp(i*th)+exp(-i*th)), ga = -w(1+exp(i*th)+exp(-i*th))
  real al=a*e*i+b*f*g+c*d*h-(c*e*g+b*d*i+a*f*h);		// = det(Tab)
  real be=-(c*g+b*d+f*h-i*e-a*e-a*i);
  real ga=a+e+i;
  
//  return fabs(ga*ga*ga*al-be*be*be);		// works in theory, but not normalized so that al==1

  real w=pow(fabs(al),1.0/3)*sign(al);
  return fabs(be/(w*w)-ga/w);
}

// actually uses Kruppa constraints
//////////////////////////////////////////////////////////////////////
real CalcModulusConstraintError3(
  const rPoint3&		q,
  const int			inx,
  const Array<ViewPairInfo>&	pairinfo,
  const Array<Transform>&	T0i_proj,
  const Array<rPoint3>&		E
)
{
  int ii=pairinfo[inx].i;
  int jj=pairinfo[inx].j;
  rPoint3 ei_j;
  Transform Fij;
  Transform Tab;
  if (ii==0) {
    Tab=T0i_proj[jj]+Transform::Outer(E[jj],q);

    ei_j=E[jj];
    Fij=pairinfo[inx].F;
//    (Fij.TransposeOf()*ei_j).Dump();
//    Fij.DumpNormalized();
//    (Transform::Cross(ei_j)*Tab).DumpNormalized();
  } else {
    Transform TTi=T0i_proj[ii]+Transform::Outer(E[ii],q);
    Transform TTj=T0i_proj[jj]+Transform::Outer(E[jj],q);
    Tab=TTj*(TTi.Inverse());
    
    //::: proj cam matrix i is:  [ T0i_proj[i] : E[i] ]
    rPoint3 oci=(T0i_proj[ii].Inverse()*E[ii]);		// (-1)* op cent of cam i in cam 0's basis
    ei_j=E[jj]-(T0i_proj[jj]*oci);			// proj of cam i into cam j
    Fij=Transform::Cross(ei_j)*Tab;
  }
///  Tab=(normalizer.Inverse()*TT*normalizer);		// should not affect modulus constraint calc
//  Tab=Tab/Tab.Determinant();
  Transform absconic=Tab*(Tab.TransposeOf());
  absconic.Normalize();

  Transform cpmat=Transform::Cross(ei_j);
  Transform A=cpmat*absconic*(cpmat.TransposeOf());
  Transform B=Fij*(Fij.TransposeOf());
//absconic.Dump();
//A.Dump("A");
//B.Dump("B");
//EXIT();
  return (A/A.Norm()-B/B.Norm()).Norm();
}

#ifdef NOTDEFINED
//////////////////////////////////////////////////////////////////////
void IntersectionByTriangles::PerformIntersectionByModc(
  LinkedList<rPoint3>&		ll_mutuals,
  const Array<int>&		draws,
  const boole			CON_b_show_stats
)
{
/*
//test that true answer is a zero
rPoint3 ans=true_alpha+rPoint3(1,3,-4)*1e-5;
VOTE_UNNORM(ans);
FORi(pairinfo.size) {
  real err=CalcModulusConstraintError(ans,i,pairinfo,T0i_proj,E);
  printf("%d   %e\n",i,err);
}
EXIT();
*/

  const int CON_n_bins=3;
  Array<LinkedList_rPoint3> bins(CON_n_bins);

  rPoint3 width(10,10,10);
//  real scale=0.000005;
//  rPoint3 q0=true_alpha-width*scale;
//  rPoint3 q1=true_alpha+width*scale;
  real scale=1;
  rPoint3 q0=rPoint3(0,0,0)-width*scale;
  rPoint3 q1=rPoint3(0,0,0)+width*scale;
  real inc=width.x*scale*2/23;
  real min=1e10;
  real max=-1;
  rPoint3 min_pt;
  rPoint3 q=q0;
  while (q.x<q1.x) {
    while (q.y<q1.y) {
      while (q.z<q1.z) {

        rPoint3 qq=q;
        VOTE_UNNORM(qq);
        real err=0;
        FORi(draws.size) err+=CalcModulusConstraintError(qq,draws[i],pairinfo,T0i_proj,E);
/*
if (rand01()<0.0001) {
q.Dump();
Tab.Dump();
     printf("%e   %e    %e    %e\n",al,be,ga,err);
}
*/
        if (err<min) { min=err; min_pt=qq; }
        else if (err>max) max=err;

        int bin_inx;
//        const real MIN=3.126343e-22/2*3;
        const real MIN=8.94e-12;
//        if (err<3.126343e-22) bin_inx=0;
        if (err<MIN) bin_inx=0;
        else {
          real err2=(err)/(MIN);
          real dum=log(err2)/log(10);
          bin_inx=_ceil(dum);
          if (bin_inx>=CON_n_bins) bin_inx=CON_n_bins-1;
        }
        if (bin_inx<CON_n_bins-1)
          bins[bin_inx].Add(new rPoint3(q));
        
        q.z+=inc;
      }
      q.z=q0.z;
      q.y+=inc;
    }
    q.y=q0.y;
    q.x+=inc;
  }
  printf("%e   %e\n",min,max);
  min_pt.Dump("calc min");
  VOTE_UNNORM(true_alpha);
  true_alpha.Dump("true ans");

  const RGB3Byte bin_color[]={ RGB3Byte(255,0,0), RGB3Byte(150,0,150), RGB3Byte(0,0,255) };
  VrmlFile::Open("modulus.constraint.wrl");
  FORi(bins.size) {
    Array<rPoint3> W;
    W=bins[i];
    VrmlFile::OutputPoints(W,bin_color[i]);
  }
  VrmlFile::Close();
  
  
  
  EXIT();
}
#endif

#ifdef NOTDEFINED
//////////////////////////////////////////////////////////////////////
struct LMCMCEStruct {
  Array<int> const*		draws;
  Array<ViewPairInfo> const*	pairinfo;
  Array<Transform> const*	T0i_proj;
  Array<rPoint3> const*		E;
};

//////////////////////////////////////////////////////////////////////
real LM_CalcModulusConstraintError(
  const rVector&		x,
  const void*			con,
  const int			inx
)
{
  LMCMCEStruct* const CON=(LMCMCEStruct*)con;
  rPoint3 q(x[0],x[1],x[2]);
  VOTE_UNNORM(q);
  real err=0;
  FORi(3) err+=CalcModulusConstraintError(q,(*(CON->draws))[i],*(CON->pairinfo),*(CON->T0i_proj),*(CON->E));
  return err;
}

#include "LevenbergMarquardt.h"

//////////////////////////////////////////////////////////////////////
void IntersectionByTriangles::PerformIntersectionByModc(
  LinkedList<rPoint3>&		ll_mutuals,
  const Array<int>&		draws,
  const boole			CON_b_show_stats
)
{
  LMCMCEStruct con;
  con.draws=&draws;
  con.pairinfo=&pairinfo;
  con.T0i_proj=&T0i_proj;
  con.E=&E;

  rPoint3 q0=true_alpha+rPoint3(4,-1,3)*0.001;  
  rVector x(3);
  x[0]=q0.x;
  x[1]=q0.y;
  x[2]=q0.z;

  LevenbergMarquardt(x,&con,&LM_CalcModulusConstraintError,1);

  rPoint3 ans(x[0],x[1],x[2]);
  VOTE_UNNORM(ans);
  ans.Dump("ans");
  VOTE_UNNORM(true_alpha);
  true_alpha.Dump("true");

  EXIT();
}
#endif

//////////////////////////////////////////////////////////////////////
#define calc_err(e,x)			\
  qq=x;					\
  VOTE_UNNORM(qq);			\
  e=0;					\
  FORl(draws.size) e+=CalcModulusConstraintError(qq,draws[l],pairinfo,T0i_proj,E);

//////////////////////////////////////////////////////////////////////
#define calc_err2(e,X)			\
  qq=X;					\
  e=qq.x*qq.x+qq.y*qq.y+qq.z*qq.z;

//////////////////////////////////////////////////////////////////////
void IntersectionByTriangles::GradientDescent(
  rPoint3&			Q,
  const Array<int>&		draws
)
{
  rPoint3 qq;			// needed by calc_err() macro

  const int CON_min_step=0;
  ASSERT(CON_min_step>=0,"CON_min_step cannot be smaller than 0");

  real STEPS[50];
  FORi(50) STEPS[i]=pow(10,real(-20+i)/4-6);  
  int step_inx=15;
  rPoint3 q0=Q;

  rPoint3 q,pos;
  q=q0;

  pos=q;
/*
  pos.Dump("start");
  VOTE_UNNORM(pos);
  real err;
  calc_err(err,q);
  printf("err= %e\n",err);
*/
  
  LinkedList<rPoint3> ll_path;
  ll_path.FreeMem();
  boole b_try_big_step=FALSE;	// once step gets small, this gives minimizer one last chance to take big steps in direction of gradient before halting search

// INFINITELOOP {
  FORj(10000) {

    //::: calc partial derivs
    real step=STEPS[step_inx];
    real scale=step/4;
    makecliphi(scale,1e-8);

    real err0,err1,err100,err010,err001;
    calc_err(err0,q);
    calc_err(err100,q+rPoint3(1,0,0)*scale);
    calc_err(err010,q+rPoint3(0,1,0)*scale);
    calc_err(err001,q+rPoint3(0,0,1)*scale);
  
    rPoint3 grad(err100-err0,err010-err0,err001-err0);
    grad=grad/scale;
  //grad.Dump("grad");
    rPoint3 disp;
    do {
      disp=grad*step;
      step*=0.5;
    } while (disp.Norm()>2);		// so don't take large leaps
    step*=2;
    rPoint3 q1=q-disp;
  
    calc_err(err1,q1);
    real diff=err0-err1;
    if (diff>0) {
      b_try_big_step=FALSE;
      ++step_inx;
      makecliphi(step_inx,49);
      q=q1;
      if (rand01()<-1) {
        grad.Dump("grad");
        rPoint3 ans=q;
        printf("dist = %e\n",(true_alpha-ans).Norm());
        ans.Dump("un pos");
        VOTE_UNNORM(ans);
        ans.Dump("pos");
        printf("%e    %d\n",err1,step_inx);
        ll_path.Add(new rPoint3(q1.x,q1.y,err1));
      }
    }
    else {
  //    if (step_inx>8) step_inx=8;
      --step_inx;
      makecliplo(step_inx,CON_min_step);
    }
  
    if (step_inx==CON_min_step) {
      if (b_try_big_step) {
//        printf("FINISHED EARLY\n");
        break;
      } else {
        b_try_big_step=TRUE;
        step_inx=49;
      }
    }
  } // end main loop

/*
  calc_err(err,q);
  printf("final err= %e\n",err);

  Array<rPoint3> path;
  path=ll_path;
//  VisualizeModcTerrain(true_alpha,0.1,0.01,40,path,draws);
//  VisualizeModcTerrain(true_alpha,0.001,0.002,40,path,draws);
//  VisualizeModcTerrain(q,0.002,0.03,40,path,true_alpha,draws);

  rPoint3 ans=q;
  ans.Dump("final un pos");
  true_alpha.Dump("true");
  VOTE_UNNORM(ans);
  VOTE_UNNORM(true_alpha);
  ans.Dump("final ans");
  true_alpha.Dump("true");
  VOTE_NORM(true_alpha);
*/
 
  Q=q;
}

// sample nearby points and return min
//////////////////////////////////////////////////////////////////////
void IntersectionByTriangles::BruteForceSearch(
  rPoint3&			q0,
  const real&			step,
  const int			perside,
  const Array<int>&		draws
)
{
  rPoint3 qq;			// needed by calc_err() macro
  real ERR;

  rPoint3 mi_pos;
  real mi_err;

//  LinkedList<rPoint3> ll_path;
//  ll_path.Add(new rPoint3(q0));		// must use correct z value...

  mi_err=1e40;
  FORi(perside) {
    FORj(perside) {
      rPoint3 pos(q0.x+(i-real(perside)/2)*step, q0.y+(j-real(perside)/2)*step, q0.z);
      FORk(perside) {
        pos.z=q0.z+(k-real(perside)/2)*step;
        calc_err(ERR,pos);
        if (ERR<mi_err) { mi_err=ERR; mi_pos=pos; }
      }
    }
  }

//  ll_path.Add(new rPoint3(q0.x,q0.y,mi_err));

  q0=mi_pos;
}

//#ifdef NOTDEFINED
// this version does nonlinear minimization; discrete step sizes
//////////////////////////////////////////////////////////////////////
void IntersectionByTriangles::PerformIntersectionByModc(
  LinkedList<rPoint3>&		ll_mutuals,
  const Array<int>&		draws,
  const boole			CON_b_show_stats
)
{
  rPoint3 qq;			// needed by calc_err() macro
  real ERR;

  real dkfja=randpm();
  dkfja=randpm();
  FORi(100) {
//    rPoint3 q0=true_alpha+rPoint3(randpm(),randpm(),randpm())*0.001;
    rPoint3 q0=rPoint3(randpm(),randpm(),randpm())*50;
    FORj(3) {
      BruteForceSearch(q0,1,20,draws);
      GradientDescent(q0,draws);
      BruteForceSearch(q0,0.1,20,draws);
      GradientDescent(q0,draws);
      BruteForceSearch(q0,0.01,20,draws);
      GradientDescent(q0,draws);
      BruteForceSearch(q0,0.001,20,draws);
      GradientDescent(q0,draws);
    }

    calc_err(ERR,q0);
    printf("err = %e\n",ERR);
    q0.Dump();
    true_alpha.Dump("true");
    printf("\n");
    
    if (ERR<1e-1) {
      Array<rPoint3> path;
      rPoint3 base=(q0-true_alpha);
      real base_dist=rPoint3(base.x,base.y,0).Norm();
//      VisualizeModcTerrain(q0,base_dist/20,0.5,100,path,true_alpha,draws);
      VisualizeModcTerrain(true_alpha+rPoint3(1,1,1)*0.001,base_dist/20,0.5,100,path,true_alpha,draws);
//      VisualizeModcTerrain(q0,base_dist/60,0.5,150,path,true_alpha,draws);
      EXIT();
    }
  }

  EXIT();
}
//#endif


#ifdef NOTDEFINED
// brute-force version
//////////////////////////////////////////////////////////////////////
void IntersectionByTriangles::PerformIntersectionByModc(
  LinkedList<rPoint3>&		ll_mutuals,
  const Array<int>&		draws,
  const boole			CON_b_show_stats
)
{
  rPoint3 qq;			// needed by calc_err() macro
  real ERR;

  const real CON_sf=0.75;

  rPoint3 q0(0,0,0);
  real step=0.1;
  int perside=40;
  rPoint3 mi_pos;
  real mi_err;

  LinkedList<rPoint3> ll_path;
//  ll_path.Add(new rPoint3(q0));		// must use correct z value...

  GradientDescent(q0,draws);

  for (int loop=0; loop<15; ++loop) {
//    if (loop==14) { perside=100; step/=(10/4); }
    BruteForceSearch(q0,step,perside,draws);
    calc_err(ERR,q0);
    printf("err = %e\n",ERR);
    q0.Dump();
    step*=CON_sf;
    ll_path.Add(new rPoint3(q0.x,q0.y,mi_err));
  }

  GradientDescent(q0,draws);

  Array<rPoint3> path;
//  path=ll_path;
//  VisualizeModcTerrain(mi_pos,0.05,0.0001,40,path,true_alpha,draws);
  VisualizeModcTerrain(q0,step/CON_sf,0.5,100,path,true_alpha,draws);
//  VisualizeModcTerrain(true_alpha*0.75+mi_pos*0.25,0.05,0.01,40,path,true_alpha,draws);

  rPoint3 ans=q0;
  ans.Dump("final un pos");
  true_alpha.Dump("true");
  real err;
  calc_err(err,true_alpha);
  printf("true err = %e\n",err);
  VOTE_UNNORM(ans);
  VOTE_UNNORM(true_alpha);
  ans.Dump("final ans");
  true_alpha.Dump("true");
  VOTE_NORM(true_alpha);

  EXIT();
}
#endif

//////////////////////////////////////////////////////////////////////
void IntersectionByTriangles::VisualizeModcTerrain(
  const rPoint3&		start,			// centering position
  const real&			step,			// unit length for grid
  const real&			height_ratio,		// scaling of terrain height relative to lattice width; 0.5 is good
  const int			perside,		// size of grid
  const Array<rPoint3>&		path,			// a path to superimpose on terrain plot
  const rPoint3&		true_alpha,		// true min
  const Array<int>&		draws
)
{
ASSERT(height_ratio==0.5,"HR should be 0.5");

  rPoint3 qq;			// needed by calc_err() macro
  real ERR;

  rPoint3 q0=start;
  rPoint3 q=q0;

////  const real step=0.1;
////  const real exagg=.01;
///  const real step=0.01;
///  const real exagg=.01;
//  const real step=0.001;
//  const real exagg=.002;
//  const int perside=40;

  real mi_err=1e40;
  real ma_err=-1e40;
  rPoint3 mi_pos;

  Array2D<rPoint3> lat(perside,perside);
  FORi(perside) {
    FORj(perside) {
/*
      rPoint3 pos(q0.x+(i-real(perside)/2)*step, q0.y+(j-real(perside)/2)*step, q0.z);
      real err;
      calc_err(err,pos);
      lat[i][j]=rPoint3(pos.x,pos.y,err);
*/

      rPoint3 pos(q0.x+(i-real(perside)/2)*step, q0.y+(j-real(perside)/2)*step, q0.z);
      real min_err=1e30;
      FORk(perside) {
        pos.z=q0.z+(k-real(perside)/2)*step;
        calc_err(ERR,pos);
ERR=log(ERR);
        if (ERR<min_err) min_err=ERR;
      }
      lat[i][j]=rPoint3(pos.x,pos.y,min_err);
      if (min_err<mi_err) { mi_err=min_err; mi_pos=lat[i][j]; }
      else if (min_err>ma_err) ma_err=min_err;
    }
  }

  //::: scale lattice so that height equals width
  real height_scale=perside*step/(ma_err-mi_err)*(height_ratio);
  real exagg=height_scale;
  FORi(lat.rows) FORj(lat.cols) lat[i][j].z*=exagg;
  mi_pos.z*=exagg;
  
//  VrmlFile::OutputLattice("blah.wrl",lat);
  VrmlFile::Open("blah.wrl");
  Array<rPoint3> bars;
  bars.SetSize(lat.rows*(lat.cols-1)*2);
  int inx=0;
  FORi(lat.rows) FORj(lat.cols-1) {
    bars[inx++]=lat[i][j];
    bars[inx++]=lat[i][j+1];
  }
  VrmlFile::OutputLines(bars,RGB3Byte_green);
  bars.SetSize((lat.rows-1)*lat.cols*2);
  inx=0;
  FORi(lat.rows-1) FORj(lat.cols) {
    bars[inx++]=lat[i][j];
    bars[inx++]=lat[i+1][j];
  }
  VrmlFile::OutputLines(bars,RGB3Byte_green);
  
  //::: output path
  bars.SetSize((path.size-1)*2);
  inx=0;
  FORi(path.size-1) { bars[inx++]=path[i]; bars[inx++]=path[i+1]; }
  FORi(bars.size) bars[i].z=log(bars[i].z)*exagg;
  VrmlFile::OutputLines(bars,RGB3Byte_yellow);
  
  //::: output lattice minimum
  bars.SetSize(1);
  bars[0]=mi_pos;
  VrmlFile::OutputPoints(bars,RGB3Byte_cyan,8);

  //::: output true_alpha
  calc_err(ERR,true_alpha);
ERR=log(ERR);
  bars[0]=true_alpha;
  bars[0].z=ERR*exagg;
  VrmlFile::OutputPoints(bars,RGB3Byte_yellow,8);
  bars.SetSize(2);
  bars[0]=true_alpha;  bars[0].z=ERR*exagg;
  bars[1]=true_alpha;  bars[1].z=ma_err*exagg;
  VrmlFile::OutputLines(bars,RGB3Byte_yellow);
  
  //::: output axis
  bars.SetSize(2);
  rPoint3 Q=lat[perside/2][perside/2];
  bars[0]=Q;
  bars[1]=Q+rPoint3(1,0,0)*step*5;
  VrmlFile::OutputLines(bars,RGB3Byte_red);
  bars[1]=Q+rPoint3(0,1,0)*step*5;
  VrmlFile::OutputLines(bars,RGB3Byte_cyan);
  bars[1]=Q+rPoint3(0,0,1)*step*5;
  VrmlFile::OutputLines(bars,RGB3Byte_white);
  
  VrmlFile::Close();
}
