#include <stdio.h>

#include "helpers.h"
#include "String.h"
#include "Matrix.h"
#include "VrmlFile.h"
#include "MiscUtility.h"
#include "PlotManifolds.h"
#include "IntersectionByMcmc.h"
#include "MeshGenerationRoutines.h"
#include "ManifoldPointStruct.3Space.h"

//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
IntersectionByMcmc::IntersectionByMcmc(
  const Array<ViewPairInfo>&	_pairinfo,
  const Array<Transform>&	_T0i_proj,
  const Array<rPoint3>&		_E,
  const AlphaCoord&		_true_alpha,
  const Transform&		_Tinorm,
  const rPoint2&		_image_size,
  Array2D<rPoint3>*		_GLOB_Kpts,
  Transform* const		_pInternal
) :
  pairinfo(_pairinfo), T0i_proj(_T0i_proj), E(_E), __true_alpha(_true_alpha),
  normalizer(_Tinorm), image_size(_image_size), GLOB_Kpts(_GLOB_Kpts),
  pInternal(_pInternal),
  
  CON_n_manifolds(pairinfo.size)
{
  true_alpha=__true_alpha;
}

//////////////////////////////////////////////////////////////////////
IntersectionByMcmc::~IntersectionByMcmc()
{
}

#ifdef NOTDEFINED
//////////////////////////////////////////////////////////////////////
void IntersectionByMcmc::NormalizeVotingSpace()
{
  printf("BEGIN NORM\n");  fflush(stdout);

  Array<rPoint3> M(lattice.size*Kappa.cols*Theta.cols);
  int inx=0;
  FORi(lattice.size) FORj(Theta.cols) FORk(Kappa.cols) M[inx++]=lattice[i][j][k];
//  M.Dump();
//  EXIT();

  rMatrix Q(3,M.size);
  rPoint3 median=rPoint3(0,0,0);
  FORi(M.size) median+=M[i];
  median/=M.size;
  FORi(M.size) {
    const rPoint3& p=M[i];
    Q[0][i]=p.x-median.x; Q[1][i]=p.y-median.y; Q[2][i]=p.z-median.z;
  }
  rMatrix QQ=Q*(Q.TransposeOf());	// covariance matrix
  rMatrix U,S,V;
  QQ.full_Svd(U,S,V);
  V=V.TransposeOf();
  //::: transform samples so major axii are principal components
  Transform T(&V[0]);
  T.Invert();
  FORi(M.size) M[i]=T*(M[i]-median);

  //::: find standard deviation in each axis direction
  rPoint3 sd(0,0,0);
  FORi(M.size) {
    sd.x+=SQ(M[i].x-median.x);
    sd.y+=SQ(M[i].y-median.y);
    sd.z+=SQ(M[i].z-median.z);
  }
  sd.x=sqrt(sd.x/M.size);
  sd.y=sqrt(sd.y/M.size);
  sd.z=sqrt(sd.z/M.size);
sd.Dump("standard deviation sd");

  //::: scale width of point set in each axis direction so that width is 1
  rPoint3 max, min;
  max=min=M[0];
  FORi(M.size) {
    const rPoint3& p=M[i];
    if (p.x>max.x) max.x=p.x;
    if (p.y>max.y) max.y=p.y;
    if (p.z>max.z) max.z=p.z;
    if (p.x<min.x) min.x=p.x;
    if (p.y<min.y) min.y=p.y;
    if (p.z<min.z) min.z=p.z;
  }
  Transform scale=Transform(1.0/(max.x-min.x),1.0/(max.y-min.y),1.0/(max.z-min.z));
////  //::: scale each axis so that standard deviation (or max width) is 1 in each direction
////  Transform scale=Transform(1.0/sd.x,1.0/sd.y,1.0/sd.z);
  Transform joe=scale*T;
  ManifoldPointStruct3Space_SetVoteNormalizer(joe,median);

/*
//now done later
  //::: normalize lattice points
  FORi(lattice.size) FORj(Theta.cols) FORk(Kappa.cols)
    VOTE_NORM(lattice[i][j][k]);
*/

/*
  //::: find covariance matrix of sample points
  //::: Q=[ q0 q1 q2 ... ] (==points as column vectors; qi=pi-median_i)
  //::: then PC's are eigenvectors of Q * Q'  ( <-- covariance matrix )
  rMatrix V,S;
  rPoint3 centroid;
  MiscUtility::FindPrincipalComponentsOf3DPointCloud(V,S,centroid,M);

  //::: create 3x3 array that transforms original alpha (voting) space into a space
  //::: where the principle components are orthonormal basis vectors; may also want
  //::: to transform centroid to origin...
  Transform norm;
  FORi(3) FORj(3) norm[i][j]=V[j][i];
norm=norm*Transform(sqrt(S[0][0]),sqrt(S[1][1]),sqrt(S[2][2]));
norm=norm.Inverse();
norm.Dump("norm");
  ManifoldPointStruct3Space_SetVoteNormalizer(norm,centroid);
*/


/*  
  //::: find covariance matrix of sample points
  //::: Q=[ q0 q1 q2 ... ] (==points as column vectors; qi=pi-median_i)
  //::: then PC's are eigenvectors of Q * Q'  ( <-- covariance matrix )
  const Array<rPoint3>& P=M;
  rMatrix Q(3,P.size);
  rPoint3 median=rPoint3(0,0,0);
  FORi(P.size) median+=P[i];
  median/=P.size;
  FORi(P.size) { Q[0][i]=P[i].x-median.x; Q[1][i]=P[i].y-median.y; Q[2][i]=P[i].z-median.z; }
  rMatrix QQ=Q*(Q.TransposeOf());	// covariance matrix
  QQ*=1.0/P.size;
  QQ=QQ.InverseOf();
median.Dump("centroid");
QQ.Dump();

  //::: create 3x3 array that transforms original alpha (voting) space into a space
  //::: where the principle components are orthonormal basis vectors; may also want
  //::: to transform centroid to origin...
  Transform norm;
  FORi(3) FORj(3) norm[i][j]=QQ[i][j];
  norm.Normalize();
  norm.Dump("norm");
  ManifoldPointStruct3Space_SetVoteNormalizer(norm,median);
*/
  printf("END NORM\n");  fflush(stdout);
}

//////////////////////////////////////////////////////////////////////
void IntersectionByMcmc::SetLattice(
  const Array<real>&		min_th,
  const Array<real>&		max_th,
  const Array<real>&		min_ka,
  const Array<real>&		max_ka,
  const int			CON_n_theta,
  const int			CON_n_kappa,
  const int			CON_n_mani
)
{
  Theta.SetSize(CON_n_mani,CON_n_theta);
  Kappa.SetSize(CON_n_mani,CON_n_kappa);
  lattice.SetSize(CON_n_mani);
  b_lattice.SetSize(CON_n_mani);
  FORi(CON_n_mani) lattice[i].SetSize(CON_n_theta,CON_n_kappa);
  FORi(CON_n_mani) b_lattice[i].SetSize(CON_n_theta,CON_n_kappa);

  //::: uniform-sample each manifold
  FORi(CON_n_mani) FORj(CON_n_kappa) Kappa[i][j]=min_ka[i]+(real(j)/(CON_n_kappa-1))*(max_ka[i]-min_ka[i]);
  FORi(CON_n_mani) FORj(CON_n_theta) Theta[i][j]=min_th[i]+(real(j)/(CON_n_theta-1))*(max_th[i]-min_th[i]);

  Transform Tab;
  rPoint3 alpha, dum;
  FORk(lattice.size) {
    FORi(CON_n_theta) FORj(CON_n_kappa) {
      if (!AffineCalibFrom2Params_efficient(Tab,dum,Kappa[k][j],Theta[k][i],pairinfo[k])) EXIT("err0913247");
      PPOMCM_FindAlpha(alpha,Tab,T0i_proj,E,pairinfo[k].i,pairinfo[k].j);
      lattice[k][i][j]=alpha;
      b_lattice[k][i][j]=TRUE;
    }
  }


  rPoint3 corners[]={rPoint3(100,0,0), rPoint3(0,100,0), rPoint3(0,0,100), rPoint3(-100,0,0), rPoint3(0,-100,0) };
  FORk(lattice.size) {
    FORi(Theta.cols) FORj(Kappa.cols) {
      rPoint3& p=lattice[k][i][j];
      if ( (!within(p.x,-100,100))||(!within(p.y,-100,100))||(!within(p.z,-100,100)) ) {
        p=corners[k];
	b_lattice[k][i][j]=FALSE;
      }
    }
  }
}
#endif

//////////////////////////////////////////////////////////////////////
real IntersectionByMcmc::Likelihood(
  const MCMC_State&		X
)
{
  if (X.size==3) {
    const rPoint3& a0=lattice[0][X.t[0]][X.k[0]];
    const rPoint3& a1=lattice[1][X.t[1]][X.k[1]];
    const rPoint3& a2=lattice[2][X.t[2]][X.k[2]];
    real distx=SQ(a0.x-a1.x)+SQ(a1.x-a2.x)+SQ(a2.x-a0.x);
    real disty=SQ(a0.y-a1.y)+SQ(a1.y-a2.y)+SQ(a2.y-a0.y);
    real distz=SQ(a0.z-a1.z)+SQ(a1.z-a2.z)+SQ(a2.z-a0.z);
    real dist=max3(distx,disty,distz);
    return 1.0/(dist*dist*dist);
  }

/*
  Array<rPoint3*> a(X.size);
  FORi(X.size) a[i]=&lattice[i][X.t[i]][X.k[i]];
//  real distx=0, disty=0, distz=0;
//  FORi(X.size) FORjIN(i+1,X.size) distx+=SQ(a[i]->x - a[j]->x);
  real dist=-1, dx, dy, dz, md;
  FORi(X.size) for (int j=i+1; j<X.size; ++j) {
    dx=abs(a[i]->x - a[j]->x);
    dy=abs(a[i]->y - a[j]->y);
    dz=abs(a[i]->z - a[j]->z);
    md=max3(dx,dy,dz);
    if (md>dist) dist=md;
  }
  return 1.0/(dist*dist*dist);
*/

  Array<rPoint3*> a(X.size);
  FORi(X.size) a[i]=&lattice[i][X.t[i]][X.k[i]];
  real dist=0, dx,dy,dz;
  FORi(X.size) for (int j=i+1; j<X.size; ++j) {
    const rPoint3& ai=*a[i];
    const rPoint3& aj=*a[j];
    dx=ai.x-aj.x;
    dy=ai.y-aj.y;
    dz=ai.z-aj.z;
    dist+=dx*dx+dy*dy+dz*dz;
  }
  return 1.0/(dist*dist);
}

//////////////////////////////////////////////////////////////////////
rPoint3 PositionGivenByDetailedState(
  const MCMC_State_Detailed&	X,
  const Array<Array2D_rPoint3>&	lattice,
  const int			mani_inx
)
{
  const int n_th=lattice[mani_inx].rows;
  const int n_ka=lattice[mani_inx].cols;
  const rPoint3& aa=lattice[mani_inx][X.t[mani_inx]][X.k[mani_inx]];
  real tt=X.tt[mani_inx];
  real kk=X.kk[mani_inx];
  boole b_invert=FALSE;
  if (tt+kk>1.0) {
    tt=1.0-tt;
    kk=1.0-kk;
    b_invert=TRUE;
  }
  rPoint3 ath, aka;
  if (b_invert) {
    if (X.t[mani_inx]==0) ath=aa;
    else ath=lattice[mani_inx][X.t[mani_inx]-1][X.k[mani_inx]];
    if (X.k[mani_inx]==0) aka=aa;
    else aka=lattice[mani_inx][X.t[mani_inx]][X.k[mani_inx]-1];
  }
  else {
    if (X.t[mani_inx]+1==n_th) ath=aa;
    else ath=lattice[mani_inx][X.t[mani_inx]+1][X.k[mani_inx]];
    if (X.k[mani_inx]+1==n_ka) aka=aa;
    else aka=lattice[mani_inx][X.t[mani_inx]][X.k[mani_inx]+1];
  }
  return aa + (ath-aa)*tt + (aka-aa)*kk;
}

//////////////////////////////////////////////////////////////////////
void TriangleContainingDetailedState(
  ClusterTriangle&		TRI,
  boole&			b_tri_defined,
  const MCMC_State_Detailed&	X,
  const Array<Array2D_rPoint3>&	lattice,
  const int			mani_inx
)
{
  const int n_th=lattice[mani_inx].rows;
  const int n_ka=lattice[mani_inx].cols;
  real tt=X.tt[mani_inx];
  real kk=X.kk[mani_inx];
  boole b_invert=FALSE;
  if (tt+kk>1.0) {
    tt=1.0-tt;
    kk=1.0-kk;
    b_invert=TRUE;
  }

  rPoint3 A,B,C;
  if (b_invert) {
    if ( (X.t[mani_inx]==0)||(X.k[mani_inx]==0) ) goto tcds_not_defined_exit;
    B=lattice[mani_inx][X.t[mani_inx]-1][X.k[mani_inx]];
    C=lattice[mani_inx][X.t[mani_inx]][X.k[mani_inx]-1];
  }
  else {
    if ( (X.t[mani_inx]+1==n_th)||(X.k[mani_inx]+1==n_ka) ) goto tcds_not_defined_exit;
    B=lattice[mani_inx][X.t[mani_inx]+1][X.k[mani_inx]];
    C=lattice[mani_inx][X.t[mani_inx]][X.k[mani_inx]+1];
  }
  A=lattice[mani_inx][X.t[mani_inx]][X.k[mani_inx]];
  TRI.Init(A,B,C);
  b_tri_defined=TRUE;
  return;
  
tcds_not_defined_exit:

  b_tri_defined=FALSE;
  return;
}

//////////////////////////////////////////////////////////////////////
real IntersectionByMcmc::Likelihood(
  const MCMC_State_Detailed&	X
)
{
  Array<rPoint3*> a00(X.size);
  FORi(X.size) a00[i]=&lattice[i][X.t[i]][X.k[i]];
//  rPoint3 ath, aka;
  Array<rPoint3> a(X.size);
  FORi(X.size) {
    /*
    const rPoint3& aa=*a00[i];
    if (X.t[i]+1==Theta.cols) ath=aa;
    else ath=lattice[i][X.t[i]+1][X.k[i]];
    if (X.k[i]+1==Kappa.cols) aka=aa;
    else aka=lattice[i][X.t[i]][X.k[i]+1];
    a[i]=aa + (ath-aa)*X.tt[i] + (aka-aa)*X.kk[i];
    */
    a[i]=PositionGivenByDetailedState(X,lattice,i);
  }
  real dist=0, dx,dy,dz;
  FORi(X.size) for (int j=i+1; j<X.size; ++j) {
    const rPoint3& ai=a[i];
    const rPoint3& aj=a[j];
    dx=ai.x-aj.x;
    dy=ai.y-aj.y;
    dz=ai.z-aj.z;
    dist+=dx*dx+dy*dy+dz*dz;
  }
//  return 1.0/(dist*dist);
  return 1.0/dist;		// more exploration this way
//  return 1.0/(dist*dist*dist);
}

//////////////////////////////////////////////////////////////////////
void IntersectionByMcmc::OutputManifoldsAndSearchResults(
  const String&			results_file,
  const Array<rPoint3>&		probes
)
{
  if (results_file=="") return;

  const RGB3Byte color[6]={ RGB3Byte_red, RGB3Byte_green, RGB3Byte_cyan,
    RGB3Byte_white, RGB3Byte_blue, RGB3Byte_yellow };
  VrmlFile::Open(results_file);
  Array<rPoint3> M;
  FORk(lattice.size) {
    M.SetSize(Theta.cols*Kappa.cols);
    int inx=0;
    FORi(Theta.cols) FORj(Kappa.cols) M[inx++]=lattice[k][i][j];
    VrmlFile::OutputPoints(M,color[k%6],4);

    //::: show manifold grid lines
    M.SetSize(2*(2*(Theta.cols-1)*(Kappa.cols-1)+Theta.cols-1+Kappa.cols-1));
    inx=0;
    FORi(Theta.cols-1) FORj(Kappa.cols-1) {
      M[inx++]=lattice[k][i][j];
      M[inx++]=lattice[k][i+1][j];
      M[inx++]=lattice[k][i][j];
      M[inx++]=lattice[k][i][j+1];
    }
    FORi(Theta.cols-1) {
      M[inx++]=lattice[k][i][Kappa.cols-1];
      M[inx++]=lattice[k][i+1][Kappa.cols-1];
    }
    FORi(Kappa.cols-1) {
      M[inx++]=lattice[k][Theta.cols-1][i];
      M[inx++]=lattice[k][Theta.cols-1][i+1];
    }
    VrmlFile::OutputLines(M,color[k%6]);
  }
  M.SetSize(1);
  FORi(probes.size) {
    M[0]=probes[i];
    VrmlFile::OutputPoints(M,color[i%6],15);
  }

  //::: show true intersection point
  M.SetSize(1);
  M[0]=true_alpha;
  VrmlFile::OutputPoints(M,RGB3Byte_yellow,15);

  VrmlFile::Close();
}

#ifdef NOTDEFINED
// input:  for each manifold, give range for grid lattice
// output: new ranges, centered around closest mutual intersection point
//////////////////////////////////////////////////////////////////////
void IntersectionByMcmc::McmcForRange(
  Array<real>&			min_th,		// for both input and output; output is centered around final lattice point for each manifold
  Array<real>&			max_th,
  Array<real>&			min_ka,
  Array<real>&			max_ka,
  const int			n_th,
  const int			n_ka,
  const String&			results_file,	// empty string means no output
  const real			f_shrink
)
{
  static int n_visits=0;
  ++n_visits;

//  SetLattice(min_th,max_th,min_ka,max_ka,n_th,n_ka,CON_n_manifolds);
  SetLattice(min_th,max_th,min_ka,max_ka,n_th,n_ka,4);

//  if (n_visits<=2) {
    VOTE_UNNORM(true_alpha);
    NormalizeVotingSpace();
    VOTE_NORM(true_alpha);
//  }

  //::: normalize lattice points
  FORi(lattice.size) FORj(Theta.cols) FORk(Kappa.cols)
    VOTE_NORM(lattice[i][j][k]);
  
  //::: use MCMC to find mutual intersection point of grid/manifolds
  //::: (1) choose initial state X at random; a state is three <theta,kappa>
  //:::     tuples, lattice points on precalc'ed grid
  //::: (2) find likelihood of X
  //::: (3) LOOP
  //:::     (3.1) choose a new state X* from neighbors of X
  //:::     (3.2) find likelihood of X*
  //:::     (3.3) choose random number r in [0,1]
  //:::     (3.4) if r<L(X*)/L(X), set X equal to X* (accept new state)
  //:::
  //::: perform loop a set number of times; keep track of "best" state reached during looping
  //::: MCMC finds expected values, but its properties can be useful in finding global mins

  //::: (1) choose initial state X at random; a state is three <theta,kappa>
  //:::     tuples, lattice points on precalc'ed grid
  MCMC_State X(lattice.size), XX, best_X;
  FORi(X.size) {
    X.t[i]=random(1,Theta.cols)-1;
    X.k[i]=random(1,Kappa.cols)-1;
  }
  //::: (2) find likelihood of X
  real X_like=Likelihood(X);
  //::: (3) LOOP
  real best_like=-1;
  for (long kk=0; kk<1000000; ++kk) {
    //::: keep track of "best" state
    if (X_like>best_like) {
      best_like=X_like; best_X=X;
      printf("%e",best_like);
      FORi(X.size) printf("  <%3d,%3d>",X.t[i],X.k[i]);
      printf("\n");
//      P[0][X.t[0]][X.k[0]].Dump();
//      P[1][X.t[1]][X.k[1]].Dump();
//      P[2][X.t[2]][X.k[2]].Dump();
    }
    //:::     (3.1) choose a new state X* from neighbors of X
    int Theta_change, Kappa_change;
    int mani_change=random(1,X.size)-1;
    do {
      Theta_change=random(-1,1);
      Kappa_change=random(-1,1);
    } while ((Theta_change==0)&&(Kappa_change==0));
    XX=X;
    XX.t[mani_change]+=Theta_change+Theta.cols;
    XX.k[mani_change]+=Kappa_change+Kappa.cols;
    XX.t[mani_change]%=Theta.cols;
    XX.k[mani_change]%=Kappa.cols;
    //:::     (3.2) find likelihood of X*
    real XX_like=Likelihood(XX);
    //:::     (3.3) choose random number r in [0,1]
    real r=rand01();
    //:::     (3.4) if r<L(X*)/L(X), set X equal to X* (accept new state)
    if (r<XX_like/X_like) { X=XX; X_like=XX_like; }
//    printf("%e   %e   %e   %e\n",r,XX_like/X_like,XX_like,X_like);
  }

  //::: show results
  if (results_file!="") {  
    Array<rPoint3> probes(best_X.size);
    FORi(best_X.size) probes[i]=lattice[i][best_X.t[i]][best_X.k[i]];
    OutputManifoldsAndSearchResults(results_file,probes);
  }
  
  //::: shrink range to produce output
  FORi(min_th.size) {
    real breadth=max_th[i]-min_th[i];
    breadth*=f_shrink;
    min_th[i]=Theta[i][best_X.t[i]]-breadth/2;
    max_th[i]=Theta[i][best_X.t[i]]+breadth/2;
  }
  FORi(min_ka.size) {
    real breadth=max_ka[i]-min_ka[i];
    breadth*=f_shrink;
    min_ka[i]=Kappa[i][best_X.k[i]]-breadth/2;
    max_ka[i]=Kappa[i][best_X.k[i]]+breadth/2;
  }
}

// straight-forward grid; use zoom-ins
//////////////////////////////////////////////////////////////////////
void IntersectionByMcmc::Run(
  const Array2D<boole>&		bPairOn
)
{
  const int	CON_n_theta	= 30;
  const int	CON_n_kappa	= 30;
  const real	CON_min_theta	= 0.01;
  const real	CON_max_theta	= PI-0.01;
  const real	CON_min_kappa	= 0.05;
  const real	CON_max_kappa	= 0.95;

  BANNER("starting MCMC search\a");  fflush(stdout);
  true_alpha.Dump("true alpha");

//  Array<real> min_th(CON_n_manifolds), max_th(CON_n_manifolds);
//  Array<real> min_ka(CON_n_manifolds), max_ka(CON_n_manifolds);
  Array<real> min_th(4), max_th(4);
  Array<real> min_ka(4), max_ka(4);
  FORi(min_th.size) min_th[i]=CON_min_theta;
  FORi(max_th.size) max_th[i]=CON_max_theta;
  FORi(min_ka.size) min_ka[i]=CON_min_kappa;
  FORi(max_ka.size) max_ka[i]=CON_max_kappa;
  
//  McmcForRange(min_th,max_th,min_ka,max_ka,CON_n_theta,CON_n_kappa,"",0.5);
//  McmcForRange(min_th,max_th,min_ka,max_ka,CON_n_theta,CON_n_kappa,"",0.5);
//  McmcForRange(min_th,max_th,min_ka,max_ka,CON_n_theta,CON_n_kappa,"",0.5);
//  McmcForRange(min_th,max_th,min_ka,max_ka,CON_n_theta,CON_n_kappa,"",0.5);
  McmcForRange(min_th,max_th,min_ka,max_ka,CON_n_theta,CON_n_kappa,"MCMC.wrl",0.5);

  EXIT();
}
#endif

//////////////////////////////////////////////////////////////////////
void IntersectionByMcmc::Run(
  const Array2D<boole>&		bPairOn
)
{
  const int	CON_n_theta	= 50;
  const int	CON_n_kappa	= 50;
  const real	CON_min_theta	= 0.01;
  const real	CON_max_theta	= PI-0.01;
  const real	CON_min_kappa	= 0.05;
  const real	CON_max_kappa	= 0.95;

  BANNER("starting MCMC search\a");
  true_alpha.Dump("true alpha");

  Array<real> min_th(CON_n_manifolds), max_th(CON_n_manifolds);
  Array<real> min_ka(CON_n_manifolds), max_ka(CON_n_manifolds);
  FORi(min_th.size) min_th[i]=CON_min_theta;
  FORi(max_th.size) max_th[i]=CON_max_theta;
  FORi(min_ka.size) min_ka[i]=CON_min_kappa;
  FORi(max_ka.size) max_ka[i]=CON_max_kappa;
  
  SetLattice(Theta,Kappa,lattice,b_lattice,min_th,max_th,min_ka,max_ka,
             CON_n_theta,CON_n_kappa,CON_n_manifolds,pairinfo,T0i_proj,E);
  NormalizeVotingSpace(Theta,Kappa,lattice,b_lattice);

  VOTE_NORM(true_alpha);
  //::: normalize lattice points
  FORi(lattice.size) FORj(Theta.cols) FORk(Kappa.cols)
    VOTE_NORM(lattice[i][j][k]);

/*
//  SetLattice(min_th,max_th,min_ka,max_ka,n_th,n_ka,CON_n_manifolds);
  SetLattice(min_th,max_th,min_ka,max_ka,CON_n_theta,CON_n_kappa,CHUCKY);

  NormalizeVotingSpace();
  VOTE_NORM(true_alpha);

  //::: normalize lattice points
  FORi(lattice.size) FORj(Theta.cols) FORk(Kappa.cols)
    VOTE_NORM(lattice[i][j][k]);
*/


  
  //::: use MCMC to find mutual intersection point of grid/manifolds
  //::: (1) choose initial state X at random; a state is three <theta,kappa>
  //:::     tuples, lattice points on precalc'ed grid
  //::: (2) find likelihood of X
  //::: (3) LOOP
  //:::     (3.1) choose a new state X* from neighbors of X
  //:::     (3.2) find likelihood of X*
  //:::     (3.3) choose random number r in [0,1]
  //:::     (3.4) if r<L(X*)/L(X), set X equal to X* (accept new state)
  //:::
  //::: perform loop a set number of times; keep track of "best" state reached during looping
  //::: MCMC finds expected values, but its properties can be useful in finding global mins

  //::: (1) choose initial state X at random; a state is three <theta,kappa>
  //:::     tuples, lattice points on precalc'ed grid
  MCMC_State_Detailed X(lattice.size), XX, best_X;
  FORi(X.size) {
    X.t[i]=random(1,Theta.cols)-1;
    X.k[i]=random(1,Kappa.cols)-1;
//    X.tt[i]=rand01();
//    X.kk[i]=rand01();
    X.itt[i]=random(0,99);
    X.ikk[i]=random(0,99);
    X.tt[i]=real(X.itt[i])/100;
    X.kk[i]=real(X.ikk[i])/100;
  }
  //::: (2) find likelihood of X
  real X_like=Likelihood(X);
  //::: (3) LOOP
  real best_like=-1;
  for (long kk=0; kk<500000; ++kk) {
    if (kk%100000==0) printf("COUNT=%ld\n",kk);
    //::: keep track of "best" state
    if (X_like>best_like) {
      best_like=X_like; best_X=X;
      printf("%e",best_like);
      FORi(X.size) printf("  <%3d,%3d>",X.t[i],X.k[i]);
      printf("\n");
//      P[0][X.t[0]][X.k[0]].Dump();
//      P[1][X.t[1]][X.k[1]].Dump();
//      P[2][X.t[2]][X.k[2]].Dump();
    }
    //:::     (3.1) choose a new state X* from neighbors of X
/*
    int Theta_change, Kappa_change;
    int mani_change=random(1,X.size)-1;
    do {
      Theta_change=random(-1,1);
      Kappa_change=random(-1,1);
    } while ((Theta_change==0)&&(Kappa_change==0));
    XX=X;
    XX.t[mani_change]+=Theta_change+Theta.cols;
    XX.k[mani_change]+=Kappa_change+Kappa.cols;
    XX.t[mani_change]%=Theta.cols;
    XX.k[mani_change]%=Kappa.cols;
    XX.tt[mani_change]=rand01();
    XX.kk[mani_change]=rand01();
*/
    //::: discrete sub-lattice transition
    XX=X;
    int Theta_change, Kappa_change;
    int mani_change=random(1,X.size)-1;
    do {
      Theta_change=random(-1,1);
      Kappa_change=random(-1,1);
    } while ((Theta_change==0)&&(Kappa_change==0));
    if (rand01()<0.50) {
      //::: note tt and kk stay same but in adjacent square
      XX.t[mani_change]+=Theta_change+Theta.cols;
      XX.k[mani_change]+=Kappa_change+Kappa.cols;
      XX.t[mani_change]%=Theta.cols;
      XX.k[mani_change]%=Kappa.cols;
    }
    else {
      /*
      //::: stay in same square; method could lead to round-off errors
      XX.tt[mani_change]+=real(Theta_change)/100;
      XX.kk[mani_change]+=real(Kappa_change)/100;
      if (XX.tt[mani_change]<0) XX.tt[mani_change]+=1.0;
      else if (XX.tt[mani_change]>=1.0) XX.tt[mani_change]-=1.0;
      if (XX.kk[mani_change]<0) XX.kk[mani_change]+=1.0;
      else if (XX.kk[mani_change]>=1.0) XX.kk[mani_change]-=1.0;
      */
      //::: stay in same square
      XX.itt[mani_change]+=Theta_change+100;
      XX.ikk[mani_change]+=Kappa_change+100;
      XX.itt[mani_change]%=100;
      XX.ikk[mani_change]%=100;
      XX.tt[mani_change]=real(XX.itt[mani_change])/100;
      XX.kk[mani_change]=real(XX.ikk[mani_change])/100;
    }

    //:::     (3.2) find likelihood of X*
    real XX_like=Likelihood(XX);
    //:::     (3.3) choose random number r in [0,1]
    real r=rand01();
    //:::     (3.4) if r<L(X*)/L(X), set X equal to X* (accept new state)
    if (r<XX_like/X_like) { X=XX; X_like=XX_like; }	// do this with ptr swapping...
  }

/*
  Array<rPoint3> probes(best_X.size);
  FORi(best_X.size) probes[i]=PositionGivenByDetailedState(best_X,lattice,i);
  OutputManifoldsAndSearchResults("MCMC.wrl",probes);
  EXIT();
*/


  Array<rPoint3> probes(best_X.size);
  FORi(best_X.size) probes[i]=PositionGivenByDetailedState(best_X,lattice,i);
  Array<ClusterTriangle> TRI(best_X.size);
  FORi(best_X.size) {
    boole b_defined;
    TriangleContainingDetailedState(TRI[i],b_defined,best_X,lattice,i);
    if (!b_defined) EXIT("undefined triangle");
  }
  boole b_intersect;
  rPoint3 P;
  IntersectionOf3Triangles(b_intersect,P,TRI[0],TRI[1],TRI[2]);
Array<rPoint3> Q(6);
VrmlFile::Open("JUNK.wrl");
RGB3Byte ccolor[]={RGB3Byte_red,RGB3Byte_green,RGB3Byte_yellow};
FORi(3) {
  int inx=0;
  Q[inx++]=TRI[i].a;
  Q[inx++]=TRI[i].b;
  Q[inx++]=TRI[i].b;
  Q[inx++]=TRI[i].c;
  Q[inx++]=TRI[i].c;
  Q[inx++]=TRI[i].a;
  VrmlFile::OutputLines(Q,ccolor[i]);
}
Q.SetSize(1);
FORi(3) { Q[0]=probes[i]; VrmlFile::OutputPoints(Q,ccolor[i],6); }
Q[0]=true_alpha;
VrmlFile::OutputPoints(Q,RGB3Byte_cyan,6);
Q[0]=P;
VrmlFile::OutputPoints(Q,RGB3Byte_white,6);
VrmlFile::Close();

  if (!b_intersect) printf("triangles do not intersect\n");
  P.Dump("mutual intersection of planes");
  probes.Dump();
  true_alpha.Dump("true alpha");

/*
  VrmlFile::Open("output.mcmc.results.wrl");
  VrmlFile::OutputPoints(probes,RGB3Byte_yellow,5);
//  probes.SetSize(1);
//  probes[1]=true_alpha;
//  VrmlFile::OutputPoints(probes,RGB3Byte_green,5);
  VrmlFile::Close();
*/

  EXIT();
}
