#ifndef INCLUDED_IntersectionByMcmc
#define INCLUDED_IntersectionByMcmc

///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////

#include "stdmacros.h"
#include "Array.h"
#include "Array2D.h"
#include "LinkedList.h"
#include "ManifoldPointStruct.3Space.h"

///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////
typedef Array2D<rPoint3> Array2D_rPoint3;
typedef Array2D<boole> Array2D_boole;
typedef LinkedList<AlphaCoord> LinkedList_AlphaCoord;

///////////////////////////////////////////////////////////////////////
struct MCMC_State {
  int				size;
  Array<int>			t,k;
  
  MCMC_State(const int _size=0) { size=_size; t.SetSize(size); k.SetSize(size); }
};

///////////////////////////////////////////////////////////////////////
struct MCMC_State_Detailed {
  int				size;
  Array<int>			t,k;
  Array<real>			tt,kk;		// meniscus
  Array<int>			itt,ikk;	// sub lattice coords
  
  MCMC_State_Detailed(const int _size=0)
    { size=_size;
      t.SetSize(size); k.SetSize(size);
      tt.SetSize(size); kk.SetSize(size);
      itt.SetSize(size); ikk.SetSize(size);
    }
};

///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////

// structure so a group of tightly-related functions can share common variables easily
//////////////////////////////////////////////////////////////////////
struct IntersectionByMcmc {

  //::: info that must be passed in to call the algorithm
  const Array<ViewPairInfo>&	pairinfo;		//holds epipolar geometry info for each pair in use
  const Array<Transform>&	T0i_proj;		//this combined with E form all camera matrices in common projective basis
  const Array<rPoint3>&		E;
  const AlphaCoord&		__true_alpha;
  CONST AlphaCoord		true_alpha;
  const Transform&		normalizer;
  Array2D<rPoint3>*		GLOB_Kpts;		//simple kludge used in outputting affine reconstructions
  Transform* const		pInternal;		//allows true internal params to be passed in for testing
  rPoint2			image_size;		//image size is used for showing camera fulcrums and for sanity checking calibration

  //::: local constants
  const int			CON_n_manifolds;	//just pairinfo.size, renamed to make code clearer
  const int			CON_vote_dim;		//number of vote voxels in region under consideration
  const int			CON_bacteria_dim;	//number of bacteria voxels in region under consideration
  const int			CON_sq_vote_dim;
  const int			CON_cube_vote_dim;
  const int			CON_sq_bacteria_dim;
  const int			CON_cube_bacteria_dim;
  const int			CON_likely_zoomin;

  //::: local variables
  Array2D<ManifoldPointStruct3Space>	p;		//attractors on manifold surfaces; [n_manifolds, n_attractors]
  word*				nVotes;
  VolumeStruct			region, maximum_region;
  Array<DitherRangeStruct>	dither_range;
  Array<boole>			bManifoldOn;
  boole				bRangeDitherOn;
  boole				bHelpLosersOn;
  
  //::: local variables for manifold coordinate grids
  real				theta_min, theta_max;	// range for theta
  real				kappa_min, kappa_max;	// range for kappa
  int				n_theta;		// steps for theta
  int				n_kappa;		// steps for kappa
  Array2D<real>			Theta;			// theta values for grid lattice for each manifold
  Array2D<real>			Kappa;			// kappa values for grid lattice for each manifold
  Array<Array2D_rPoint3>	lattice;		// lattice points of the grid imposed on each manifold
  Array<Array2D_boole>		b_lattice;		// whether a given lattice point has meaning (or is out of bounds)

  //::: methods / local functions
//  void SeedInitialRegion();
  void Run(const Array2D<boole>&);
//  boole MoveAlongManifoldTowardsDesiredPosition2(ManifoldPointStruct3Space&,
//    const VoxelCoord3Space&, const AlphaCoord&, const int);
//  void RunStateReport();
//  void FindCalibrationFromTabs();
//  void ReconstructScene(const AlphaCoord& a);
  real Likelihood(const MCMC_State& X);
  real Likelihood(const MCMC_State_Detailed& X);
  void McmcForRange(Array<real>&, Array<real>&, Array<real>&, Array<real>&, const int, const int, const String& outfile="", const real shrink=0.5);
  void OutputManifoldsAndSearchResults(const String&, const Array<rPoint3>&);

  IntersectionByMcmc(const Array<ViewPairInfo>&, const Array<Transform>&,
    const Array<rPoint3>&, const rPoint3&, const Transform&, const rPoint2&,
    Array2D<rPoint3>*, Transform* const _pInternal=NULL);
  ~IntersectionByMcmc();
};

///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////

#endif
