#include <stdio.h>

#include "helpers.h"
#include "String.h"
#include "Matrix.h"
#include "VrmlFile.h"
#include "MiscUtility.h"
#include "ManifoldPointStruct.3Space.h"
#include "IntersectionByTriangles.h"
#include "PlotManifolds.h"
#include "RunningTimeDisplay.h"
#include "TestRunDataStruct.h"
#include "MeshGenerationRoutines.h"

#include "Image3ByteChannels.h"
#include "KltFeatureTable.h"

//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

// taken from another program; put in standard library!
//////////////////////////////////////////////////////////////////////
void RandomIndexSet(
  Array<int>&			set,
  const int			max
)
{
  int top=0;
  FORi(set.size) {
    int x;
    INFINITELOOP {
      x=random(0,max-1);
      if (top==0) break;
      boole bFound=FALSE;
      FORj(top) if (set[j]==x) { bFound=TRUE; break; }
      if (!bFound) break;
    }
    set[top++]=x;
  }
}

//////////////////////////////////////////////////////////////////////
void RandomIndexSet(
  Array<int>&			set,
  const int			n,
  const int			max
)
{
  set.SetSize(n);
  RandomIndexSet(set,max);
}

//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
// put in standard library

#ifdef NOTDEFINED
// working permutation creator
//////////////////////////////////////////////////////////////////////
void CreatePermutations(
  Array2D<uint8>&		P,
  const int			n
)
{
  P.FreeMem();
  if (n<=0) return;
  if (n>8) EXIT("CreatePermutations() probably can't handle n");
  uint32 nfact=1;
  FORiIN(2,n) nfact*=i;
  P.SetSize(nfact,n);
  
  //::: fill P with all permutations in special form
  FORi(n) {
    int r=0;
    int m=n-1-i;
    int mfact=1;
    FORjIN(1,m) mfact*=j;
    int inx=0;
    while (r<nfact) {
      FORj(mfact) P[r++][i]=inx;
      ++inx;
      if (inx==m+1) inx=0;
    }
  }

  //::: convert special form to standard form
  Array<uint8> avail(n);
  const int nminus1=n-1;
  FORi(nfact) {
    FORj(n) avail[j]=j;
    FORj(n) {
      int t=P[i][j];
      P[i][j]=avail[t];
      FORkIN(t+1,nminus1) avail[k-1]=avail[k];
    }
  }

  FORi(nfact) {
    FORj(n) printf("%d  ",P[i][j]);
    printf("\n");
  }
  EXIT();
}
#endif

// return array of all combinations "n choose q"
//////////////////////////////////////////////////////////////////////
void CreateCombinations(
  Array2D<uint8>&		P,
  const int			n,
  const int			q
)
{
  P.FreeMem();
  if (n<1) return;
  if (q<1) return;
  ASSERT(q<=n,"bad input to CreateCombinations()");
  if (n>8) EXIT("CreateCombinations() probably can't handle n");
  uint32 nchooseq=1;
  FORiIN(q+1,n) nchooseq*=i;
  FORiIN(2,n-q) nchooseq/=i;
  P.SetSize(nchooseq,q);

  if (q==1) {
    FORi(P.rows) P[i][0]=i;
  }
  else {
    int qm1=q-1;
    FORi(q) P[0][i]=i;
    
    for (int r=1; r<P.rows; ++r) {
      FORi(q) P[r][i]=P[r-1][i];
      ++P[r][qm1];
      if (P[r][qm1]==n) {
        int j=q-2;
        while (P[r][j]==n-1-(qm1-j)) --j;
        ++P[r][j];
        FORkIN(j+1,qm1) P[r][k]=P[r][k-1]+1;
      }
    }
  }

/*
  FORi(P.rows) {
    FORj(P.cols) printf("%d  ",P[i][j]);
    printf("\n");
  }
  printf("\n");
*/
}

//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
//const RGB3Byte color[6]={ RGB3Byte_red, RGB3Byte_green, RGB3Byte_cyan,
//  RGB3Byte_white, RGB3Byte_blue, RGB3Byte_yellow };

//////////////////////////////////////////////////////////////////////
IntersectionByTriangles::IntersectionByTriangles(
  const Array<ViewPairInfo>&	_pairinfo,
  const Array<Transform>&	_T0i_proj,
  const Array<rPoint3>&		_E,
  const AlphaCoord&		_true_alpha,
  const Transform&		_Tinorm,
  const rPoint2&		_image_size,
  Array2D<rPoint3>*		_GLOB_Kpts,
  Transform* const		_pInternal
) :
  pairinfo(_pairinfo), T0i_proj(_T0i_proj), E(_E), __true_alpha(_true_alpha),
  normalizer(_Tinorm), image_size(_image_size), GLOB_Kpts(_GLOB_Kpts),
  pInternal(_pInternal),

  CON_n_manifolds(pairinfo.size)
{
  true_alpha=__true_alpha;
}

//////////////////////////////////////////////////////////////////////
IntersectionByTriangles::~IntersectionByTriangles()
{
}

#include "IntByTriRelated.cc"
#include "IntByMcmcRelated.cc"
//#include "IntByModcRelated.cc"

//////////////////////////////////////////////////////////////////////
void IntersectionByTriangles::RansacIteration(
  rPoint3&			best_ans,		// best intersection point
  Transform&			best_K,			// internal calibration
  int&				best_n_good,		// number of manifolds that fit the answer well
  real&				best_err,		// error induced by K
  const int			method,
  const Array<int>&		draws,
  const Array<LinkedList_ClusterLatticeCluster>&	ll_clusters
)
{
  RunningTimeDisplay::SwissTimePush();
  const boole CON_b_show_stats = FALSE;

  LinkedList<rPoint3> ll_mutuals;
  switch (method) {
    case METHOD_TRI:
      PerformIntersectionByTriangles(ll_mutuals,draws,ll_clusters,CON_b_show_stats);
      break;
      
    case METHOD_MCMC:
      PerformIntersectionByMcmc(ll_mutuals,draws,CON_b_show_stats);
      FORLISTnow(ll_mutuals, rPoint3)
        rPoint3 ans=now;
        VOTE_UNNORM(ans);
      }
      break;

    case METHOD_MODC:
      PerformIntersectionByModc(ll_mutuals,draws,CON_b_show_stats);
      break;

    default:
      EXIT("unknown method");
  }

  //::: convert each intersection point to K matrix; calc goodness of solution
  best_n_good=-1;
//  const real CON_good_limit=1.0e-1;
  const real CON_good_limit=0.6e-1;	// worked for many cases, although just barely
//  const real CON_good_limit=1.5e-1;

  //::: init some variables used in testing
  real best_frob_with_true_K=-1;
  Transform trueK;
  if (pInternal!=NULL) {
    trueK=(normalizer.Inverse()*(*pInternal));
    trueK=trueK/trueK.Norm();
  }
  
  //::: check each mutual intersection point
  SortedFixedSizeLinkedList<PerIntersectionErrData> intersection_err(2);
  SCREENPUT printf("mutual intersections = %d\n",ll_mutuals.size);
  FORl(ll_mutuals.size) {
//    rPoint3 ans=true_alpha;
    rPoint3 ans=ll_mutuals[l];
    VOTE_UNNORM(ans);
    //::: use ans to find relative calibrations for each view pair
    Array<Transform> allTabs(pairinfo.size);
    FORi(pairinfo.size) {
      int ii=pairinfo[i].i;
      int jj=pairinfo[i].j;
      Transform TT;
      if (ii==0)
        TT=T0i_proj[jj]+Transform::Outer(E[jj],ans);
      else {
        Transform TTi=T0i_proj[ii]+Transform::Outer(E[ii],ans);
        Transform TTj=T0i_proj[jj]+Transform::Outer(E[jj],ans);
        TT=TTj*(TTi.Inverse());
      }
      allTabs[i]=(normalizer.Inverse()*TT*normalizer);
      allTabs[i].Normalize();
    }

    //::: find K; only use the three manifolds involved in the intersection
    const int ONE=draws[0], TWO=draws[1], THREE=draws[2];
    Array<Transform> threeTabs(3);
    threeTabs[0]=allTabs[ONE];
    threeTabs[1]=allTabs[TWO];
    threeTabs[2]=allTabs[THREE];
    Transform absconic;
    MiscUtility::FindAbsoluteConicFromRelativeCalibrations(absconic,threeTabs);
    rMatrix rmKK(&absconic[0][0],3,3);
    rmKK/=(rmKK[0][0]);
    rmKK=rmKK.InverseOf();
    int bGoodCholesky=rmKK.Cholesky();
//    if (!bGoodCholesky) { ALERT("Cholesky decomposition failed\n"); continue; }
    if (!bGoodCholesky) continue;
    rmKK=rmKK.InverseOf();
    Transform K(rmKK.data);
    Transform invK=K.Inverse();

    PerIntersectionErrData* inter_err;
    inter_err=new PerIntersectionErrData;
    inter_err->err_frob=-1;
    //::: if true K is known, check how well K matches true K; keep records
    if (pInternal!=NULL) {
      real frob=((K/K.Norm())-trueK).Norm();
//      SCREENPUT printf("frob = %lf\n",frob);
      if ((best_frob_with_true_K<0)||(best_frob_with_true_K>frob)) best_frob_with_true_K=frob;
      inter_err->err_frob=frob;
    }

    //::: test how well K fits with all fmats/Hinfs
    real tot_err=0;
    int n_good=0;
    FORi(allTabs.size) {
      real det=allTabs[i].Determinant();
      real scale=1.0/pow(fabs(det),1.0/3)*sign(det);
      Transform R=(invK*allTabs[i]*K)*scale;
      Transform RR=R*(R.TransposeOf());
//      RR.Dump();
      real err=(Transform_Id-RR).Norm();		// frobenius norm
      SCREENPUT if (CON_b_show_stats) { printf("err = %e          %d  %d\n",err,pairinfo[i].i,pairinfo[i].j); }
//      if (draws.bHolds(i)) {		// use this method to maximize fit to just the three manifolds used in iteration
      if (err<CON_good_limit) {
        ++n_good;
        tot_err+=err*err;
      }
    }
    SCREENPUT if (CON_b_show_stats) { printf("tot err = %e\n",tot_err); }

/*
    //::: following test eliminates solutions that are not realistic cameras
    if ( (!within(K[0][0]/K[1][1],0.95,1.05)) ||
         (!within(K[0][2]/1024/K[1][2]*768,0.90,1.1)) )
          n_good=min(n_good,1);		// kludgy way to indicate bad camera
*/

    //::: test if ans did not fit any Hinf well
    if (n_good==0) {
      tot_err=888.0;
    }
    else
    //::: eliminate solutions that are not realistic cameras
    if ( (!within(K[0][0]/K[1][1],0.85,1.15)) ||
//         (!within((K[0][2]/K[1][2])*(image_size.y/image_size.x),0.50,2.0)) ) {
         (!within((K[0][2]/image_size.x),0.35,0.65)) ||
         (!within((K[1][2]/image_size.y),0.35,0.65)) ) {
          n_good=min(n_good,1);		// kludgy way to indicate bad camera
          tot_err=99.0+sqrt(tot_err);
    }
    else
      tot_err=sqrt(tot_err);

    inter_err->err_ransac=tot_err;
    intersection_err.TryToAdd(inter_err);

    //::: keep track of best answer so far
    if ((best_n_good==-1) || (n_good>best_n_good) || ((n_good==best_n_good)&&(tot_err<best_err))) {
      best_n_good=n_good;
      best_err=tot_err;
      best_K=K;
      best_ans=ans;
    }
  }	// end loop

  //::: for testing, output "per trial" info
  GLOBAL_testrun_data.iter_n_intersection_pts=ll_mutuals.size;
  GLOBAL_testrun_data.iter_best_frob_error=best_frob_with_true_K;
  GLOBAL_testrun_data.iter_time_iteration=RunningTimeDisplay::SwissTimePop();
  GLOBAL_testrun_data.iter_1st_ransac_error=(intersection_err.size>=1) ? intersection_err[0].err_ransac : -1;
  GLOBAL_testrun_data.iter_2nd_ransac_error=(intersection_err.size>=2) ? intersection_err[1].err_ransac : -1;
  GLOBAL_testrun_data.iter_1st_frob_error=(intersection_err.size>=1) ? intersection_err[0].err_frob : -1;
  GLOBAL_testrun_data.iter_2nd_frob_error=(intersection_err.size>=2) ? intersection_err[1].err_frob : -1;
  GLOBAL_testrun_data.Write_TriangleIntersect_Iteration_TR1428(GLOBAL_testrun_data.file_per_iteration_data);

  //::: free mem
  ll_mutuals.FreeMem();
  intersection_err.FreeMem();
}

//////////////////////////////////////////////////////////////////////
void IntersectionByTriangles::Run(
  const String&			s_method,
  const Array2D<boole>&		bPairOn
)
{
  const int	CON_n_kappa	= 300;
  const int	CON_n_theta	= 100;
  const int	CON_ka_step	= 2;
  const int	CON_th_step	= 2;
  const real	CON_min_theta	= 0.01;
  const real	CON_max_theta	= PI-0.01;
  const real	CON_min_kappa	= 0.05;
  const real	CON_max_kappa	= 0.95;

  int method;
  if (s_method=="mcmc") method=METHOD_MCMC;
  else
  if (s_method=="tri") method=METHOD_TRI;
  else
  if (s_method=="modc") method=METHOD_MODC;
  else
  throw IntersectionByTrianglesError(String("unknown method:  ")+s_method,-1);

  SCREENPUT {
    switch (method) {
      case METHOD_MCMC:
        BANNER("starting mcmc search\a");
        break;
        
      case METHOD_TRI:
        BANNER("starting exhaustive triangle intersection search\a");
        break;

      case METHOD_MODC:
        BANNER("starting modulus constraint search\a");
        break;
        
      default:  EXIT("unknown method");
    }
    fflush(stdout);
  }
  SCREENPUT true_alpha.Dump("true alpha");

  //::: create meshes
  Array<real> min_th(CON_n_manifolds), max_th(CON_n_manifolds);
  Array<real> min_ka(CON_n_manifolds), max_ka(CON_n_manifolds);
  FORi(min_th.size) min_th[i]=CON_min_theta;
  FORi(max_th.size) max_th[i]=CON_max_theta;
  FORi(min_ka.size) min_ka[i]=CON_min_kappa;
  FORi(max_ka.size) max_ka[i]=CON_max_kappa;
  
  SetLattice(Theta,Kappa,lattice,b_lattice,min_th,max_th,min_ka,max_ka,
             CON_n_theta,CON_n_kappa,CON_n_manifolds,pairinfo,T0i_proj,E);
  NormalizeVotingSpace(Theta,Kappa,lattice,b_lattice);
  VOTE_NORM(true_alpha);

  //::: normalize mesh points
  FORi(lattice.size) FORj(Theta.cols) FORk(Kappa.cols)
    VOTE_NORM(lattice[i][j][k]);

  SCREENPUT printf("\a");  fflush(stdout);
  Array<LinkedList_ClusterLatticeCluster> ll_clusters(CON_n_manifolds);
  if (method==METHOD_TRI) CreateTriangleClusters(ll_clusters,lattice,b_lattice,CON_n_kappa,CON_n_theta,CON_ka_step,CON_th_step);

  //::: use ransac to find K
  int CON_n_ransac_trials=100;
  
  //::: (*) if ransac_trials is larger than (n_manifolds!) then create list of all possible draws (combinations)
  boole b_all_combinations=FALSE;
  Array2D<uint8> combinations;
  if (CON_n_manifolds<=6) {
    CreateCombinations(combinations,CON_n_manifolds,3);
    if (combinations.rows<=CON_n_ransac_trials)	{ b_all_combinations=TRUE; CON_n_ransac_trials=combinations.rows; }
    else					{ b_all_combinations=FALSE; combinations.FreeMem(); }
  }

  if (method==METHOD_MCMC) {
    b_all_combinations=FALSE;
    CON_n_ransac_trials=50;
  }

  Array<int> draws(3);
  real err, best_err;
  Transform K, best_K;
  rPoint3 ans, best_ans;
  int n_good, best_n_good=-1;
  int n_trials_performed=0;

  RunningTimeDisplay::SwissTimePush();
  SCREENPUT RunningTimeDisplay::Stopwatch();
  for (int loop=0; loop<CON_n_ransac_trials; ++loop) {
    //::: pick 3 manifolds to intersect
    if (b_all_combinations)
      FORi(3) draws[i]=combinations[loop][i];		// step through all combinations
    else
      RandomIndexSet(draws,CON_n_manifolds);		// pick three indices at random from [0,CON_n_manifolds-1]

    //::: try to intersect them
    RansacIteration(ans,K,n_good,err,method,draws,ll_clusters);
    ++n_trials_performed;

    //::: check if best answer so far
    real err2=err/((n_good>0) ? n_good : 1);	// err per accepted manifold
    if ((best_n_good==-1) || (n_good>best_n_good) || ((n_good==best_n_good)&&(err2<best_err))) {
      best_n_good=n_good;
      best_err=err2;
      best_K=K;
      best_ans=ans;
      SCREENPUT printf("(%5d)   %e    [%d]  ",loop,best_err,best_n_good);
      SCREENPUT best_ans.Dump();
    }

    if (method==METHOD_MCMC) {
//      if ((n_trials_performed>20) && (best_err<1e-2)) break;
    }
  }
  SCREENPUT RunningTimeDisplay::Stopwatch();
  GLOBAL_testrun_data.time_all_ransac=RunningTimeDisplay::SwissTimePop()/n_trials_performed;   

  SCREENPUT best_ans.Dump("best ans");
  VOTE_UNNORM(true_alpha);
  SCREENPUT true_alpha.Dump("true ans");
  
  SCREENPUT OutputResults(best_ans,best_K);
//Transform trueK=(normalizer.Inverse()*(*pInternal));
//SCREENPUT OutputResults(true_alpha,trueK);

//  if ((best_n_good<3) || (best_err<0))
//    throw IntersectionByTrianglesError("no mutual intersection",-1);

  //::: tabulate results for experiments
  Transform answer=best_K;
  real frob_err, frob_err2;
  Transform per_entry_diff(0,0,0);
  if (pInternal!=NULL) {
    Transform trueK=(normalizer.Inverse()*(*pInternal));
    Transform diff=(answer/answer.Norm())-(trueK/trueK.Norm());
    SCREENPUT (answer/answer[2][2]).Dump("best K");
    SCREENPUT (trueK/trueK[2][2]).Dump("true K");
    per_entry_diff=diff;
    frob_err=diff.Norm();
    frob_err/=6;		// ...since all 6 nonzero entries are being measured
    diff[0][2]=diff[1][2]=0;	// use to just check error in main components
    frob_err2=diff.Norm();
    frob_err2/=4;		// ...since only 4 entries being measured for error
  } else
    frob_err=frob_err2=-1;

  GLOBAL_testrun_data.error_frobenius_main=frob_err;
  GLOBAL_testrun_data.error_frobenius=frob_err2;
  GLOBAL_testrun_data.error_per_entry=per_entry_diff;
  GLOBAL_testrun_data.n_grid_size_ka=CON_n_kappa;
  GLOBAL_testrun_data.n_grid_size_th=CON_n_theta;
  GLOBAL_testrun_data.n_grid_block_size_ka=CON_ka_step;
  GLOBAL_testrun_data.n_grid_block_size_th=CON_th_step;
  GLOBAL_testrun_data.n_ransac_trials=n_trials_performed;

// not throwing bad flag because my test for good and bad not
// reliable; algo finds decent mutual intersection point but may not
// realize it; just want raw, uninterpreted results
//  if ((best_n_good<3) || (best_err<0))
//    throw IntersectionByTrianglesError("no mutual intersection",-1);
//  else
    throw IntersectionByTrianglesError("SUCCESS!",0);
}

//////////////////////////////////////////////////////////////////////
void IntersectionByTriangles::OutputResults(
  const rPoint3&		best_ans,
  const Transform&		best_K
)
{
  Transform T_norm0_met=(normalizer*best_K).Inverse();
  Transform T_met_norm0=T_norm0_met.Inverse();
  Transform inv_norm=normalizer.Inverse();

  const rPoint3 a=best_ans;
  //::: create projective camera matrices
  Array<rMatrix> cams(E.size), unnorm_cams;
  FORi(cams.size) if (i>0) {
    cams[i].SetSize(3,4);
    Transform Tab=T0i_proj[i]+Transform::Outer(E[i],a);
    FORj(3) FORk(3) cams[i][j][k]=Tab[j][k];
    cams[i][0][3]=E[i].x;
    cams[i][1][3]=E[i].y;
    cams[i][2][3]=E[i].z;
  }
  cams[0].SetSize(3,4);
  cams[0].FillWith(0);
  cams[0][0][0]=cams[0][1][1]=cams[0][2][2]=1;

  //::: use normalizing transform on camera matrices
  rMatrix rm44_T_met_norm0(4,4);
  rm44_T_met_norm0.MakeIdentity();
  FORi(3) FORj(3) rm44_T_met_norm0[i][j]=T_met_norm0[i][j];
  FORi(cams.size) cams[i]=cams[i]*rm44_T_met_norm0;

  unnorm_cams=cams;
//  rMatrix rm_inv_norm44(4,4);
//  rm_inv_norm44.MakeIdentity();
//  FORi(3) FORj(3) rm_inv_norm44[i][j]=normalizer[i][j];
  rMatrix rm_inv_norm(&inv_norm[0][0],3,3);
  FORi(cams.size) unnorm_cams[i]=rm_inv_norm*unnorm_cams[i];
best_K.DumpNormalized("best_K");

  //::: reconstruct scene
  Array<rPoint3> W;
  MiscUtility::Triangulate(W,cams,(*GLOB_Kpts));
  VrmlFile::OutputPoints("output.recon.wrl",W,RGB3Byte_green);
  Array<rPoint3> nul(0);  
  VrmlFile::OutputPointsAndCameras("output.world.and.cameras.wrl",W,nul,unnorm_cams,image_size);




  //::: write metric camera matrices for future use
  FILE* file=fopen("output.cameras","wb");
  ASSERT(file,"err632147");
  fprintf(file,"%d\n",cams.size);
//  rMatrix rm_inv_norm(&inv_norm[0][0],3,3);
  FORi(cams.size) {
    rMatrix cam=rm_inv_norm*cams[i];
    fprintf(file,"\n");
    FORj(3) fprintf(file,"%.9e %.9e %.9e %.9e\n",cam[j][0],cam[j][1],cam[j][2],cam[j][3]);
  }
  fclose(file);

  //::: output 3D --> 2D correspondence (for later texture mapping)
  file=fopen("given.dense.matching","rb");
  if (file) {
    int nViews, nPts;
    fscanf(file,"%d\n",&nPts);
    fscanf(file,"%d\n",&nViews);
    Array2D<rPoint3> dense(nViews,nPts);
    Transform T_normi_i=normalizer.Inverse();
    FORj(nViews) {
      FORk(nPts) {
	fscanf(file,"%lf %lf\n",&dense[j][k].x,&dense[j][k].y);
	dense[j][k].z=1.0;
	dense[j][k]=normalizer*dense[j][k];
	dense[j][k]/=dense[j][k].z;
      }
      fscanf(file,"\n");
    }
    fclose(file);
//	Transform T_met_0=normalizer.Inverse()*T_met_norm0;
////        MiscUtility::AffineTriangulate_WithInternalParams(W,Tab,T_met_norm0,E[i],dense[0],dense[i]);
////        file=fopen((String("mapping.")+String(i,"%02d")+".0").Ptr(),"w");
    FORk(W.size) {
      rPoint3 p=T_normi_i*dense[0][k];
      p/=p.z;
      fprintf(file,"%8e  %8e  %8e      %8e  %8e\n",W[k].x,W[k].y,W[k].z,p.x,p.y);
    }
    fclose(file);
  } else {
    const int view=0;
    FILE* file2=fopen((String("output.mapping.into.view")+String(view,"%02d")).Ptr(),"w");
    FORk(W.size) {
//      rPoint3 p=(*GLOB_Kpts)[view][k];
      rVector pp, P(4);
      P[0]=W[k].x;  P[1]=W[k].y;  P[2]=W[k].z;  P[3]=1.0;
      pp=cams[view]*P;
      rPoint3 p(pp[0],pp[1],pp[2]);
      p=inv_norm*p;
      p/=p.z;
      fprintf(file2,"%8e  %8e  %8e      %8e  %8e\n",W[k].x,W[k].y,W[k].z,p.x,p.y);
    }
    fclose(file2);
  }

EXIT("exit from OutputResults()");


  //::: bundle adjust:  use reconstruction to refine features and remove outlying features
  Array2D<rPoint3> newK(GLOB_Kpts->rows,GLOB_Kpts->cols);
  rVector P(4), p(3);
  P[3]=1.0;
  FORi(newK.rows) {
    const rMatrix& cam=cams[i];
    FORj(W.size) {
      P[0]=W[j].x;  P[1]=W[j].y;  P[2]=W[j].z;
      p=cam*P;
      newK[i][j]=rPoint3(p[0]/p[2], p[1]/p[2], 1.0);
    }
  }

  //::: show reconstruction projected into each view
  const int HHH=768, WWW=1024;
  Image3ByteChannels img(HHH,WWW);
  KltFeatureTable klt;
  klt.SetSize(newK[0].size,cams.size);
  FORk(cams.size) {
    img.FillWithColor(RGB3Byte_black);
    FORi(newK[k].size) {
      rPoint3 p=inv_norm*newK[k][i];
      p/=p.z;
      img.DrawSquare(rPoint2(p.x,p.y),3,RGB3Byte_green);
      rPoint3 q=inv_norm*(*GLOB_Kpts)[k][i];
      q/=q.z;
      img.DrawSquare(rPoint2(q.x,q.y),3,RGB3Byte_red);
      img.DrawLine(rPoint2(p.x,p.y),rPoint2(q.x,q.y),RGB3Byte_cyan);
      klt[i][k].msg=(k==0) ? 1 : 0;
      klt[i][k].x=q.x;		// store original pt
      klt[i][k].y=q.y;
    /*
    // for moving points slowly towards consensus
      klt[i][k].x=(p.x+q.x)/2;
      klt[i][k].y=(p.y+q.y)/2;
    //  if (sqrt(SQ(q.x-p.x)+SQ(q.y-p.y))>2) klt[i][0].msg=-1;
    */
    }
    img.WriteToFile((String("test")+String(k,"%02d")+".jpg").Ptr());
  }

  //::: here because of compiler bug
  Array<rMatrix> cms(klt.cols-1);
  FORi(cms.size) cms[i].SetSize(3,4);

  //::: find point that is most in disagreement with consensus and replace it with consensus
  FORi(klt.rows) {
    real worst_dist=-1;
    int worst_inx=-1;
//    rPoint3 worst_pt;
    FORj(klt.cols) {
      rPoint3 p=inv_norm*newK[j][i];
      p/=p.z;
      rPoint3 q=inv_norm*(*GLOB_Kpts)[j][i];
      q/=q.z;
      real dist=sqrt(SQ(q.x-p.x)+SQ(q.y-p.y));
      if ((worst_inx==-1)||(worst_dist<dist)) {
	worst_inx=j;
	worst_dist=dist;
//        worst_pt=p;		// store consensus pt for this bad feature
      }
    }
    //::: replace feature with consensus of *remaining* features
    Array2D<rPoint3> good_fe(klt.cols-1,1);
    int inx=0;
//  FORi(cms.size) FORk(3) FORl(4) cms[i][k][l]=cams[i][k][l];
    FORj(klt.cols) if (j!=worst_inx) {
      good_fe[inx][0]=(*GLOB_Kpts)[j][i];	// work in normalized image space
//      cms[inx].MakeCopyOf(cams[j]);	// very inefficient -- use ptrs? (must modify Triangulate)
      ++inx;
    }
    //::: DUE TO APPARENT COMPILER PROBLEM, HAVE TO COPY CAMS IN SEPARATE LOOP?!?!?!?!
    inx=0;
    FORk(klt.cols) if (k!=worst_inx) {
      cms[inx].MakeCopyOf(cams[k]);	// very inefficient -- use ptrs? (must modify Triangulate)
      ++inx;
    }
    Array<rPoint3> WW;
    MiscUtility::Triangulate(WW,cms,good_fe);
    rVector p, P(4);
    rPoint3 pp;
    P[0]=WW[0].x;  P[1]=WW[0].y;  P[2]=WW[0].z;  P[3]=1.0;
    p=cams[worst_inx]*P;
    pp=inv_norm*rPoint3(p[0],p[1],p[2]);
    pp/=pp.z;
    klt[i][worst_inx].x=pp.x;
    klt[i][worst_inx].y=pp.y;
//    klt[i][worst_inx].x=worst_pt.x;
//    klt[i][worst_inx].y=worst_pt.y;
  }

  //::: remove any row (feature) that has a -1 msg in the first view
  int c=0;
  FORi(klt.rows) if (klt[i][0].msg==1) ++c;
  KltFeatureTable klt2;
  klt2.SetSize(c,klt.cols);
  c=0;
  FORi(klt.rows) { if (klt[i][0].msg!=-1) { FORj(klt[i].size) klt2[c][j]=klt[i][j]; ++c; } }

  klt2.WriteToFile("given.features",HHH,WWW);
  EXIT();
}

#ifdef NOTDEFINED
//////////////////////////////////////////////////////////////////////
void IntersectionByTriangles::ReconstructScene(
  const AlphaCoord&		a
)
{
  //::: create projective camera matrices
  Array<rMatrix> cams(E.size);
  FORi(cams.size) if (i>0) {
    cams[i].SetSize(3,4);
    Transform Tab=T0i_proj[i]+Transform::Outer(E[i],a);
    FORj(3) FORk(3) cams[i][j][k]=Tab[j][k];
    cams[i][0][3]=E[i].x;
    cams[i][1][3]=E[i].y;
    cams[i][2][3]=E[i].z;
  }
  cams[0].SetSize(3,4);
  cams[0].FillWith(0);
  cams[0][0][0]=cams[0][1][1]=cams[0][2][2]=1;

  //::: use normalizing transform on camera matrices
  rMatrix KKK(4,4);
  KKK.MakeIdentity();
  FORi(3) FORj(3) KKK[i][j]=T_met_norm0[i][j];
  FORi(cams.size) cams[i]=cams[i]*KKK;

  //::: reconstruct scene
  Array<rPoint3> W;
  MiscUtility::Triangulate(W,cams,(*GLOB_Kpts));

/*
//::: write metric camera matrices for future use
FILE* file=fopen((String("recon.cameras.")+String(nZoominsPerformed,"%02d")).Ptr(),"wb");
ASSERT(file,"err0934034");
fprintf("%d\n",cams.size);
FORi(cams.size) {
  fprintf(file,"\n");
  FORj(3) fprintf(file,"%.9e %.9e %.9e %.9e\n",cams[i][j][0],cams[i][j][1],cams[i][j][2],cams[i][j][3]);
}
fclose(file);
*/

  InventorFile::OutputPoints(String("recon")+String(1,"%02d")+".iv",W,RGB3Byte_green);

#ifdef NOTDEFINED    
    //::: output 3D --> 2D correspondence for each view (for later texture mapping)
    if (i==1) {		//only doing first pair for now
      FILE* file=fopen("given.dense.matching","rb");
      if (file) {
        int nViews, nPts;
        fscanf(file,"%d\n",&nPts);
        fscanf(file,"%d\n",&nViews);
	Array2D<rPoint3> dense(nViews,nPts);
        Transform T_normi_i=normalizer.Inverse();
	FORj(nViews) {
	  FORk(nPts) {
	    fscanf(file,"%lf %lf\n",&dense[j][k].x,&dense[j][k].y);
	    dense[j][k].z=1.0;
	    dense[j][k]=normalizer*dense[j][k];
	    dense[j][k]/=dense[j][k].z;
	  }
	  fscanf(file,"\n");
	}
        fclose(file);
//	Transform T_met_0=normalizer.Inverse()*T_met_norm0;
        MiscUtility::AffineTriangulate_WithInternalParams(W,Tab,T_met_norm0,E[i],dense[0],dense[i]);
        file=fopen((String("mapping.")+String(i,"%02d")+".0").Ptr(),"w");
        FORk(W.size) {
          rPoint3 p=T_normi_i*dense[0][k];
	  p/=p.z;
          fprintf(file,"%8e  %8e  %8e      %8e  %8e\n",W[k].x,W[k].y,W[k].z,p.x,p.y);
        }
        fclose(file);
      } else {
        Transform T_normi_i=normalizer.Inverse();
        FILE* file2=fopen((String("mapping.")+String(i,"%02d")+".0").Ptr(),"w");
        FORk(W.size) {
          rPoint3 p=T_normi_i*((*GLOB_Kpts)[0][k]);
	  p/=p.z;
          fprintf(file2,"%8e  %8e  %8e      %8e  %8e\n",W[k].x,W[k].y,W[k].z,p.x,p.y);
        }
        fclose(file2);
      }
    }
#endif
/*
    if (pInternal!=NULL) {
      //::: use known internal calibration (NOTE: assumes already multiplied by Tinorm)
//      MiscUtility::AffineTriangulate_WithInternalParams(W,Tab,normalizer*(*pInternal),E[i],(*GLOB_Kpts)[0],(*GLOB_Kpts)[i]);
      MiscUtility::AffineTriangulate_WithInternalParams(W,T0i_proj[i]+Transform::Outer(E[i],true_alpha),*pInternal,E[i],(*GLOB_Kpts)[0],(*GLOB_Kpts)[i]);

      SCREENPUT InventorFile::OutputPoints(String("true.recon")+String(i,"%02d")+".iv",W,RGB3Byte_green);
//throw GeneticFillOutManifoldGroup3SpaceError();
    }
*/

}
#endif


