#include "ManifoldPointStruct.3Space.h"
#include "PlotManifolds.h"

#include "helpers.h"

///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
void VoxelCoord3Space::FindFrom(const AlphaCoord& p, const VolumeStruct& vol, const int dim)
{
  x1=int16(floor((p.x-vol.min.x)/(vol.max.x-vol.min.x)*dim));
  x2=int16(floor((p.y-vol.min.y)/(vol.max.y-vol.min.y)*dim));
  x3=int16(floor((p.z-vol.min.z)/(vol.max.z-vol.min.z)*dim));
}

//////////////////////////////////////////////////////////////////////
AlphaCoord VoxelCoord3Space::FindCenterInSpace(const VolumeStruct& vol, const int dim) const
{
  return AlphaCoord(
    (vol.max.x-vol.min.x)/dim*(x1+0.5)+vol.min.x,
    (vol.max.y-vol.min.y)/dim*(x2+0.5)+vol.min.y,
    (vol.max.z-vol.min.z)/dim*(x3+0.5)+vol.min.z
  );
}

//////////////////////////////////////////////////////////////////////
VolumeStruct VoxelCoord3Space::FindVolume(const VolumeStruct& vol, const int dim) const
{
  VolumeStruct V;
  V.min.x=(vol.max.x-vol.min.x)/dim*x1+vol.min.x;
  V.min.y=(vol.max.y-vol.min.y)/dim*x2+vol.min.y;
  V.min.z=(vol.max.z-vol.min.z)/dim*x3+vol.min.z;
  V.max.x=(vol.max.x-vol.min.x)/dim*(x1+1)+vol.min.x;
  V.max.y=(vol.max.y-vol.min.y)/dim*(x2+1)+vol.min.y;
  V.max.z=(vol.max.z-vol.min.z)/dim*(x3+1)+vol.min.z;
  return V;
}

//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
void DitherRangeStruct::CalcDitherRange(const real& sc)
{
  dk=(kmax-kmin)*sc;
  dth=(thmax-thmin)*sc;
  if ((abs(dk)>10*sc)||(abs(dth)>0.1*sc)) bSet=FALSE;		//range limits useless until zoomed-in enough
  bSet=TRUE;
}

//////////////////////////////////////////////////////////////////////
void DitherRangeStruct::Dump() const
{
  printf("kmax=%f kmin=%f thmax=%f thmin=%f dth=%f dk=%f\n",kmax,kmin,thmax,thmin,dth,dk);
  if (bSet) printf("VALID RANGE\n");
}

//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
boole		ManifoldPointStruct3Space_voting_space_set_flag=FALSE;
Transform	ManifoldPointStruct3Space_voting_space_normalizer;
Transform	ManifoldPointStruct3Space_voting_space_normalizer_inverse;
AlphaCoord	ManifoldPointStruct3Space_voting_space_centroid;

uint32 ManifoldPointStruct3Space::nBorn;         // for algorithm analysis

//////////////////////////////////////////////////////////////////////
ManifoldPointStruct3Space* ManifoldPointStruct3Space::Dopple() const
{
  return new ManifoldPointStruct3Space(*this);
}

//////////////////////////////////////////////////////////////////////
void ManifoldPointStruct3Space::Init(
  const real&			k,
  const real&			th,
  const ViewPairInfo&		pairinfo,
  const Array<Transform>&	T0i_proj,
  const Array<rPoint3>&		E
)
{
  ++nBorn;
  m.k=k;
  m.th=th;
  Transform Tij;
  AlphaCoord dum;
  bLegal=TRUE;
  if (!AffineCalibFrom2Params_efficient(Tij,dum,k,th,pairinfo)) { bLegal=FALSE; return; }
  PPOMCM_FindAlpha(a,Tij,T0i_proj,E,pairinfo.i,pairinfo.j);
  VOTE_NORM(a);
}

//////////////////////////////////////////////////////////////////////
ManifoldPointStruct3Space::ManifoldPointStruct3Space(
  const real&			k,
  const real&			th,
  const ViewPairInfo&		pairinfo,
  const Array<Transform>&	T0i_proj,
  const Array<rPoint3>&		E,
  const VolumeStruct&		volume,
  const int			CON_bacteria_dim,
  const int			CON_vote_dim
)
{
  Init(k,th,pairinfo,T0i_proj,E);
  if (!bLegal) return;

  vb.FindFrom(a,volume,CON_bacteria_dim);
  vv.FindFrom(a,volume,CON_vote_dim);
}

//////////////////////////////////////////////////////////////////////
ManifoldPointStruct3Space::ManifoldPointStruct3Space(
  const real&			k,
  const real&			th,
  const ViewPairInfo&		pairinfo,
  const Array<Transform>&	T0i_proj,
  const Array<rPoint3>&		E
)
{
  Init(k,th,pairinfo,T0i_proj,E);
}

//////////////////////////////////////////////////////////////////////
void ManifoldPointStruct3Space::RecalcVoxels(
  const VolumeStruct&		volume,
  const int			CON_bacteria_dim,
  const int			CON_vote_dim
)
{
  vb.FindFrom(a,volume,CON_bacteria_dim);
  vv.FindFrom(a,volume,CON_vote_dim);
}

//////////////////////////////////////////////////////////////////////
ManifoldPointStruct3Space ManifoldPointStruct3Space::Dither(
  const ViewPairInfo&		pairinfo,
  const Array<Transform>&	T0i_proj,
  const Array<rPoint3>&		E,
  const VolumeStruct&		volume,
  const int			CON_bacteria_dim,
  const int			CON_vote_dim
)
{
  real k, th;
  if (rand01()>0.1) {
    k=m.k+randpm()*pow(10,random(-4,4));
    th=m.th+randpm()*pow(10,random(-3,1));
  } else {
    k=m.k+randpm()*pow(10,random(-8,8));
    th=randpm()*PI;
  }
  return ManifoldPointStruct3Space(k,th,pairinfo,T0i_proj,E,volume,CON_bacteria_dim,CON_vote_dim);
}

//////////////////////////////////////////////////////////////////////
ManifoldPointStruct3Space ManifoldPointStruct3Space::DitherSmall(
  const ViewPairInfo&		pairinfo,
  const Array<Transform>&	T0i_proj,
  const Array<rPoint3>&		E,
  const VolumeStruct&		volume,
  const int			CON_bacteria_dim,
  const int			CON_vote_dim
)
{
  real k, th;
  real r=rand01();
  if (r>0.8) {
    k=m.k+randpm()*pow(10,random(-6,-1));
    th=m.th+randpm()*pow(10,random(-6,-1));
  } else
  if (r>0.1) {
    k=m.k+randpm()*pow(10,random(-4,1));
    th=m.th+randpm()*pow(10,random(-3,0));
  } else {
    k=m.k+randpm()*pow(10,random(-8,1));
    th=randpm()*PI;
  }
  return ManifoldPointStruct3Space(k,th,pairinfo,T0i_proj,E,volume,CON_bacteria_dim,CON_vote_dim);
}

//////////////////////////////////////////////////////////////////////
ManifoldPointStruct3Space ManifoldPointStruct3Space::RangeDither(
  const ViewPairInfo&		pairinfo,
  const Array<Transform>&	T0i_proj,
  const Array<rPoint3>&		E,
  const VolumeStruct&		volume,
  const int			CON_bacteria_dim,
  const int			CON_vote_dim,
  const DitherRangeStruct&	dither_range
)
{
  real k, th;
  real r=rand01();
  if (r>0.80) {
    k=m.k+randpm()*dither_range.dk*0.1;
    th=m.th+randpm()*dither_range.dth*0.1;
  } else
  if (r>0.30) {
    k=m.k+randpm()*dither_range.dk;
    th=m.th+randpm()*dither_range.dth;
  } else
  if (r>0.05) {
    k=m.k+randpm()*dither_range.dk*10;
    th=m.th+randpm()*dither_range.dth*10;
  } else {
//    k=m.k+randpm()*(1e20);
    k=m.k+randpm();
    th=randpm()*PI;
  }
  return ManifoldPointStruct3Space(k,th,pairinfo,T0i_proj,E,volume,CON_bacteria_dim,CON_vote_dim);
}

