#ifndef INCLUDED_ManifoldPointStruct3Space
#define INCLUDED_ManifoldPointStruct3Space

///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////

#include "Array.h"
#include "LinkedList.h"
#include "String.h"
#include "Point.h"
#include "Transform.h"
#include "RM_SameCameraManip.h"

///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////

typedef rPoint3 AlphaCoord;
typedef Array<AlphaCoord> Array_AlphaCoord;
typedef Array<rPoint3> Array_rPoint3;

///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
struct ManifoldCoord {
  real k;
  real th;

  void PutThetaInRange() { th=fmod(th,PI2); if (th<0) th+=PI2; if (th==PI2) th=0; }

  ManifoldCoord() { }  
  ManifoldCoord(const real _k, const real _th) { k=_k; th=_th; }
};

typedef Array<ManifoldCoord>	Array_ManifoldCoord;

//////////////////////////////////////////////////////////////////////
struct VolumeStruct {
  AlphaCoord min, max;
  
  boole bContains(const AlphaCoord& p) const
    { return (((p.x>=min.x)&&(p.x<max.x))&&
              ((p.y>=min.y)&&(p.y<max.y))&&
	      ((p.z>=min.z)&&(p.z<max.z))); }
  void Dump() const { min.Dump("volume min");  max.Dump("volume max"); }
  VolumeStruct* Dopple() const { return new VolumeStruct(*this); }

  VolumeStruct() { }  
  VolumeStruct(const AlphaCoord& MIN, const AlphaCoord& MAX) { min=MIN; max=MAX; }
  VolumeStruct(const VolumeStruct& V) { min=V.min; max=V.max; }
};

//////////////////////////////////////////////////////////////////////
struct VoxelCoord3Space {
  int16 x1,x2,x3;

  void FindFrom(const AlphaCoord& p, const VolumeStruct& vol, const int dim);
  AlphaCoord FindCenterInSpace(const VolumeStruct& vol, const int dim) const;
  VolumeStruct FindVolume(const VolumeStruct& vol, const int dim) const;
  boole Equals(const VoxelCoord3Space& v) { return ((x1==v.x1)&&(x2==v.x2)&&(x3==v.x3)); }
  void Dump() const { printf("%d  %d  %d\n",x1,x2,x3); }

  VoxelCoord3Space() { }  
  VoxelCoord3Space(const int y1, const int y2, const int y3) { x1=y1; x2=y2; x3=y3; }
};

//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
struct DitherRangeStruct {
  real			kmax, kmin;
  real			thmax, thmin;
  real			dth, dk;
  boole			bSet;			//flag to indicate whether dither range is meaningful

  void Dump() const;
  void CalcDitherRange(const real& sc=0.1);
  
  DitherRangeStruct() { bSet=FALSE; }
};

//////////////////////////////////////////////////////////////////////
extern boole		ManifoldPointStruct3Space_voting_space_set_flag;	//tells if normalization has been set yet
extern Transform	ManifoldPointStruct3Space_voting_space_normalizer;	//normalizes alpha coords
extern Transform	ManifoldPointStruct3Space_voting_space_normalizer_inverse;
extern AlphaCoord	ManifoldPointStruct3Space_voting_space_centroid;	//needed for normalization

#define VOTE_NORM(x) if (ManifoldPointStruct3Space_voting_space_set_flag) x=ManifoldPointStruct3Space_voting_space_normalizer*(x-ManifoldPointStruct3Space_voting_space_centroid)
#define VOTE_UNNORM(x) if (ManifoldPointStruct3Space_voting_space_set_flag) x=(ManifoldPointStruct3Space_voting_space_normalizer_inverse*x)+ManifoldPointStruct3Space_voting_space_centroid

//////////////////////////////////////////////////////////////////////
inline void ManifoldPointStruct3Space_SetVoteNormalizer(
  const Transform&	norm,
  const rPoint3&	centroid
)
{
  ManifoldPointStruct3Space_voting_space_normalizer=norm;
  ManifoldPointStruct3Space_voting_space_normalizer_inverse=norm.Inverse();
  ManifoldPointStruct3Space_voting_space_centroid=centroid;
  ManifoldPointStruct3Space_voting_space_set_flag=TRUE;
}

//////////////////////////////////////////////////////////////////////
struct ManifoldPointStruct3Space {
  AlphaCoord		a;	//alpha coords
  ManifoldCoord		m;	//manifold coords
  VoxelCoord3Space	vb;	//voxel coords of bacteria cell
  VoxelCoord3Space	vv;	//voxel coords of vote cell
  boole			bLegal;	//is this a legitimate manifold point?
  static uint32		nBorn;	//counts how many manifolds points generated overall (# calls to Init()...)

  void Init(const real&, const real&, const ViewPairInfo&, const Array<Transform>&, const Array<rPoint3>&);
  ManifoldPointStruct3Space* Dopple() const;
  ManifoldPointStruct3Space Dither(const ViewPairInfo&, const Array<Transform>&,
    const Array<rPoint3>&, const VolumeStruct&, const int, const int);
  ManifoldPointStruct3Space DitherSmall(const ViewPairInfo&, const Array<Transform>&,
    const Array<rPoint3>&, const VolumeStruct&, const int, const int);
  ManifoldPointStruct3Space RangeDither(const ViewPairInfo&, const Array<Transform>&,
    const Array<rPoint3>&, const VolumeStruct&, const int, const int, const DitherRangeStruct&);
  void RecalcVoxels(const VolumeStruct&, const int, const int);

  ManifoldPointStruct3Space() { }
  ManifoldPointStruct3Space(const ManifoldPointStruct3Space& x) { a=x.a; m=x.m; vb=x.vb; vv=x.vv; bLegal=x.bLegal; }
  ManifoldPointStruct3Space(const real&, const real&, const ViewPairInfo&, const Array<Transform>&,
    const Array<rPoint3>&, const VolumeStruct&, const int, const int);
//  ManifoldPointStruct3Space(const real&, const real&, const ViewPairInfo&,
//    const Array<Transform>&, const Array<rPoint3>&, const int, const int);
  ManifoldPointStruct3Space(const real&, const real&, const ViewPairInfo&,
    const Array<Transform>&, const Array<rPoint3>&);
};

///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////

#endif
