#include <stdio.h>

#include "helpers.h"
#include "String.h"
#include "Matrix.h"
#include "MiscUtility.h"
#include "ManifoldPointStruct.3Space.h"
#include "IntersectionByTriangles.h"
#include "PlotManifolds.h"
#include "RunningTimeDisplay.h"
#include "TestRunDataStruct.h"

//#include "VrmlFile.h"
//#include "Image3ByteChannels.h"
//#include "KltFeatureTable.h"

//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
void NormalizeVotingSpace(
  Array2D<real>&		Theta,
  Array2D<real>&		Kappa,
  Array<Array2D_rPoint3>&	lattice,
  Array<Array2D_boole>&		b_lattice
)
{
  SCREENPUT printf("BEGIN NORM\n");  fflush(stdout);

  const int CON_max_Msize		= 30000;
  const int CON_n_manifolds_to_use	= 3;		// max is lattice.size
  int32 n_all_lattice_pts = int32(real(CON_n_manifolds_to_use)*Kappa.cols*Theta.cols);
  Array<rPoint3> M;
  if (n_all_lattice_pts<=CON_max_Msize) {
    // full method -- original way, but often uses too many samples
    M.SetSize(CON_n_manifolds_to_use*Kappa.cols*Theta.cols);
    int inx=0;
    FORi(CON_n_manifolds_to_use) FORj(Theta.cols) FORk(Kappa.cols) M[inx++]=lattice[i][j][k];
  }
  else {
    // sampling method
    M.SetSize(CON_max_Msize);
    FORi(M.size) {
      int ii,jj,kk;
      do {
        ii=random(0,CON_n_manifolds_to_use-1);
        jj=random(0,Theta.cols-1);
        kk=random(0,Kappa.cols-1);
      } while (!b_lattice[ii][jj][kk]);
      M[i]=lattice[ii][jj][kk];
    }
  }

  rPoint3 median=rPoint3(0,0,0);
  FORi(M.size) median+=M[i];
  median/=M.size;
SCREENPUT median.Dump("MEDIAN");

  rMatrix Q(3,M.size);
  FORi(M.size) {
    const rPoint3& p=M[i];
    Q[0][i]=p.x-median.x; Q[1][i]=p.y-median.y; Q[2][i]=p.z-median.z;
  }
  rMatrix QQ=Q*(Q.TransposeOf());	// covariance matrix
  rMatrix U,S,V;
  QQ.full_Svd(U,S,V);
  V=V.TransposeOf();

  //::: transform samples so major axii are principal components
  Transform T(&V[0]);
  T.Invert();
  FORi(M.size) M[i]=T*(M[i]-median);

  //::: find standard deviation in each axis direction
  rPoint3 sd(0,0,0);
  FORi(M.size) {
    sd.x+=SQ(M[i].x-median.x);
    sd.y+=SQ(M[i].y-median.y);
    sd.z+=SQ(M[i].z-median.z);
  }

  sd.x=sqrt(sd.x/M.size);
  sd.y=sqrt(sd.y/M.size);
  sd.z=sqrt(sd.z/M.size);
SCREENPUT sd.Dump("standard deviation sd");

  //::: scale width of point set in each axis direction so that width is 1
  rPoint3 max, min;
  max=min=M[0];
  FORi(M.size) {
    const rPoint3& p=M[i];
    if (p.x>max.x) max.x=p.x;
    if (p.y>max.y) max.y=p.y;
    if (p.z>max.z) max.z=p.z;
    if (p.x<min.x) min.x=p.x;
    if (p.y<min.y) min.y=p.y;
    if (p.z<min.z) min.z=p.z;
  }
  Transform scale=Transform(1.0/(max.x-min.x),1.0/(max.y-min.y),1.0/(max.z-min.z));
////  //::: scale each axis so that standard deviation (or max width) is 1 in each direction
////  Transform scale=Transform(1.0/sd.x,1.0/sd.y,1.0/sd.z);
  Transform joe=scale*T;
  ManifoldPointStruct3Space_SetVoteNormalizer(joe,median);

/*
//now done later
  //::: normalize lattice points
  FORi(lattice.size) FORj(Theta.cols) FORk(Kappa.cols)
    VOTE_NORM(lattice[i][j][k]);
*/

/*
  //::: find covariance matrix of sample points
  //::: Q=[ q0 q1 q2 ... ] (==points as column vectors; qi=pi-median_i)
  //::: then PC's are eigenvectors of Q * Q'  ( <-- covariance matrix )
  rMatrix V,S;
  rPoint3 centroid;
  MiscUtility::FindPrincipalComponentsOf3DPointCloud(V,S,centroid,M);

  //::: create 3x3 array that transforms original alpha (voting) space into a space
  //::: where the principle components are orthonormal basis vectors; may also want
  //::: to transform centroid to origin...
  Transform norm;
  FORi(3) FORj(3) norm[i][j]=V[j][i];
norm=norm*Transform(sqrt(S[0][0]),sqrt(S[1][1]),sqrt(S[2][2]));
norm=norm.Inverse();
SCREENPUT norm.Dump("norm");
  ManifoldPointStruct3Space_SetVoteNormalizer(norm,centroid);
*/


/*  
  //::: find covariance matrix of sample points
  //::: Q=[ q0 q1 q2 ... ] (==points as column vectors; qi=pi-median_i)
  //::: then PC's are eigenvectors of Q * Q'  ( <-- covariance matrix )
  const Array<rPoint3>& P=M;
  rMatrix Q(3,P.size);
  rPoint3 median=rPoint3(0,0,0);
  FORi(P.size) median+=P[i];
  median/=P.size;
  FORi(P.size) { Q[0][i]=P[i].x-median.x; Q[1][i]=P[i].y-median.y; Q[2][i]=P[i].z-median.z; }
  rMatrix QQ=Q*(Q.TransposeOf());	// covariance matrix
  QQ*=1.0/P.size;
  QQ=QQ.InverseOf();
SCREENPUT median.Dump("centroid");
SCREENPUT QQ.Dump();

  //::: create 3x3 array that transforms original alpha (voting) space into a space
  //::: where the principle components are orthonormal basis vectors; may also want
  //::: to transform centroid to origin...
  Transform norm;
  FORi(3) FORj(3) norm[i][j]=QQ[i][j];
  norm.Normalize();
SCREENPUT norm.Dump("norm");
  ManifoldPointStruct3Space_SetVoteNormalizer(norm,median);
*/
  SCREENPUT printf("END NORM\n");  fflush(stdout);
}

//////////////////////////////////////////////////////////////////////
void SetLattice(
  Array2D<real>&		Theta,
  Array2D<real>&		Kappa,
  Array<Array2D_rPoint3>&	lattice,
  Array<Array2D_boole>&		b_lattice,

  const Array<real>&		min_th,
  const Array<real>&		max_th,
  const Array<real>&		min_ka,
  const Array<real>&		max_ka,
  const int			CON_n_theta,
  const int			CON_n_kappa,
  const int			CON_n_mani,

  const Array<ViewPairInfo>&	pairinfo,
  const Array<Transform>&	T0i_proj,
  const Array<rPoint3>&		E
)
{
/*
//::: for testing precise timer
INFINITELOOP {
  RunningTimeDisplay::SwissTimePush();
  FORi(1000) FORj(1000) FORk(1000) ;
  real t=RunningTimeDisplay::SwissTimePop();
  printf("%lf\n",t);  
}
*/

  Theta.SetSize(CON_n_mani,CON_n_theta);
  Kappa.SetSize(CON_n_mani,CON_n_kappa);
  lattice.SetSize(CON_n_mani);
  b_lattice.SetSize(CON_n_mani);
  FORi(CON_n_mani) lattice[i].SetSize(CON_n_theta,CON_n_kappa);
  FORi(CON_n_mani) b_lattice[i].SetSize(CON_n_theta,CON_n_kappa);

  //::: uniform-sample each manifold
  FORi(CON_n_mani) FORj(CON_n_kappa) Kappa[i][j]=min_ka[i]+(real(j)/(CON_n_kappa-1))*(max_ka[i]-min_ka[i]);
  FORi(CON_n_mani) FORj(CON_n_theta) Theta[i][j]=min_th[i]+(real(j)/(CON_n_theta-1))*(max_th[i]-min_th[i]);

  Transform Tab;
  rPoint3 alpha, dum;
  RunningTimeDisplay::SwissTimePush();
  SCREENPUT RunningTimeDisplay::Reset(lattice.size);
  FORk(lattice.size) {
    SCREENPUT RunningTimeDisplay::Tick();
    FORi(CON_n_theta) FORj(CON_n_kappa) {
      if (!AffineCalibFrom2Params_efficient(Tab,dum,Kappa[k][j],Theta[k][i],pairinfo[k])) EXIT("err0913247");
      PPOMCM_FindAlpha(alpha,Tab,T0i_proj,E,pairinfo[k].i,pairinfo[k].j);
      lattice[k][i][j]=alpha;
      b_lattice[k][i][j]=TRUE;
    }
  }
  SCREENPUT RunningTimeDisplay::Done();
  real runtime=RunningTimeDisplay::SwissTimePop();
  real time_per_node=runtime/(real(lattice.size)*CON_n_theta*CON_n_kappa);
  SCREENPUT printf("time to calc each lattice pt = %lf\n",time_per_node);

  GLOBAL_testrun_data.time_lattice_pt=time_per_node;

  const rPoint3 corners[]={rPoint3(100,0,0), rPoint3(0,100,0), rPoint3(0,0,100), rPoint3(-100,0,0), rPoint3(0,-100,0) };
  FORk(lattice.size) {
    FORi(Theta.cols) FORj(Kappa.cols) {
      rPoint3& p=lattice[k][i][j];
      if ( (!within(p.x,-100,100))||(!within(p.y,-100,100))||(!within(p.z,-100,100)) ) {
        p=corners[k%5];
	b_lattice[k][i][j]=FALSE;
      }
    }
  }
}

//////////////////////////////////////////////////////////////////////
void CreateTriangleClusters(
  Array<LinkedList_ClusterLatticeCluster>&	ll_clusters,
  const Array<Array2D_rPoint3>&			lattice,
  const Array<Array2D_boole>&			b_lattice,
  const int					CON_n_kappa,
  const int					CON_n_theta,
  const int					CON_ka_step,
  const int					CON_th_step
)
{
  //::: create clusters
  SCREENPUT { BANNER("creating clusters"); fflush(stdout); }
  FORi(ll_clusters.size) {
    ll_clusters[i].FreeMem();
    for (int j=0; j<CON_n_theta; j+=CON_th_step) {
      for (int k=0; k<CON_n_kappa; k+=CON_ka_step) {
        ClusterLatticeCluster* pc=new ClusterLatticeCluster;
        ClusterLatticeCluster& c=*pc;
        c.th0=j;
        c.th1=j+CON_th_step;
        if (c.th1 > CON_n_theta-1) c.th1=CON_n_theta-1;
        c.ka0=k;
        c.ka1=k+CON_ka_step;
        if (c.ka1 > CON_n_kappa-1) c.ka1=CON_n_kappa-1;
        c.b_legal=FALSE;
        c.b_tri_exist=FALSE;
        ll_clusters[i].Add(pc);
      }
    }
  }
  
  //::: find bounding box for each cluster
  FORi(ll_clusters.size) {
    FORLISTnow(ll_clusters[i], ClusterLatticeCluster)
      boole b_first_legal=TRUE;
      FORjIN(now.th0,now.th1) FORkIN(now.ka0,now.ka1) {
        if (!b_lattice[i][j][k]) continue;	// non-existent point--doesn't count
        const rPoint3& p=lattice[i][j][k];
        if (b_first_legal) {
          now.lo=now.hi=p;
          now.b_legal=TRUE;
          b_first_legal=FALSE;
        }
        else {
          if (p.x<now.lo.x)		now.lo.x=p.x;
          else if (p.x>now.hi.x)	now.hi.x=p.x;
          if (p.y<now.lo.y)		now.lo.y=p.y;
          else if (p.y>now.hi.y)	now.hi.y=p.y;
          if (p.z<now.lo.z)		now.lo.z=p.z;
          else if (p.z>now.hi.z)	now.hi.z=p.z;
        }
      }
      if ((now.lo.x==now.hi.x) || (now.lo.y==now.hi.y) || (now.lo.z==now.hi.z))
        now.b_legal=FALSE;	// trivial cluster--won't lead to meaningful mutual intersection pt
    }
  }

/*
  //::: display bounding boxes
  VrmlFile::Open("bounding.boxes.wrl");
  FORi(2) {
    FORLISTnow(ll_clusters[i], ClusterLatticeCluster*)
      if (!now->b_legal) continue;
      VrmlFile::OutputBox(now->lo,now->hi,color[i]);
    }
  }
  VrmlFile::Close();
*/

}
