#ifndef INCLUDED_MultiCameraWorkbench
#define INCLUDED_MultiCameraWorkbench

///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////

#include "Transform.h"
#include "Array.h"
#include "Array2D.h"
#include "String.h"
#include "ManifoldPointStruct.3Space.h"
#include "Matrix.h"

///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////

struct MultiCameraWorkbench {
  Array<rPoint3>		W;		// scene in world coords

  //::: normalized versions
  Array2D<rPoint3>		Kpts;		// scene projected onto retinas; [camera, point]
  Array<rMatrix>		CM;		// camera matrices
  Array2D<Transform>		Hinf;		// Hinf = T0i = Twi*inv(Tw0)
  Array2D<Transform>		F;
  Array2D<rPoint3>		eR, eL;		// eL' * F == F * eR == 0
  Array2D<Transform>		M;		// F == [eL]_x * M

  //::: misc
  Array<Transform>		Twi;		// affine transform from world to camera i
//  Array2D<Transform>		Hinf;		// Hinf = T0i = Twi*inv(Tw0)
  Array2D<rVector>		VectorA;	// Hinf == M + eL * VectorA'
  Array2D<Array_rPoint3>	R;		// projective reconstructions
  Array2D<rMatrix>		T;		// collineations between reconstructions
  Array2D<boole>			bPairOn;	// signals whether a view pair is in use
  Transform			internal;	// if same camera used throughout
  Transform			normalization;	// pretransform to roughly normalize the point data (needed to solve linear systems that have noise)

  CONST int			CON_nworldpts;
  CONST int			CON_ncameras;
  CONST int			CON_pairs;
  CONST real			CON_noiseradius;
  CONST iPoint2			CON_simulated_image_size;

  Array<real>			k, th, g;	// parameterization of affine and metric calibration
  Array2D<AlphaCoord>		alpha;

  void MakeInitialReconstructionRounder(const int one, const int two);
  void MakeInitialReconstructionRounder_2(const int one, const int two);
  void CalcAllVarsFromCameraMatrices(const int one, const int two);

  MultiCameraWorkbench(const int, const int, const real&, const int, const int,
			const real& CON_min_retinal_obj_size=-1.0, const iPoint2& _image_size=iPoint2(10,10));
  MultiCameraWorkbench(const String& file_features, const String& file_fmats, const int, const int);
};

///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////

#endif
