#include "PlotManifolds.h"
#include "helpers.h"
#include "RM_SameCameraManip.h"
#include "InventorFile.h"

///////////////////////////////////////////////////////////////////////
void PlotPointsOnModulusConstraintManifold(
  const Array<Transform>&	F,
  const Array<Transform>&	M,
  const Array<rPoint3>&		eB,
  const AlphaCoord&		true_alpha
)
{
  const int CON_npts=200;
  Transform Tab;
  Array2D<AlphaCoord> alpha(F.size,CON_npts);
  FORi(alpha.rows) { BANNER(""); FORj(CON_npts) {
  printf("."); fflush(stdout);
    do {
      real k=randpm()*3;
      real th=randpm()*PI;
      AffineCalibFrom2Params(Tab,alpha[i][j],k,th,F[i],M[i],eB[i]);
    } while (alpha[i][j].Norm()>15);
  } }
  static const RGB3Byte CON_cols[6]={RGB3Byte_green, RGB3Byte_red, RGB3Byte_cyan, RGB3Byte_blue, RGB3Byte_white, RGB3Byte_yellow };
  InventorFile::Open("hello.iv");
  FORi(alpha.rows) InventorFile::OutputPoints(alpha[i],CON_cols[i%6]);
//  Array_AlphaCoord extra(2); extra[0]=true_alpha;  extra[1]=true_alpha*(-1);
//  InventorFile::OutputPoints(extra,RGB3Byte_gray,10);
//  InventorFile::OutputLines(extra,RGB3Byte_blue);
  Array_AlphaCoord extra(1); extra[0]=true_alpha;
  InventorFile::OutputPoints(extra,RGB3Byte_gray,10);
  InventorFile::Close();
}

///////////////////////////////////////////////////////////////////////
void PPOMCM_FindAlpha(
  AlphaCoord&			alpha,
  const Transform&		Tij_met,
  const Array<Transform>&	T0i_proj,
  const Array<rPoint3>&		E,
  const int			i,
  const int			j
)
{
  if (i==0) {		//simpler case to handle because T00 is identity
    // Tij_met == T0j_proj + E[j]*alpha'
    static const rPoint3 v0(0,0,0);
    rMatrix Q(9,5);
    FillColumnWithTransform(Q,0,Tij_met);
    FillColumnWithTransform(Q,1,T0i_proj[j]);
    FillColumnWithTransform(Q,2,Transform(E[j],v0,v0));
    FillColumnWithTransform(Q,3,Transform(v0,E[j],v0));
    FillColumnWithTransform(Q,4,Transform(v0,v0,E[j]));
    rVector v;
    Q.FindNullEigenvector(v);
    alpha.x=v[2]/v[1];
    alpha.y=v[3]/v[1];
    alpha.z=v[4]/v[1];
  } else {
    /*  // old way--not least squares
    rPoint3 m1(T0i_proj[j][0][0],T0i_proj[j][1][0],T0i_proj[j][2][0]);
    rPoint3 m2(T0i_proj[j][0][1],T0i_proj[j][1][1],T0i_proj[j][2][1]);
    rPoint3 m3(T0i_proj[j][0][2],T0i_proj[j][1][2],T0i_proj[j][2][2]);
    Transform D=Tij_met*T0i_proj[i];
    rPoint3 q1(D[0][0],D[1][0],D[2][0]);
    rPoint3 q2(D[0][1],D[1][1],D[2][1]);
    rPoint3 q3(D[0][2],D[1][2],D[2][2]);
    rPoint3 v1=m1.Cross(m2);
    rPoint3 v2=(q1.Cross(m2)+m1.Cross(q2))*(-1);
    rPoint3 v3=q1.Cross(q2);
    // phi^2 * v1 + phi*v2 + v3 = 0
    real scale=v1.x/v1.y;
    real phi=-(v3.x-scale*v3.y)/(v2.x-scale*v2.y);
    rPoint3 t1=m1*phi-q1;
    rPoint3 t2=m2*phi-q2;
    rPoint3 t3=m3*phi-q3;
//    (t1).Cross(t2).Dump("0");
//    (t1).Cross(t3).Dump("0");
//    (t2).Cross(t3).Dump("0");
    rPoint3 e=(E[j]*phi-Tij_met*E[i])*(-1);
//    e.Cross(t1).Dump("0");
    alpha.x=t1.x/e.x;			// should use least squares...
    alpha.y=t2.x/e.x;
    alpha.z=t3.x/e.x;
    */

    rPoint3 m1(T0i_proj[j][0][0],T0i_proj[j][1][0],T0i_proj[j][2][0]);
    rPoint3 m2(T0i_proj[j][0][1],T0i_proj[j][1][1],T0i_proj[j][2][1]);
//    rPoint3 m3(T0i_proj[j][0][2],T0i_proj[j][1][2],T0i_proj[j][2][2]);
    Transform D=Tij_met*T0i_proj[i];
    rPoint3 q1(D[0][0],D[1][0],D[2][0]);
    rPoint3 q2(D[0][1],D[1][1],D[2][1]);
//    rPoint3 q3(D[0][2],D[1][2],D[2][2]);
    rPoint3 v1=m1.Cross(m2);
    rPoint3 v2=q1.Cross(m2)+m1.Cross(q2);
    rPoint3 v3=q1.Cross(q2);
    // phi^2 * v1 + phi*v2 + v3 = 0
    //::: NOTE!!! error in original ECCV02 paper:  phi in paper has wrong sign; corrected below
    real phi=-(v3.x*v1.y-v3.y*v1.x)/(v2.x*v1.y-v2.y*v1.x);
/*    if (rand01()<0.01) {
      (v1*(phi*phi)+v2*phi+v3).Dump();
      EXIT("uiodsafu");
    }*/
    Transform U=Tij_met*T0i_proj[i]+(T0i_proj[j]*phi);
    rPoint3 w=Tij_met*E[i]+(E[j]*phi);
    static const rPoint3 v0(0,0,0);
    rMatrix Q(9,4);
    FillColumnWithTransform(Q,0,U);
    FillColumnWithTransform(Q,1,Transform(w,v0,v0));
    FillColumnWithTransform(Q,2,Transform(v0,w,v0));
    FillColumnWithTransform(Q,3,Transform(v0,v0,w));
    rVector v;
    Q.FindNullEigenvector(v);
    alpha.x=v[1]/v[0];
    alpha.y=v[2]/v[0];
    alpha.z=v[3]/v[0];
  }

/*
printf("%d   %d\n",i,j);  
alpha.Dump("alpha");
Transform ANS1=T0i_proj[j]+Transform::Outer(E[j],alpha);
Transform ANS2=T0i_proj[i]+Transform::Outer(E[i],alpha);
ANS1.DumpNormalized("j");
ANS2.DumpNormalized("i");
Tij_met.DumpNormalized("target");
//EXIT("");
LINE();
*/

}

/*\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
----------------------------------------------------------------------
the given M and eB determine the parametrization of manifold for F;
T0i_met == T0i_proj + E*alpha'    (met=metric, proj=projective)
Tij_met == (T0j_met)*inv(T0i_met)
algorithm:  guess Tij_met using parameterization; use this guess to find alpha;
plot alpha to show manifold
NOTE: for i=0, T0i_met=identity and method reduces to earlier method
----------------------------------------------------------------------
\*////////////////////////////////////////////////////////////////////
void PlotPointsOnModulusConstraintManifold(
  const Array<Transform>&	T0i,	//relative transforms in chosen projective space
  const Array<rPoint3>&		E,
  const Array2D<Transform>&	F,
  const Array2D<Transform>&	M,
  const Array2D<rPoint3>&	eB,
  const int			CON_nmanifolds_to_plot
)
{
  const int CON_npts=500;
  const int CON_ncameras=F.rows;

  Array<Array_AlphaCoord> alpha(CON_nmanifolds_to_plot);	//set of manifold points for each view pair
  FORi(alpha.size) alpha[i].SetSize(CON_npts);

  int index=0;
  AlphaCoord dum;
  Transform Tij;
  FORi(CON_ncameras) FORj(CON_ncameras) if (i<j) {
    BANNER("");
    FORl(CON_npts) {
      printf("."); fflush(stdout);
      do {
        real k=randpm()*3;
        real th=randpm()*PI;
        AffineCalibFrom2Params(Tij,dum,k,th,F[i][j],M[i][j],eB[i][j]);
	PPOMCM_FindAlpha(alpha[index][l],Tij,T0i,E,i,j);	//could utilize dum here for speed up, if M,eB chosen carefully
      } while (alpha[index][l].Norm()>15);
    }
    ++index;
    if (index==CON_nmanifolds_to_plot) goto plot_manifolds;
  }

plot_manifolds:
/*
  static const RGB3Byte CON_cols[6]={RGB3Byte_green, RGB3Byte_red, RGB3Byte_cyan, RGB3Byte_blue, RGB3Byte_white, RGB3Byte_yellow };
  InventorFile::Open("hello.iv");
  FORi(alpha.size) InventorFile::OutputPoints(alpha[i],CON_cols[i%6]);
//  Array_AlphaCoord extra(2); extra[0]=true_alpha;  extra[1]=true_alpha*(-1);
//  InventorFile::OutputPoints(extra,RGB3Byte_gray,10);
//  InventorFile::OutputLines(extra,RGB3Byte_blue);
  Array_AlphaCoord extra(1); extra[0]=true_alpha;
  InventorFile::OutputPoints(extra,RGB3Byte_gray,10);
  InventorFile::Close();
*/

  //::: manifold-based voting procedure
  const int CON_voxel_dim=20;
  const AlphaCoord min(-10,-10,-10), max(10,10,10);
  byte vote[CON_voxel_dim][CON_voxel_dim][CON_voxel_dim];
  boole bmarked[CON_voxel_dim][CON_voxel_dim][CON_voxel_dim];
  FORi(CON_voxel_dim) FORj(CON_voxel_dim) FORk(CON_voxel_dim) vote[i][j][k]=0;
  FORi(alpha.size) {
    FORj(CON_voxel_dim) FORk(CON_voxel_dim) FORl(CON_voxel_dim) bmarked[j][k][l]=FALSE;
    FORj(alpha[i].size) {
      AlphaCoord& a=alpha[i][j];
      if (((a.x>=min.x)&&(a.x<max.x))&&
          ((a.y>=min.y)&&(a.y<max.y))&&
	  ((a.z>=min.z)&&(a.z<max.z))) {
        int ix=_floor((a.x-min.x)/(max.x-min.x)*CON_voxel_dim);
        int iy=_floor((a.y-min.y)/(max.y-min.y)*CON_voxel_dim);
        int iz=_floor((a.z-min.z)/(max.z-min.z)*CON_voxel_dim);
        if (!bmarked[ix][iy][iz]) {
	  bmarked[ix][iy][iz]=TRUE;
	  ++vote[ix][iy][iz];
	}
      }
    }
  }

  uint32 nvotes=0;
  FORi(CON_voxel_dim) FORj(CON_voxel_dim) FORk(CON_voxel_dim)
    if (vote[i][j][k]>0) ++nvotes;
  printf("nvotes=%d\n",nvotes);
  Array<RGB3Byte> cols(nvotes);
  Array_AlphaCoord pts(nvotes);
  uint32 ind=0;
  FORi(CON_voxel_dim) FORj(CON_voxel_dim) FORk(CON_voxel_dim)
    if (vote[i][j][k]>0) {
      pts[ind]=AlphaCoord( (max.x-min.x)/CON_voxel_dim*(i+0.5),
        (max.y-min.y)/CON_voxel_dim*(j+0.5), (max.z-min.z)/CON_voxel_dim*(k+0.5));
      cols[ind]=RGB3Byte_white*(real(vote[i][j][k])/CON_nmanifolds_to_plot);
      if (vote[i][j][k]==CON_nmanifolds_to_plot) cols[ind]=RGB3Byte_green;
      if (vote[i][j][k]==CON_nmanifolds_to_plot-1) cols[ind]=RGB3Byte_red;
///        cols[ind]=(vote[i][j][k]==1) ? RGB3Byte_blue : (vote[i][j][k]==2) ? RGB3Byte_red :
///	          (vote[i][j][k]==CON_nmanifolds_to_plot) ? RGB3Byte_white : RGB3Byte_green;
//      cols[ind]=RGB3Byte_green;
      ++ind;
    }
  InventorFile::OutputColoredPoints("vote.iv",pts,cols,20);
  
  
}

///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////

#ifdef NOTDEFINED
///////////////////////////////////////////////////////////////////////
void ShowManifoldFixedTheta(
  const Array<Array_Transform>&	F,
  const Array<Array_Transform>&	M,
  const Array<Array_rPoint3>&	eB,
  const AlphaCoord&		true_alpha,

  const Array<Transform>&	T0i_proj,
  const Array<rPoint3>&		E,
  const int			one,
  const int			two
)
{
  real th=PI*.1;
//  real th2=PI*.12;
//  real th3=PI*.14;
  real k=1e2;
  ManifoldPointStruct p;
  LinkedList<AlphaCoord> man;
  do {
    p=ManifoldPointStruct(k,th,F[one][two],M[one][two],eB[one][two],T0i_proj,E,one,two);
    man.Add(new AlphaCoord(p.a));
    p=ManifoldPointStruct(-k,th,F[one][two],M[one][two],eB[one][two],T0i_proj,E,one,two);
    man.Add(new AlphaCoord(p.a));
/*    p=ManifoldPointStruct(-k,th2,F[one][two],M[one][two],eB[one][two],T0i_proj,E,one,two);
    man.Add(new AlphaCoord(p.a));
    p=ManifoldPointStruct(-k,th2,F[one][two],M[one][two],eB[one][two],T0i_proj,E,one,two);
    man.Add(new AlphaCoord(p.a));
    p=ManifoldPointStruct(-k,th3,F[one][two],M[one][two],eB[one][two],T0i_proj,E,one,two);
    man.Add(new AlphaCoord(p.a));
    p=ManifoldPointStruct(-k,th3,F[one][two],M[one][two],eB[one][two],T0i_proj,E,one,two);
    man.Add(new AlphaCoord(p.a));*/
    k/=1.2;
  } while (k>1e-3);

  Array_AlphaCoord manifold;
  manifold.MakeCopyOf(man);
  InventorFile::Open("fixed.theta.iv");
  InventorFile::OutputPoints(manifold,RGB3Byte_green);
  Array_AlphaCoord extra(1); extra[0]=true_alpha;
  InventorFile::OutputPoints(extra,RGB3Byte_gray,10);
  Array_AlphaCoord box(2);
  box[0]=AlphaCoord(-20,-20,-20); box[1]=AlphaCoord(20,20,20);
  InventorFile::OutputPoints(box,RGB3Byte_orange,10);
  InventorFile::Close();
}

/*
//////////////////////////////////////////////////////////////////////
void ShowManifold(
  const Array2D<Transform>&	F,
  const Array2D<Transform>&	M,
  const Array2D<rPoint3>&	eB,
  const AlphaCoord&		true_alpha,

  const Array<Transform>&	T0i_proj,
  const Array<rPoint3>&		E,
  const int			one,
  const int			two
)
{
  VolumeStruct region(AlphaCoord(-20,-20,-20),AlphaCoord(20,20,20));
  real k, th;
  ManifoldPointStruct p;
  LinkedList<AlphaCoord> man;
  FORi(10000) {
    k=random_real();
    th=random_angle();
    p=ManifoldPointStruct(k,th,F[one][two],M[one][two],eB[one][two],T0i_proj,E,one,two);
    if (p.bLegal && region.bContains(p.a)) man.Add(new AlphaCoord(p.a));
  }

  Array_AlphaCoord manifold;
  manifold.MakeCopyOf(man);
  InventorFile::Open("all.manifold.iv");
  InventorFile::OutputPoints(manifold,RGB3Byte_green);
  Array_AlphaCoord extra(1); extra[0]=true_alpha;
  InventorFile::OutputPoints(extra,RGB3Byte_gray,10);
  Array_AlphaCoord box(8);
  box[0]=region.min;
  box[1]=region.max;
  box[2]=AlphaCoord(region.min.x,region.min.y,region.max.z);
  box[3]=AlphaCoord(region.min.x,region.max.y,region.max.z);
  box[4]=AlphaCoord(region.max.x,region.max.y,region.max.z);
  box[5]=AlphaCoord(region.max.x,region.max.y,region.min.z);
  box[6]=AlphaCoord(region.max.x,region.min.y,region.min.z);
  box[7]=AlphaCoord(region.min.x,region.max.y,region.min.z);
  InventorFile::OutputPoints(box,RGB3Byte_orange,8);
  InventorFile::Close();
}*/

#endif

