#include <stdio.h>

#include "helpers.h"
#include "String.h"
#include "Matrix.h"
#include "VrmlFile.h"
#include "MiscUtility.h"
#include "ManifoldPointStruct.3Space.h"
#include "IntersectionByTriangles.h"
#include "PlotManifolds.h"
#include "RunningTimeDisplay.h"
#include "TestRunDataStruct.h"

#include "Image3ByteChannels.h"
#include "KltFeatureTable.h"

//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

// taken from another program; put in standard library!
//////////////////////////////////////////////////////////////////////
void RandomIndexSet(
  Array<int>&			set,
  const int			max
)
{
  int top=0;
  FORi(set.size) {
    int x;
    INFINITELOOP {
      x=random(0,max-1);
      if (top==0) break;
      boole bFound=FALSE;
      FORj(top) if (set[j]==x) { bFound=TRUE; break; }
      if (!bFound) break;
    }
    set[top++]=x;
  }
}

//////////////////////////////////////////////////////////////////////
void RandomIndexSet(
  Array<int>&			set,
  const int			n,
  const int			max
)
{
  set.SetSize(n);
  RandomIndexSet(set,max);
}

//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
// put in standard library

#ifdef NOTDEFINED
// working permutation creator
//////////////////////////////////////////////////////////////////////
void CreatePermutations(
  Array2D<uint8>&		P,
  const int			n
)
{
  P.FreeMem();
  if (n<=0) return;
  if (n>8) EXIT("CreatePermutations() probably can't handle n");
  uint32 nfact=1;
  FORiIN(2,n) nfact*=i;
  P.SetSize(nfact,n);
  
  //::: fill P with all permutations in special form
  FORi(n) {
    int r=0;
    int m=n-1-i;
    int mfact=1;
    FORjIN(1,m) mfact*=j;
    int inx=0;
    while (r<nfact) {
      FORj(mfact) P[r++][i]=inx;
      ++inx;
      if (inx==m+1) inx=0;
    }
  }

  //::: convert special form to standard form
  Array<uint8> avail(n);
  const int nminus1=n-1;
  FORi(nfact) {
    FORj(n) avail[j]=j;
    FORj(n) {
      int t=P[i][j];
      P[i][j]=avail[t];
      FORkIN(t+1,nminus1) avail[k-1]=avail[k];
    }
  }

  FORi(nfact) {
    FORj(n) printf("%d  ",P[i][j]);
    printf("\n");
  }
  EXIT();
}
#endif

// return array of all combinations "n choose q"
//////////////////////////////////////////////////////////////////////
void CreateCombinations(
  Array2D<uint8>&		P,
  const int			n,
  const int			q
)
{
  P.FreeMem();
  if (n<1) return;
  if (q<1) return;
  ASSERT(q<=n,"bad input to CreateCombinations()");
  if (n>8) EXIT("CreateCombinations() probably can't handle n");
  uint32 nchooseq=1;
  FORiIN(q+1,n) nchooseq*=i;
  FORiIN(2,n-q) nchooseq/=i;
  P.SetSize(nchooseq,q);

  if (q==1) {
    FORi(P.rows) P[i][0]=i;
  }
  else {
    int qm1=q-1;
    FORi(q) P[0][i]=i;
    
    for (int r=1; r<P.rows; ++r) {
      FORi(q) P[r][i]=P[r-1][i];
      ++P[r][qm1];
      if (P[r][qm1]==n) {
        int j=q-2;
        while (P[r][j]==n-1-(qm1-j)) --j;
        ++P[r][j];
        FORkIN(j+1,qm1) P[r][k]=P[r][k-1]+1;
      }
    }
  }

/*
  FORi(P.rows) {
    FORj(P.cols) printf("%d  ",P[i][j]);
    printf("\n");
  }
  printf("\n");
*/
}

//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
struct ClusterTriangle {
  rPoint3			a,b,c;			// vertices in search space coords
  
  rPoint3			u, v, n, u_cr_v;	// precomputed values for faster triangle intersection testing
  real				a_th, a_ka;		// theta and kappa coords for vertex a
  real				b_th, b_ka;
  real				c_th, c_ka;

  void Init(const rPoint3& A, const rPoint3& B, const rPoint3& C);

  ClusterTriangle() { }
  ClusterTriangle(const rPoint3& A, const rPoint3& B, const rPoint3& C) { Init(A,B,C); }
};

void ClusterTriangle::Init(const rPoint3& A, const rPoint3& B, const rPoint3& C)
{
  a=A;  b=B;  c=C;
  u=b-a;
  v=c-a;
  n=(u.Cross(v)).Hat();
  u_cr_v=u.Cross(v);
}

//////////////////////////////////////////////////////////////////////
struct ClusterLatticeCluster {
  int				th0, th1;
  int				ka0, ka1;
  rPoint3			lo, hi;		// the bounding box
  boole				b_legal;	// does cluster have at least one legal lattice point?
  LinkedList<ClusterTriangle>	tri;		// triangles for this cluster
  boole				b_tri_exist;	// have tris been generated yet?
};

//////////////////////////////////////////////////////////////////////
struct ClusterPair {
  ClusterLatticeCluster*	a;
  ClusterLatticeCluster*	b;
};

//////////////////////////////////////////////////////////////////////
struct ClusterTrianglePair {
  ClusterTriangle		a, b;
  rPoint3			lo, hi;		// bounding box for pair
};

//////////////////////////////////////////////////////////////////////
struct ClusterTriangleTriplet {
  ClusterTriangle		a, b, c;
};

//////////////////////////////////////////////////////////////////////
typedef LinkedList<ClusterTriangle> LinkedList_ClusterTriangle;

// my own code -- one I downloaded didn't work!
//////////////////////////////////////////////////////////////////////
int tri_tri_intersect(
  const ClusterTriangle&	S,
  const ClusterTriangle&	T
)
{
/*
  rPoint3 T_u=T.b-T.a;		// should precompute in tri data struct
  rPoint3 T_v=T.c-T.a;
  rPoint3 T_n=(T_u.Cross(T_v)).Hat();
  rPoint3 u_cr_v=T_u.Cross(T_v);
//  rPoint3 S_n=((S.b-S.a).Cross(S.c-S.a)).Hat();
//  if (T_n.Cross(S_n).Norm()<1e-8) return 0;	// triangles almost parallel
*/

  //::: determine which side of T each vertex of S is on
  rPoint3 A=S.a-T.a;
  rPoint3 B=S.b-T.a;
  rPoint3 C=S.c-T.a;
  real A_alpha=(T.n).Dot(A);
  real B_alpha=(T.n).Dot(B);
  real C_alpha=(T.n).Dot(C);
  int A_sign=(A_alpha<0) ? -1 : 1;
  int B_sign=(B_alpha<0) ? -1 : 1;
  int C_sign=(C_alpha<0) ? -1 : 1;
  if ((A_sign==B_sign)&&(B_sign==C_sign)) return 0;	// all on same side
  //::: (*) now know 2 of the signs are equal
  if (A_sign==B_sign) {		// C is odd
    swap(A,C);			// use pointers for speed ?
  }
  else if (A_sign==C_sign) {	// B is odd
    swap(A,B);
  }
  // ...else A is odd, do nothing

/*
Transform M(C-A,T.u,T.v);
Transform invM=M.Inverse();
(invM*(A)).Dump();
(invM*(A)).Dump();
//EXIT("dkjsaf");
invM.Dump("invM");
(T.u.Cross(T.v)).Dump();
(T.v.Cross(C-A)).Dump();
((C-A).Cross(T.u)).Dump();
*/
  //::: find coords of where edges of S hit plane of T, in T-based coords
  rPoint3 AB=B-A;
  rPoint3 AC=C-A;
  rPoint3 AB_cp_A=AB.Cross(A);
  real inv_trip_prod_AB=1.0/(AB.Dot(T.u_cr_v));	// scale for inverse matrix
  real AB_beta  =   (T.v).Dot(AB_cp_A)*inv_trip_prod_AB;
  real AB_gamma = -((T.u).Dot(AB_cp_A))*inv_trip_prod_AB;
  rPoint3 AC_cp_A=AC.Cross(A);
  real inv_trip_prod_AC=1.0/(AC.Dot(T.u_cr_v));	// scale for inverse matrix
  real AC_beta  =   (T.v).Dot(AC_cp_A)*inv_trip_prod_AC;
  real AC_gamma = -((T.u).Dot(AC_cp_A))*inv_trip_prod_AC;

  //::: test if line segment formed by [S intersect T-plane] hits T
  if ((AB_beta<0)&&(AC_beta<0)) return 0;	// both hitpoints in quad II or III
  if ((AB_gamma<0)&&(AC_gamma<0)) return 0;     // both hitpoints in quad IV or III
  //::: (*) sort hitpoints by beta coord
  if (AB_beta>AC_beta) {
    swap(AB_beta,AC_beta);
    swap(AB_gamma,AC_gamma);
  }
  //::: (*) check if AC hits within triangle T
  if ((AC_gamma>=0)&&(AC_beta+AC_gamma<=1)) return 1;	// yes; we know AC_beta>=0

  if (AB_beta<0) {	// know AB hits in quad II or III and AC hits in quad I or IV
    // neither hitpoint within triangle T, so check if line between hitpoints is
    // (*) find where line hits gamma-axis
    rPoint3 dum=rPoint3(AB_beta,AB_gamma,1.0).Cross(rPoint3(AC_beta,AC_gamma,1.0));
    rPoint3 hit=rPoint3(1,0,0).Cross(dum);
    hit.y/=hit.z;
    if (hit.y<=0) {
      // check if hitpoint for AC is in quad I
      if (AC_gamma<=0) return 0;	// line seg hits below beta axis (in quad IV)
      // check where seg crosses beta axis (done later)
    }
    else if (hit.y<=1) return 1;	// line seg hits left-hand wall of T
    else if (AC_gamma>=0) return 0;	// so can't hit T since AC hits in quad I
    // (*) last chance: must hit bottom wall of T (so intersect with beta-axis)
    hit=rPoint3(0,1,0).Cross(dum);
    hit.x/=hit.z;
    return ((hit.x>=0.0)&&(hit.x<=1.0));
  }
  else {		// know both hitpoints are in quads I or IV
    if (AB_beta>=1.0) return 0;	// both hitpoints right of line beta=1
    //::: test if AB hits in triangle T
    if ((AB_gamma>=0)&&(AB_beta+AB_gamma<=1)) return 1;	// yes it does
    // ...else neither hitpoint within triangle T, so check if line between hitpoints is
    // check if both hitpoints in quad I (but neither in T)
    if ((AB_gamma>=0)&&(AC_gamma>=0)) return 0;
    // one hitpoint in quad I other in quad IV; see where between hits beta axis
    rPoint3 dum=rPoint3(AB_beta,AB_gamma,1.0).Cross(rPoint3(AC_beta,AC_gamma,1.0));
    rPoint3 hit=rPoint3(0,1,0).Cross(dum);
    hit.x/=hit.z;
    return (hit.x<=1);
  }
}

//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
//const RGB3Byte color[6]={ RGB3Byte_red, RGB3Byte_green, RGB3Byte_cyan,
//  RGB3Byte_white, RGB3Byte_blue, RGB3Byte_yellow };

//////////////////////////////////////////////////////////////////////
IntersectionByTriangles::IntersectionByTriangles(
  const Array<ViewPairInfo>&	_pairinfo,
  const Array<Transform>&	_T0i_proj,
  const Array<rPoint3>&		_E,
  const AlphaCoord&		_true_alpha,
  const Transform&		_Tinorm,
  Array2D<rPoint3>*		_GLOB_Kpts,
  Transform* const		_pInternal
) :
  pairinfo(_pairinfo), T0i_proj(_T0i_proj), E(_E), __true_alpha(_true_alpha),
  normalizer(_Tinorm), GLOB_Kpts(_GLOB_Kpts), pInternal(_pInternal),

  CON_n_manifolds(pairinfo.size)
{
  true_alpha=__true_alpha;
}

//////////////////////////////////////////////////////////////////////
IntersectionByTriangles::~IntersectionByTriangles()
{
}

//////////////////////////////////////////////////////////////////////
void IntersectionByTriangles::NormalizeVotingSpace()
{
  SCREENPUT printf("BEGIN NORM\n");  fflush(stdout);

  Array<rPoint3> M(lattice.size*Kappa.cols*Theta.cols);
  int inx=0;
  FORi(lattice.size) FORj(Theta.cols) FORk(Kappa.cols) M[inx++]=lattice[i][j][k];
//  M.Dump();
//  EXIT();

  rMatrix Q(3,M.size);
  rPoint3 median=rPoint3(0,0,0);
  FORi(M.size) median+=M[i];
  median/=M.size;
  FORi(M.size) {
    const rPoint3& p=M[i];
    Q[0][i]=p.x-median.x; Q[1][i]=p.y-median.y; Q[2][i]=p.z-median.z;
  }
  rMatrix QQ=Q*(Q.TransposeOf());	// covariance matrix
  rMatrix U,S,V;
  QQ.full_Svd(U,S,V);
  V=V.TransposeOf();
  //::: transform samples so major axii are principal components
  Transform T(&V[0]);
  T.Invert();
  FORi(M.size) M[i]=T*(M[i]-median);

  //::: find standard deviation in each axis direction
  rPoint3 sd(0,0,0);
  FORi(M.size) {
    sd.x+=SQ(M[i].x-median.x);
    sd.y+=SQ(M[i].y-median.y);
    sd.z+=SQ(M[i].z-median.z);
  }
  sd.x=sqrt(sd.x/M.size);
  sd.y=sqrt(sd.y/M.size);
  sd.z=sqrt(sd.z/M.size);
SCREENPUT sd.Dump("standard deviation sd");

  //::: scale width of point set in each axis direction so that width is 1
  rPoint3 max, min;
  max=min=M[0];
  FORi(M.size) {
    const rPoint3& p=M[i];
    if (p.x>max.x) max.x=p.x;
    if (p.y>max.y) max.y=p.y;
    if (p.z>max.z) max.z=p.z;
    if (p.x<min.x) min.x=p.x;
    if (p.y<min.y) min.y=p.y;
    if (p.z<min.z) min.z=p.z;
  }
  Transform scale=Transform(1.0/(max.x-min.x),1.0/(max.y-min.y),1.0/(max.z-min.z));
////  //::: scale each axis so that standard deviation (or max width) is 1 in each direction
////  Transform scale=Transform(1.0/sd.x,1.0/sd.y,1.0/sd.z);
  Transform joe=scale*T;
  ManifoldPointStruct3Space_SetVoteNormalizer(joe,median);

/*
//now done later
  //::: normalize lattice points
  FORi(lattice.size) FORj(Theta.cols) FORk(Kappa.cols)
    VOTE_NORM(lattice[i][j][k]);
*/

/*
  //::: find covariance matrix of sample points
  //::: Q=[ q0 q1 q2 ... ] (==points as column vectors; qi=pi-median_i)
  //::: then PC's are eigenvectors of Q * Q'  ( <-- covariance matrix )
  rMatrix V,S;
  rPoint3 centroid;
  MiscUtility::FindPrincipalComponentsOf3DPointCloud(V,S,centroid,M);

  //::: create 3x3 array that transforms original alpha (voting) space into a space
  //::: where the principle components are orthonormal basis vectors; may also want
  //::: to transform centroid to origin...
  Transform norm;
  FORi(3) FORj(3) norm[i][j]=V[j][i];
norm=norm*Transform(sqrt(S[0][0]),sqrt(S[1][1]),sqrt(S[2][2]));
norm=norm.Inverse();
SCREENPUT norm.Dump("norm");
  ManifoldPointStruct3Space_SetVoteNormalizer(norm,centroid);
*/


/*  
  //::: find covariance matrix of sample points
  //::: Q=[ q0 q1 q2 ... ] (==points as column vectors; qi=pi-median_i)
  //::: then PC's are eigenvectors of Q * Q'  ( <-- covariance matrix )
  const Array<rPoint3>& P=M;
  rMatrix Q(3,P.size);
  rPoint3 median=rPoint3(0,0,0);
  FORi(P.size) median+=P[i];
  median/=P.size;
  FORi(P.size) { Q[0][i]=P[i].x-median.x; Q[1][i]=P[i].y-median.y; Q[2][i]=P[i].z-median.z; }
  rMatrix QQ=Q*(Q.TransposeOf());	// covariance matrix
  QQ*=1.0/P.size;
  QQ=QQ.InverseOf();
SCREENPUT median.Dump("centroid");
SCREENPUT QQ.Dump();

  //::: create 3x3 array that transforms original alpha (voting) space into a space
  //::: where the principle components are orthonormal basis vectors; may also want
  //::: to transform centroid to origin...
  Transform norm;
  FORi(3) FORj(3) norm[i][j]=QQ[i][j];
  norm.Normalize();
SCREENPUT norm.Dump("norm");
  ManifoldPointStruct3Space_SetVoteNormalizer(norm,median);
*/
  SCREENPUT printf("END NORM\n");  fflush(stdout);
}

//////////////////////////////////////////////////////////////////////
void IntersectionByTriangles::SetLattice(
  const Array<real>&		min_th,
  const Array<real>&		max_th,
  const Array<real>&		min_ka,
  const Array<real>&		max_ka,
  const int			CON_n_theta,
  const int			CON_n_kappa,
  const int			CON_n_mani
)
{
/*
//::: for testing precise timer
INFINITELOOP {
  RunningTimeDisplay::SwissTime();
  FORi(1000) FORj(1000) FORk(1000) ;
  real t=RunningTimeDisplay::SwissTime();
  printf("%lf\n",t);  
}
*/

  Theta.SetSize(CON_n_mani,CON_n_theta);
  Kappa.SetSize(CON_n_mani,CON_n_kappa);
  lattice.SetSize(CON_n_mani);
  b_lattice.SetSize(CON_n_mani);
  FORi(CON_n_mani) lattice[i].SetSize(CON_n_theta,CON_n_kappa);
  FORi(CON_n_mani) b_lattice[i].SetSize(CON_n_theta,CON_n_kappa);

  //::: uniform-sample each manifold
  FORi(CON_n_mani) FORj(CON_n_kappa) Kappa[i][j]=min_ka[i]+(real(j)/(CON_n_kappa-1))*(max_ka[i]-min_ka[i]);
  FORi(CON_n_mani) FORj(CON_n_theta) Theta[i][j]=min_th[i]+(real(j)/(CON_n_theta-1))*(max_th[i]-min_th[i]);

  Transform Tab;
  rPoint3 alpha, dum;
  RunningTimeDisplay::SwissTime();
  SCREENPUT RunningTimeDisplay::Reset(lattice.size);
  FORk(lattice.size) {
    SCREENPUT RunningTimeDisplay::Tick();
    FORi(CON_n_theta) FORj(CON_n_kappa) {
      if (!AffineCalibFrom2Params_efficient(Tab,dum,Kappa[k][j],Theta[k][i],pairinfo[k])) EXIT("err0913247");
      PPOMCM_FindAlpha(alpha,Tab,T0i_proj,E,pairinfo[k].i,pairinfo[k].j);
      lattice[k][i][j]=alpha;
      b_lattice[k][i][j]=TRUE;
    }
  }
  SCREENPUT RunningTimeDisplay::Done();
  real runtime=RunningTimeDisplay::SwissTime();
  real time_per_node=runtime/(real(lattice.size)*CON_n_theta*CON_n_kappa);
  SCREENPUT printf("time to calc each lattice pt = %lf\n",time_per_node);

  rPoint3 corners[]={rPoint3(100,0,0), rPoint3(0,100,0), rPoint3(0,0,100), rPoint3(-100,0,0), rPoint3(0,-100,0) };
  FORk(lattice.size) {
    FORi(Theta.cols) FORj(Kappa.cols) {
      rPoint3& p=lattice[k][i][j];
      if ( (!within(p.x,-100,100))||(!within(p.y,-100,100))||(!within(p.z,-100,100)) ) {
        p=corners[k];
	b_lattice[k][i][j]=FALSE;
      }
    }
  }
}

//////////////////////////////////////////////////////////////////////
void IntersectionByTriangles::RansacIteration(
  rPoint3&			best_ans,		// best intersection point
  Transform&			best_K,			// internal calibration
  int&				best_n_good,		// number of manifolds that fit the answer well
  real&				best_err,		// error induced by K
  const Array<int>&		draws,
  const Array<LinkedList_ClusterLatticeCluster>&	ll_clusters,
  const boole			b_output
)
{
    CONST boole CON_b_show_stats = FALSE;

    const int ONE=draws[0], TWO=draws[1], THREE=draws[2];
  //  const int ONE=0, TWO=1, THREE=2;

    //::: exhaustively compare each cluster bounding box for two manifolds
    if (CON_b_show_stats) { BANNER("intersecting bounding boxes");  fflush(stdout); }
    LinkedList<ClusterPair> ll_overlap;
    FORLIST(ll_clusters[ONE]) {
      ClusterLatticeCluster& Ca=ll_clusters[ONE].Now();
      if (!Ca.b_legal) continue;
      FORLIST(ll_clusters[TWO]) {
	ClusterLatticeCluster& Cb=ll_clusters[TWO].Now();
	if (!Cb.b_legal) continue;
	if (Ca.lo.x>Cb.hi.x) continue;
	if (Cb.lo.x>Ca.hi.x) continue;
	if (Ca.lo.y>Cb.hi.y) continue;
	if (Cb.lo.y>Ca.hi.y) continue;
	if (Ca.lo.z>Cb.hi.z) continue;
	if (Cb.lo.z>Ca.hi.z) continue;
	//::: boxes overlap
	ClusterPair* pair=new ClusterPair;
	pair->a=&Ca;
	pair->b=&Cb;
	ll_overlap.Add(pair);
      }
    }

  /*
    //::: show boxes that are part of an overlap pair
    //::: (*) mark all boxes as illegal
    FORi(2) {
      FORLISTnow(ll_clusters[i], ClusterLatticeCluster*)
	now->b_legal=FALSE;
      }
    }

    //::: (*) mark as legal only those boxes that overlap another
    FORLISTnow(ll_overlap, ClusterPair)
      now.a->b_legal=TRUE;
      now.b->b_legal=TRUE;
    }

    //::: (*) display bounding boxes
    BANNER("displaying bounding boxes");
    VrmlFile::Open("bounding.boxes.wrl");
    FORi(2) {
      FORLISTnow(ll_clusters[i], ClusterLatticeCluster*)
	if (!now->b_legal) continue;
	VrmlFile::OutputBox(now->lo,now->hi,color[i]);
      }
    }
    VrmlFile::Close();
  */

    if (CON_b_show_stats) SCREENPUT { printf("OVERLAP PAIRS:  %d\n",ll_overlap.size);  fflush(stdout); }

    //::: for each overlapping cluster pair, exhaustively check pairwise triangle intersections
    if (CON_b_show_stats) { BANNER("checking pairwise triangle intersection for overlapping clusters");  fflush(stdout); }
    long n_test_count=0;
    LinkedList<ClusterTrianglePair> ll_inters;	// stores each pair of intersecting triangles
    FORLISTnow(ll_overlap, ClusterPair)
      //::: create list of triangles for each cluster
      FORl(2) {
	const int i=(l==0) ? ONE : TWO;
	ClusterLatticeCluster* pcl=(l==0) ? now.a : now.b;
	ClusterLatticeCluster& cl=*pcl;
	if (cl.b_tri_exist) continue;		// only realize triangles if haven't previously done so
  // make into function; used another place, too
	cl.b_tri_exist=TRUE;
	int th0=cl.th0;
	int th1=cl.th1-1;
	int ka0=cl.ka0;
	int ka1=cl.ka1-1;
	FORjIN(th0,th1) FORkIN(ka0,ka1) {
	  //::: only allow reasonably small triangles with legal vertices
	  if (!b_lattice[i][j+1][k]) continue;
	  if (!b_lattice[i][j][k+1]) continue;
	  const rPoint3& b=lattice[i][j+1][k];
	  const rPoint3& c=lattice[i][j][k+1];
	  real d3=(b-c).Norm();
	  if (d3>=0.1) continue;
	  if (b_lattice[i][j][k]) {
	    const rPoint3& a=lattice[i][j][k];
	    real d1=(a-b).Norm();
	    real d2=(a-c).Norm();
	    if ((d1<.1)||(d2<.1))
	      cl.tri.Add(new ClusterTriangle(a,b,c));
	  }
	  if (b_lattice[i][j+1][k+1]) {
	    const rPoint3& a=lattice[i][j+1][k+1];
	    real d1=(a-b).Norm();
	    real d2=(a-c).Norm();
	    if ((d1<.1)||(d2<.1))
	      cl.tri.Add(new ClusterTriangle(a,b,c));
	  }

  /*
	  if (!b_lattice[i][j+1][k]) continue;
	  if (!b_lattice[i][j][k+1]) continue;
	  if (b_lattice[i][j][k])
	    cl.tri.Add(new ClusterTriangle(lattice[i][j][k],     lattice[i][j+1][k], lattice[i][j][k+1]));
	  if (b_lattice[i][j+1][k+1])
	    cl.tri.Add(new ClusterTriangle(lattice[i][j+1][k+1], lattice[i][j+1][k], lattice[i][j][k+1]));
  */
	}
      }
      //::: try to intersect each triangle pair
      const LinkedList<ClusterTriangle>& tri0=now.a->tri;
      const LinkedList<ClusterTriangle>& tri1=now.b->tri;
      FORLIST(tri0) {
	ClusterTriangle& A=tri0.Now();
	FORLIST(tri1) {
	  ClusterTriangle& B=tri1.Now();
	  boole b_inters=tri_tri_intersect(A,B);
	  if (b_inters) {
	    ClusterTrianglePair* pair=new ClusterTrianglePair;
	    pair->a=A;
	    pair->b=B;
	    pair->lo=A.a;
	    if (pair->lo.x>A.b.x) pair->lo.x=A.b.x;
	    if (pair->lo.x>A.c.x) pair->lo.x=A.c.x;
	    if (pair->lo.x>B.a.x) pair->lo.x=B.a.x;
	    if (pair->lo.x>B.b.x) pair->lo.x=B.b.x;
	    if (pair->lo.x>B.c.x) pair->lo.x=B.c.x;
	    if (pair->lo.y>A.b.y) pair->lo.y=A.b.y;
	    if (pair->lo.y>A.c.y) pair->lo.y=A.c.y;
	    if (pair->lo.y>B.a.y) pair->lo.y=B.a.y;
	    if (pair->lo.y>B.b.y) pair->lo.y=B.b.y;
	    if (pair->lo.y>B.c.y) pair->lo.y=B.c.y;
	    if (pair->lo.z>A.b.z) pair->lo.z=A.b.z;
	    if (pair->lo.z>A.c.z) pair->lo.z=A.c.z;
	    if (pair->lo.z>B.a.z) pair->lo.z=B.a.z;
	    if (pair->lo.z>B.b.z) pair->lo.z=B.b.z;
	    if (pair->lo.z>B.c.z) pair->lo.z=B.c.z;

	    pair->hi=A.a;
	    if (pair->hi.x<A.b.x) pair->hi.x=A.b.x;
	    if (pair->hi.x<A.c.x) pair->hi.x=A.c.x;
	    if (pair->hi.x<B.a.x) pair->hi.x=B.a.x;
	    if (pair->hi.x<B.b.x) pair->hi.x=B.b.x;
	    if (pair->hi.x<B.c.x) pair->hi.x=B.c.x;
	    if (pair->hi.y<A.b.y) pair->hi.y=A.b.y;
	    if (pair->hi.y<A.c.y) pair->hi.y=A.c.y;
	    if (pair->hi.y<B.a.y) pair->hi.y=B.a.y;
	    if (pair->hi.y<B.b.y) pair->hi.y=B.b.y;
	    if (pair->hi.y<B.c.y) pair->hi.y=B.c.y;
	    if (pair->hi.z<A.b.z) pair->hi.z=A.b.z;
	    if (pair->hi.z<A.c.z) pair->hi.z=A.c.z;
	    if (pair->hi.z<B.a.z) pair->hi.z=B.a.z;
	    if (pair->hi.z<B.b.z) pair->hi.z=B.b.z;
	    if (pair->hi.z<B.c.z) pair->hi.z=B.c.z;

	    ll_inters.Add(pair);
	  }
	  ++n_test_count;
	}
      }
    }
    SCREENPUT if (CON_b_show_stats) { printf("TOTAL TRI INTERSECTION TESTS:  %ld\n",n_test_count);  fflush(stdout); }
    SCREENPUT if (CON_b_show_stats) { printf("TOTAL PAIRWISE TRI INTERSECTIONS:  %ld\n",ll_inters.size);  fflush(stdout); }

    //::: free unneeded memory
    // ll_inters must be freed at end of each RANSAC trial!!!!!!!!

  /*
    //::: display pairwise intersections
    LinkedList<rPoint3> ll_TRI_a;
    FORLISTnow(ll_inters, ClusterTrianglePair)
      ll_TRI_a.Add(new rPoint3(now.a.a));
      ll_TRI_a.Add(new rPoint3(now.a.b));
      ll_TRI_a.Add(new rPoint3(now.a.b));
      ll_TRI_a.Add(new rPoint3(now.a.c));
      ll_TRI_a.Add(new rPoint3(now.a.c));
      ll_TRI_a.Add(new rPoint3(now.a.a));
    }
    LinkedList<rPoint3> ll_TRI_b;
    FORLISTnow(ll_inters, ClusterTrianglePair)
      ll_TRI_b.Add(new rPoint3(now.b.a));
      ll_TRI_b.Add(new rPoint3(now.b.b));
      ll_TRI_b.Add(new rPoint3(now.b.b));
      ll_TRI_b.Add(new rPoint3(now.b.c));
      ll_TRI_b.Add(new rPoint3(now.b.c));
      ll_TRI_b.Add(new rPoint3(now.b.a));
    }
    VrmlFile::Open("pairwise.intersecting.triangles.wrl");
    Array<rPoint3> M;
    M=ll_TRI_a;	ll_TRI_a.FreeMem();
    VrmlFile::OutputLines(M,color[0]);
    M=ll_TRI_b;	ll_TRI_b.FreeMem();
    VrmlFile::OutputLines(M,color[1]);
    VrmlFile::Close();
    M.FreeMem();
    EXIT();
  */

    //::: now intersect overlapping triangles with third manifold
    //::: (*) compare bb of each cluster in third manifold with bb of each triangle pair
    //::: (*) if bb's overlap, compare each triangle within cluster with triangle pair
    LinkedList<ClusterTriangleTriplet> ll_mutual_inters;		// possible triples stored here
    FORLIST(ll_clusters[THREE]) {
      CONST ClusterLatticeCluster& cl=ll_clusters[THREE].Now();
      if (!cl.b_legal) continue;
      FORLISTnow(ll_inters, ClusterTrianglePair)
	// compare bounding boxes
	if (cl.lo.x>now.hi.x) continue;
	if (now.lo.x>cl.hi.x) continue;
	if (cl.lo.y>now.hi.y) continue;
	if (now.lo.y>cl.hi.y) continue;
	if (cl.lo.z>now.hi.z) continue;
	if (now.lo.z>cl.hi.z) continue;

	// check each triangle in cluster C vs. pair
	if (!cl.b_tri_exist) {

	  // make into function; used another place, too
	  cl.b_tri_exist=TRUE;
	  int th0=cl.th0;
	  int th1=cl.th1-1;
	  int ka0=cl.ka0;
	  int ka1=cl.ka1-1;
	  FORjIN(th0,th1) FORkIN(ka0,ka1) {
	    //::: only allow reasonably small triangles with legal vertices
	    if (!b_lattice[THREE][j+1][k]) continue;
	    if (!b_lattice[THREE][j][k+1]) continue;
	    const rPoint3& b=lattice[THREE][j+1][k];
	    const rPoint3& c=lattice[THREE][j][k+1];
	    real d3=(b-c).Norm();
	    if (d3>=0.1) continue;
	    if (b_lattice[THREE][j][k]) {
	      const rPoint3& a=lattice[THREE][j][k];
	      real d1=(a-b).Norm();
	      real d2=(a-c).Norm();
	      if ((d1<.1)||(d2<.1))
		cl.tri.Add(new ClusterTriangle(a,b,c));
	    }
	    if (b_lattice[THREE][j+1][k+1]) {
	      const rPoint3& a=lattice[THREE][j+1][k+1];
	      real d1=(a-b).Norm();
	      real d2=(a-c).Norm();
	      if ((d1<.1)||(d2<.1))
		cl.tri.Add(new ClusterTriangle(a,b,c));
	    }
	  }

	}
	FORLIST(cl.tri) {
	  CONST ClusterTriangle& C=cl.tri.Now();
	  CONST ClusterTriangle& A=now.a;
	  CONST ClusterTriangle& B=now.b;
	  boole b_inters_AC=tri_tri_intersect(A,C);
	  boole b_inters_BC=tri_tri_intersect(B,C);
	  if (b_inters_AC && b_inters_BC) {
	    ClusterTriangleTriplet* trip=new ClusterTriangleTriplet;
	    trip->a=A;
	    trip->b=B;
	    trip->c=C;
	    ll_mutual_inters.Add(trip);
	  }
	}
      }
    }
    SCREENPUT if (CON_b_show_stats) { printf("TOTAL THREEWAY INTERSECTIONS:  %ld\n",ll_mutual_inters.size);  fflush(stdout); }

    //::: find exact mutual intersection points for triples that seem to intersect
    LinkedList<rPoint3> ll_mutuals;
    FORLISTnow(ll_mutual_inters, ClusterTriangleTriplet)
      // NORMa*(x-a) = 0 if x is in plane of triangle a
      rPoint3 na=((now.a.b-now.a.a).Cross(now.a.c-now.a.a)).Hat();
      rPoint3 nb=((now.b.b-now.b.a).Cross(now.b.c-now.b.a)).Hat();
      rPoint3 nc=((now.c.b-now.c.a).Cross(now.c.c-now.c.a)).Hat();
      real dpa=-(na.Dot(now.a.a));
      real dpb=-(nb.Dot(now.b.a));
      real dpc=-(nc.Dot(now.c.a));
      rMatrix M(3,4);
      M[0][0]=na.x;    M[0][1]=na.y;    M[0][2]=na.z;    M[0][3]=dpa;
      M[1][0]=nb.x;    M[1][1]=nb.y;    M[1][2]=nb.z;    M[1][3]=dpb;
      M[2][0]=nc.x;    M[2][1]=nc.y;    M[2][2]=nc.z;    M[2][3]=dpc;
      rVector v;
      M.FindNullEigenvector(v);
      rPoint3 P(v[0],v[1],v[2]);
      P/=v[3];

      //::: put P in triangle-based coords and check if P really is inside triangle
      Transform T_triA_world(now.a.b-now.a.a,now.a.c-now.a.a,na);
      Transform T_world_triA=T_triA_world.Inverse();
      rPoint3 P_triA=T_world_triA*(P-now.a.a);

      Transform T_triB_world(now.b.b-now.b.a,now.b.c-now.b.a,nb);
      Transform T_world_triB=T_triB_world.Inverse();
      rPoint3 P_triB=T_world_triB*(P-now.b.a);

      Transform T_triC_world(now.c.b-now.c.a,now.c.c-now.c.a,nc);
      Transform T_world_triC=T_triC_world.Inverse();
      rPoint3 P_triC=T_world_triC*(P-now.c.a);

      if ( ((P_triA.x>=0)&&(P_triA.y>=0)&&(P_triA.x+P_triA.y<=1))&&
	   ((P_triB.x>=0)&&(P_triB.y>=0)&&(P_triB.x+P_triB.y<=1))&&
	   ((P_triC.x>=0)&&(P_triC.y>=0)&&(P_triC.x+P_triC.y<=1)) )
	ll_mutuals.Add(new rPoint3(P));
    }
    SCREENPUT if (CON_b_show_stats) { printf("TOTAL TRUE THREEWAY INTERSECTIONS:  %ld\n",ll_mutuals.size);  fflush(stdout); }

/*
    //::: display threeway intersections
    LinkedList<rPoint3> ll_TRI_a;
    FORLISTnow(ll_mutual_inters, ClusterTriangleTriplet)
      ll_TRI_a.Add(new rPoint3(now.a.a));
      ll_TRI_a.Add(new rPoint3(now.a.b));
      ll_TRI_a.Add(new rPoint3(now.a.b));
      ll_TRI_a.Add(new rPoint3(now.a.c));
      ll_TRI_a.Add(new rPoint3(now.a.c));
      ll_TRI_a.Add(new rPoint3(now.a.a));
    }
    LinkedList<rPoint3> ll_TRI_b;
    FORLISTnow(ll_mutual_inters, ClusterTriangleTriplet)
      ll_TRI_b.Add(new rPoint3(now.b.a));
      ll_TRI_b.Add(new rPoint3(now.b.b));
      ll_TRI_b.Add(new rPoint3(now.b.b));
      ll_TRI_b.Add(new rPoint3(now.b.c));
      ll_TRI_b.Add(new rPoint3(now.b.c));
      ll_TRI_b.Add(new rPoint3(now.b.a));
    }
    LinkedList<rPoint3> ll_TRI_c;
    FORLISTnow(ll_mutual_inters, ClusterTriangleTriplet)
      ll_TRI_c.Add(new rPoint3(now.c.a));
      ll_TRI_c.Add(new rPoint3(now.c.b));
      ll_TRI_c.Add(new rPoint3(now.c.b));
      ll_TRI_c.Add(new rPoint3(now.c.c));
      ll_TRI_c.Add(new rPoint3(now.c.c));
      ll_TRI_c.Add(new rPoint3(now.c.a));
    }
    VrmlFile::Open("threeway.intersecting.triangles.wrl");
    Array<rPoint3> M;
    M=ll_TRI_a;	ll_TRI_a.FreeMem();
    VrmlFile::OutputLines(M,color[0]);
    M=ll_TRI_b;	ll_TRI_b.FreeMem();
    VrmlFile::OutputLines(M,color[1]);
    M=ll_TRI_c;	ll_TRI_c.FreeMem();
    VrmlFile::OutputLines(M,color[2]);
    //::: show true intersection point
    M.SetSize(1);
    M[0]=true_alpha;
    VrmlFile::OutputPoints(M,RGB3Byte_yellow,15);
    //::: show calc'd mutual intersection points
    M=ll_mutuals;
    VrmlFile::OutputPoints(M,RGB3Byte_pink,16);
    VrmlFile::Close();
    M.FreeMem();
*/

    //::: convert each intersection point to K matrix; calc goodness of solution
    best_n_good=-1;
//    const real CON_good_limit=1.0e-1;
    const real CON_good_limit=0.6e-1;	// worked for many cases, although just barely
//    const real CON_good_limit=1.5e-1;

    //::: init some variables used in testing
    real best_frob_with_true_K=-1;
    Transform trueK;
    if (pInternal!=NULL) {
      trueK=(normalizer.Inverse()*(*pInternal));
      trueK=trueK/trueK.Norm();
    }
    
    //::: check each mutual intersection point
    SCREENPUT printf("mutual intersections = %d\n",ll_mutuals.size);
    FORl(ll_mutuals.size) {
//      rPoint3 ans=true_alpha;
      rPoint3 ans=ll_mutuals[l];
      VOTE_UNNORM(ans);
      //::: use ans to find relative calibrations for each view pair
      Array<Transform> allTabs(pairinfo.size);
      FORi(pairinfo.size) {
	int ii=pairinfo[i].i;
	int jj=pairinfo[i].j;
	Transform TT;
	if (ii==0)
	  TT=T0i_proj[jj]+Transform::Outer(E[jj],ans);
	else {
	  Transform TTi=T0i_proj[ii]+Transform::Outer(E[ii],ans);
	  Transform TTj=T0i_proj[jj]+Transform::Outer(E[jj],ans);
	  TT=TTj*(TTi.Inverse());
	}
	allTabs[i]=(normalizer.Inverse()*TT*normalizer);
	allTabs[i].Normalize();
      }

      //::: find K; only use the three manifolds involved in the intersection
      Array<Transform> threeTabs(3);
      threeTabs[0]=allTabs[ONE];
      threeTabs[1]=allTabs[TWO];
      threeTabs[2]=allTabs[THREE];
      Transform absconic;
      MiscUtility::FindAbsoluteConicFromRelativeCalibrations(absconic,threeTabs);
      rMatrix rmKK(&absconic[0][0],3,3);
      rmKK/=(rmKK[0][0]);
      rmKK=rmKK.InverseOf();
      int bGoodCholesky=rmKK.Cholesky();
      //if (!bGoodCholesky) { ALERT("Cholesky decomposition failed\n"); continue; }
      if (!bGoodCholesky) continue;
      rmKK=rmKK.InverseOf();
      Transform K(rmKK.data);
      Transform invK=K.Inverse();
      
      //::: if true K is known, check how well K matches true K; keep records
      if (pInternal!=NULL) {
        real frob=((K/K.Norm())-trueK).Norm();
//        SCREENPUT printf("frob = %lf\n",frob);
        if ((best_frob_with_true_K<0)||(best_frob_with_true_K>frob)) best_frob_with_true_K=frob;
      }

      //::: test how well K fits with all fmats/Hinfs
      real tot_err=0;
      int n_good=0;
      FORi(allTabs.size) {
        real det=allTabs[i].Determinant();
	real scale=1.0/pow(fabs(det),1.0/3)*sign(det);
	Transform R=(invK*allTabs[i]*K)*scale;
	Transform RR=R*(R.TransposeOf());
//	RR.Dump();
	real err=(Transform_Id-RR).Norm();		// frobenius norm
	SCREENPUT if (CON_b_show_stats) { printf("err = %e          %d  %d\n",err,pairinfo[i].i,pairinfo[i].j); }
	if (err<CON_good_limit) {
//if (draws.bHolds(i)) {		// use this method to maximize fit to just the three manifolds used in iteration
	  ++n_good;
	  tot_err+=err*err;
	}
      }
      SCREENPUT if (CON_b_show_stats) { printf("tot err = %e\n",tot_err); }

      //::: following test eliminates solutions that are not realistic cameras
      if ( (!within(K[0][0]/K[1][1],0.95,1.05)) ||
           (!within(K[0][2]/1024/K[1][2]*768,0.90,1.1)) )
            n_good=min(n_good,1);		// kludgy way to indicate bad camera

      //::: keep track of best answer so far
      if ((best_n_good==-1) || (n_good>best_n_good) || ((n_good==best_n_good)&&(tot_err<best_err))) {
        best_n_good=n_good;
	best_err=tot_err;
	best_K=K;
	best_ans=ans;
      }
    }	// end loop

    //::: for testing, output "per trial" info

    //::: free mem
    ll_inters.FreeMem();
    ll_overlap.FreeMem();
    ll_mutuals.FreeMem();
    ll_mutual_inters.FreeMem();
}

//////////////////////////////////////////////////////////////////////
void IntersectionByTriangles::Run(
  const Array2D<boole>&		bPairOn
)
{
  const int	CON_n_theta	= 50;
  const int	CON_n_kappa	= 50;
  const real	CON_min_theta	= 0.01;
  const real	CON_max_theta	= PI-0.01;
  const real	CON_min_kappa	= 0.05;
  const real	CON_max_kappa	= 0.95;

  SCREENPUT BANNER("starting exhaustive triangle intersection search\a");  fflush(stdout);
  SCREENPUT true_alpha.Dump("true alpha");

  Array<real> min_th(CON_n_manifolds), max_th(CON_n_manifolds);
  Array<real> min_ka(CON_n_manifolds), max_ka(CON_n_manifolds);
  FORi(min_th.size) min_th[i]=CON_min_theta;
  FORi(max_th.size) max_th[i]=CON_max_theta;
  FORi(min_ka.size) min_ka[i]=CON_min_kappa;
  FORi(max_ka.size) max_ka[i]=CON_max_kappa;
  
  SetLattice(min_th,max_th,min_ka,max_ka,CON_n_theta,CON_n_kappa,CON_n_manifolds);

  NormalizeVotingSpace();
  VOTE_NORM(true_alpha);

  //::: normalize lattice points
  FORi(lattice.size) FORj(Theta.cols) FORk(Kappa.cols)
    VOTE_NORM(lattice[i][j][k]);

SCREENPUT printf("\a");  fflush(stdout);
  //::: create clusters
  BANNER("creating clusters");  fflush(stdout);
  Array<LinkedList_ClusterLatticeCluster> ll_clusters(CON_n_manifolds);
  FORi(ll_clusters.size) {
    ll_clusters[i].FreeMem();
    const int th_step=2;
    const int ka_step=2;
    for (int j=0; j<CON_n_theta; j+=th_step) {
      for (int k=0; k<CON_n_kappa; k+=ka_step) {
	ClusterLatticeCluster* pc=new ClusterLatticeCluster;
	ClusterLatticeCluster& c=*pc;
	c.th0=j;
	c.th1=j+th_step;
	if (c.th1 > CON_n_theta-1) c.th1=CON_n_theta-1;
	c.ka0=k;
	c.ka1=k+ka_step;
	if (c.ka1 > CON_n_kappa-1) c.ka1=CON_n_kappa-1;
	c.b_legal=FALSE;
	c.b_tri_exist=FALSE;
	ll_clusters[i].Add(pc);
      }
    }
  }
  
  //::: find bounding box for each cluster
  FORi(ll_clusters.size) {
    FORLISTnow(ll_clusters[i], ClusterLatticeCluster)
      boole b_first_legal=TRUE;
      FORjIN(now.th0,now.th1) FORkIN(now.ka0,now.ka1) {
        if (!b_lattice[i][j][k]) continue;	// non-existent point--doesn't count
        const rPoint3& p=lattice[i][j][k];
	if (b_first_legal) {
	  now.lo=now.hi=p;
	  now.b_legal=TRUE;
	  b_first_legal=FALSE;
	}
	else {
	  if (p.x<now.lo.x)		now.lo.x=p.x;
	  else if (p.x>now.hi.x)	now.hi.x=p.x;
	  if (p.y<now.lo.y)		now.lo.y=p.y;
	  else if (p.y>now.hi.y)	now.hi.y=p.y;
	  if (p.z<now.lo.z)		now.lo.z=p.z;
	  else if (p.z>now.hi.z)	now.hi.z=p.z;
	}
      }
      if ((now.lo.x==now.hi.x) || (now.lo.y==now.hi.y) || (now.lo.z==now.hi.z))
        now.b_legal=FALSE;	// trivial cluster--won't lead to meaningful mutual intersection pt
    }
  }

/*
  //::: display bounding boxes
  VrmlFile::Open("bounding.boxes.wrl");
  FORi(2) {
    FORLISTnow(ll_clusters[i], ClusterLatticeCluster*)
      if (!now->b_legal) continue;
      VrmlFile::OutputBox(now->lo,now->hi,color[i]);
    }
  }
  VrmlFile::Close();
*/

  //::: use ransac to find K
  int CON_n_ransac_trials=40;
  
  //::: (*) if ransac_trials is larger than (n_manifolds!) then create list of all possible draws (combinations)
  boole b_all_combinations=FALSE;
  Array2D<uint8> combinations;
  if (CON_n_manifolds<=6) {
    CreateCombinations(combinations,CON_n_manifolds,3);
    if (combinations.rows<=CON_n_ransac_trials)	{ b_all_combinations=TRUE; CON_n_ransac_trials=combinations.rows; }
    else					{ b_all_combinations=FALSE; combinations.FreeMem(); }
  }

  Array<int> draws(3);
  real err, best_err;
  Transform K, best_K;
  rPoint3 ans, best_ans;
  int n_good, best_n_good=-1;

  SCREENPUT RunningTimeDisplay::Stopwatch();
  for (int loop=0; loop<CON_n_ransac_trials; ++loop) {
    //::: pick 3 manifolds to intersect
    if (b_all_combinations)
      FORi(3) draws[i]=combinations[loop][i];		// step through all combinations
    else
      RandomIndexSet(draws,CON_n_manifolds);		// pick three indices at random from [0,CON_n_manifolds-1]

    //::: try to intersect them
    RansacIteration(ans,K,n_good,err,draws,ll_clusters,FALSE);

    //::: check if best answer so far
    if ((best_n_good==-1) || (n_good>best_n_good) || ((n_good==best_n_good)&&(err<best_err))) {
      best_n_good=n_good;
      best_err=err/((best_n_good>0) ? best_n_good : 1);	// err per accepted manifold
      best_K=K;
      best_ans=ans;
      SCREENPUT printf("(%5d)   %e    [%d]  ",loop,best_err,best_n_good);
      SCREENPUT best_ans.Dump();
    }
  }
  SCREENPUT RunningTimeDisplay::Stopwatch();
  
  SCREENPUT best_K.Dump("best K");
  SCREENPUT best_ans.Dump("best ans");
  VOTE_UNNORM(true_alpha);
  SCREENPUT true_alpha.Dump("true ans");
  
  SCREENPUT OutputResults(best_ans,best_K);



  //::: tabulate results for experiments
  Transform answer=best_K;
  real frob_err, frob_err2;
  Transform per_entry_diff;
  if (pInternal!=NULL) {
    Transform trueK=(normalizer.Inverse()*(*pInternal));
    Transform diff=(answer/answer.Norm())-(trueK/trueK.Norm());
    per_entry_diff=diff;
    frob_err=diff.Norm();
    diff[0][2]=diff[1][2]=0;	// use to just check error in main components
    frob_err2=diff.Norm();
  } else
    frob_err=frob_err2=-1;

  GLOBAL_testrun_data.error_frobenius_main=frob_err;
  GLOBAL_testrun_data.error_frobenius=frob_err2;
  GLOBAL_testrun_data.error_per_entry=per_entry_diff;
  GLOBAL_testrun_data.n_grid_size=CON_n_theta;			// assumes same as CON_n_kappa
  GLOBAL_testrun_data.n_ransac_trials=CON_n_ransac_trials;

  throw IntersectionByTrianglesError("SUCCESS!",0);
}

//////////////////////////////////////////////////////////////////////
void IntersectionByTriangles::OutputResults(
  const rPoint3&		best_ans,
  const Transform&		best_K
)
{
  Transform T_norm0_met=(normalizer*best_K).Inverse();
  Transform T_met_norm0=T_norm0_met.Inverse();
  Transform inv_norm=normalizer.Inverse();
  const rPoint3 a=best_ans;
  //::: create projective camera matrices
  Array<rMatrix> cams(E.size);
  FORi(cams.size) if (i>0) {
    cams[i].SetSize(3,4);
    Transform Tab=T0i_proj[i]+Transform::Outer(E[i],a);
    FORj(3) FORk(3) cams[i][j][k]=Tab[j][k];
    cams[i][0][3]=E[i].x;
    cams[i][1][3]=E[i].y;
    cams[i][2][3]=E[i].z;
  }
  cams[0].SetSize(3,4);
  cams[0].FillWith(0);
  cams[0][0][0]=cams[0][1][1]=cams[0][2][2]=1;

  //::: use normalizing transform on camera matrices
  rMatrix KKK(4,4);
  KKK.MakeIdentity();
  FORi(3) FORj(3) KKK[i][j]=T_met_norm0[i][j];
  FORi(cams.size) cams[i]=cams[i]*KKK;

  //::: reconstruct scene
  Array<rPoint3> W;
  MiscUtility::Triangulate(W,cams,(*GLOB_Kpts));
  VrmlFile::OutputPoints("output.recon.wrl",W,RGB3Byte_green);
  Array<rPoint3> nul(0);  
  VrmlFile::OutputPointsAndCameras("output.world.and.cameras.wrl",W,nul,cams);

  //::: write metric camera matrices for future use
  FILE* file=fopen("output.cameras","wb");
  ASSERT(file,"err632147");
  rMatrix rm_inv_norm(&inv_norm[0][0],3,3);
  fprintf(file,"%d\n",cams.size);
  FORi(cams.size) {
    rMatrix cam=rm_inv_norm*cams[i];
    fprintf(file,"\n");
    FORj(3) fprintf(file,"%.9e %.9e %.9e %.9e\n",cam[j][0],cam[j][1],cam[j][2],cam[j][3]);
  }
  fclose(file);

  //::: output 3D --> 2D correspondence (for later texture mapping)
  file=fopen("given.dense.matching","rb");
  if (file) {
    int nViews, nPts;
    fscanf(file,"%d\n",&nPts);
    fscanf(file,"%d\n",&nViews);
    Array2D<rPoint3> dense(nViews,nPts);
    Transform T_normi_i=normalizer.Inverse();
    FORj(nViews) {
      FORk(nPts) {
	fscanf(file,"%lf %lf\n",&dense[j][k].x,&dense[j][k].y);
	dense[j][k].z=1.0;
	dense[j][k]=normalizer*dense[j][k];
	dense[j][k]/=dense[j][k].z;
      }
      fscanf(file,"\n");
    }
    fclose(file);
//	Transform T_met_0=normalizer.Inverse()*T_met_norm0;
////        MiscUtility::AffineTriangulate_WithInternalParams(W,Tab,T_met_norm0,E[i],dense[0],dense[i]);
////        file=fopen((String("mapping.")+String(i,"%02d")+".0").Ptr(),"w");
    FORk(W.size) {
      rPoint3 p=T_normi_i*dense[0][k];
      p/=p.z;
      fprintf(file,"%8e  %8e  %8e      %8e  %8e\n",W[k].x,W[k].y,W[k].z,p.x,p.y);
    }
    fclose(file);
  } else {
    const int view=0;
    FILE* file2=fopen((String("output.mapping.into.view")+String(view,"%02d")).Ptr(),"w");
    FORk(W.size) {
//      rPoint3 p=(*GLOB_Kpts)[view][k];
      rVector pp, P(4);
      P[0]=W[k].x;  P[1]=W[k].y;  P[2]=W[k].z;  P[3]=1.0;
      pp=cams[view]*P;
      rPoint3 p(pp[0],pp[1],pp[2]);
      p=inv_norm*p;
      p/=p.z;
      fprintf(file2,"%8e  %8e  %8e      %8e  %8e\n",W[k].x,W[k].y,W[k].z,p.x,p.y);
    }
    fclose(file2);
  }

EXIT();


  //::: bundle adjust:  use reconstruction to refine features and remove outlying features
  Array2D<rPoint3> newK(GLOB_Kpts->rows,GLOB_Kpts->cols);
  rVector P(4), p(3);
  P[3]=1.0;
  FORi(newK.rows) {
    const rMatrix& cam=cams[i];
    FORj(W.size) {
      P[0]=W[j].x;  P[1]=W[j].y;  P[2]=W[j].z;
      p=cam*P;
      newK[i][j]=rPoint3(p[0]/p[2], p[1]/p[2], 1.0);
    }
  }

  //::: show reconstruction projected into each view
  const int HHH=768, WWW=1024;
  Image3ByteChannels img(HHH,WWW);
  KltFeatureTable klt;
  klt.SetSize(newK[0].size,cams.size);
  FORk(cams.size) {
    img.FillWithColor(RGB3Byte_black);
    FORi(newK[k].size) {
      rPoint3 p=inv_norm*newK[k][i];
      p/=p.z;
      img.DrawSquare(rPoint2(p.x,p.y),3,RGB3Byte_green);
      rPoint3 q=inv_norm*(*GLOB_Kpts)[k][i];
      q/=q.z;
      img.DrawSquare(rPoint2(q.x,q.y),3,RGB3Byte_red);
      img.DrawLine(rPoint2(p.x,p.y),rPoint2(q.x,q.y),RGB3Byte_cyan);
      klt[i][k].msg=(k==0) ? 1 : 0;
      klt[i][k].x=q.x;		// store original pt
      klt[i][k].y=q.y;
    /*
    // for moving points slowly towards consensus
      klt[i][k].x=(p.x+q.x)/2;
      klt[i][k].y=(p.y+q.y)/2;
    //  if (sqrt(SQ(q.x-p.x)+SQ(q.y-p.y))>2) klt[i][0].msg=-1;
    */
    }
    img.WriteToFile((String("test")+String(k,"%02d")+".jpg").Ptr());
  }

  //::: here because of compiler bug
  Array<rMatrix> cms(klt.cols-1);
  FORi(cms.size) cms[i].SetSize(3,4);

  //::: find point that is most in disagreement with consensus and replace it with consensus
  FORi(klt.rows) {
    real worst_dist=-1;
    int worst_inx=-1;
//    rPoint3 worst_pt;
    FORj(klt.cols) {
      rPoint3 p=inv_norm*newK[j][i];
      p/=p.z;
      rPoint3 q=inv_norm*(*GLOB_Kpts)[j][i];
      q/=q.z;
      real dist=sqrt(SQ(q.x-p.x)+SQ(q.y-p.y));
      if ((worst_inx==-1)||(worst_dist<dist)) {
	worst_inx=j;
	worst_dist=dist;
//        worst_pt=p;		// store consensus pt for this bad feature
      }
    }
    //::: replace feature with consensus of *remaining* features
    Array2D<rPoint3> good_fe(klt.cols-1,1);
    int inx=0;
//  FORi(cms.size) FORk(3) FORl(4) cms[i][k][l]=cams[i][k][l];
    FORj(klt.cols) if (j!=worst_inx) {
      good_fe[inx][0]=(*GLOB_Kpts)[j][i];	// work in normalized image space
//      cms[inx].MakeCopyOf(cams[j]);	// very inefficient -- use ptrs? (must modify Triangulate)
      ++inx;
    }
    //::: DUE TO APPARENT COMPILER PROBLEM, HAVE TO COPY CAMS IN SEPARATE LOOP?!?!?!?!
    inx=0;
    FORk(klt.cols) if (k!=worst_inx) {
      cms[inx].MakeCopyOf(cams[k]);	// very inefficient -- use ptrs? (must modify Triangulate)
      ++inx;
    }
    Array<rPoint3> WW;
    MiscUtility::Triangulate(WW,cms,good_fe);
    rVector p, P(4);
    rPoint3 pp;
    P[0]=WW[0].x;  P[1]=WW[0].y;  P[2]=WW[0].z;  P[3]=1.0;
    p=cams[worst_inx]*P;
    pp=inv_norm*rPoint3(p[0],p[1],p[2]);
    pp/=pp.z;
    klt[i][worst_inx].x=pp.x;
    klt[i][worst_inx].y=pp.y;
//    klt[i][worst_inx].x=worst_pt.x;
//    klt[i][worst_inx].y=worst_pt.y;
  }

  //::: remove any row (feature) that has a -1 msg in the first view
  int c=0;
  FORi(klt.rows) if (klt[i][0].msg==1) ++c;
  KltFeatureTable klt2;
  klt2.SetSize(c,klt.cols);
  c=0;
  FORi(klt.rows) { if (klt[i][0].msg!=-1) { FORj(klt[i].size) klt2[c][j]=klt[i][j]; ++c; } }

  klt2.WriteToFile("given.features",HHH,WWW);
  EXIT();
}

#ifdef NOTDEFINED
//////////////////////////////////////////////////////////////////////
void IntersectionByTriangles::ReconstructScene(
  const AlphaCoord&		a
)
{
  //::: create projective camera matrices
  Array<rMatrix> cams(E.size);
  FORi(cams.size) if (i>0) {
    cams[i].SetSize(3,4);
    Transform Tab=T0i_proj[i]+Transform::Outer(E[i],a);
    FORj(3) FORk(3) cams[i][j][k]=Tab[j][k];
    cams[i][0][3]=E[i].x;
    cams[i][1][3]=E[i].y;
    cams[i][2][3]=E[i].z;
  }
  cams[0].SetSize(3,4);
  cams[0].FillWith(0);
  cams[0][0][0]=cams[0][1][1]=cams[0][2][2]=1;

  //::: use normalizing transform on camera matrices
  rMatrix KKK(4,4);
  KKK.MakeIdentity();
  FORi(3) FORj(3) KKK[i][j]=T_met_norm0[i][j];
  FORi(cams.size) cams[i]=cams[i]*KKK;

  //::: reconstruct scene
  Array<rPoint3> W;
  MiscUtility::Triangulate(W,cams,(*GLOB_Kpts));

/*
//::: write metric camera matrices for future use
FILE* file=fopen((String("recon.cameras.")+String(nZoominsPerformed,"%02d")).Ptr(),"wb");
ASSERT(file,"err0934034");
fprintf("%d\n",cams.size);
FORi(cams.size) {
  fprintf(file,"\n");
  FORj(3) fprintf(file,"%.9e %.9e %.9e %.9e\n",cams[i][j][0],cams[i][j][1],cams[i][j][2],cams[i][j][3]);
}
fclose(file);
*/

  InventorFile::OutputPoints(String("recon")+String(1,"%02d")+".iv",W,RGB3Byte_green);

#ifdef NOTDEFINED    
    //::: output 3D --> 2D correspondence for each view (for later texture mapping)
    if (i==1) {		//only doing first pair for now
      FILE* file=fopen("given.dense.matching","rb");
      if (file) {
        int nViews, nPts;
        fscanf(file,"%d\n",&nPts);
        fscanf(file,"%d\n",&nViews);
	Array2D<rPoint3> dense(nViews,nPts);
        Transform T_normi_i=normalizer.Inverse();
	FORj(nViews) {
	  FORk(nPts) {
	    fscanf(file,"%lf %lf\n",&dense[j][k].x,&dense[j][k].y);
	    dense[j][k].z=1.0;
	    dense[j][k]=normalizer*dense[j][k];
	    dense[j][k]/=dense[j][k].z;
	  }
	  fscanf(file,"\n");
	}
        fclose(file);
//	Transform T_met_0=normalizer.Inverse()*T_met_norm0;
        MiscUtility::AffineTriangulate_WithInternalParams(W,Tab,T_met_norm0,E[i],dense[0],dense[i]);
        file=fopen((String("mapping.")+String(i,"%02d")+".0").Ptr(),"w");
        FORk(W.size) {
          rPoint3 p=T_normi_i*dense[0][k];
	  p/=p.z;
          fprintf(file,"%8e  %8e  %8e      %8e  %8e\n",W[k].x,W[k].y,W[k].z,p.x,p.y);
        }
        fclose(file);
      } else {
        Transform T_normi_i=normalizer.Inverse();
        FILE* file2=fopen((String("mapping.")+String(i,"%02d")+".0").Ptr(),"w");
        FORk(W.size) {
          rPoint3 p=T_normi_i*((*GLOB_Kpts)[0][k]);
	  p/=p.z;
          fprintf(file2,"%8e  %8e  %8e      %8e  %8e\n",W[k].x,W[k].y,W[k].z,p.x,p.y);
        }
        fclose(file2);
      }
    }
#endif
/*
    if (pInternal!=NULL) {
      //::: use known internal calibration (NOTE: assumes already multiplied by Tinorm)
//      MiscUtility::AffineTriangulate_WithInternalParams(W,Tab,normalizer*(*pInternal),E[i],(*GLOB_Kpts)[0],(*GLOB_Kpts)[i]);
      MiscUtility::AffineTriangulate_WithInternalParams(W,T0i_proj[i]+Transform::Outer(E[i],true_alpha),*pInternal,E[i],(*GLOB_Kpts)[0],(*GLOB_Kpts)[i]);

      SCREENPUT InventorFile::OutputPoints(String("true.recon")+String(i,"%02d")+".iv",W,RGB3Byte_green);
//throw GeneticFillOutManifoldGroup3SpaceError();
    }
*/

}
#endif


