#include <stdio.h>
#include <math.h>
#include <string.h>

#include "InventorFile.h"
#include "RM_SameCameraManip.h"
#include "MultiCameraWorkbench.h"
#include "IntersectionByTriangles.h"
#include "GeneticFillOutManifoldGroup.3Space.h"
#include "TestRunDataStruct.h"
#include "KltFeatureTable.h"
#include "helpers.h"

#include "MiscUtility.h"   //temporary

/*/////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////

GENERAL NOTES:

+ Tij_proj means "basis transform between camera i and camera j in
  some particular projective reconstruction of the world"; this could
  have been called Hij_ref, meaning "homography induced by some fixed
  'reference' plane" (Pollefeys uses that notation); choice of reference
  plane equals choice of projective basis/reconstruction
  
+ similarly, Tij_met means "basis transform between camera i and camera
  j in a (certain) metric reconstruction"; this could have been called Hij_inf
  for "homography induced by the plane at infinity (in the desired
  metric frame)"; problem is, plane at infinity depends on your chosen
  reference frame; Pollefeys' thesis uses latter notation

///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////
*//////////////////////////////////////////////////////////////////////

#ifdef NOTDEFINED
/*/////////////////////////////////////////////////////////////////////
the common projective basis will have camera 0's matrix be [ I | 0 ];
in the projective reconstruction built from cameras 0 and i, the matrix
for camera i is [ M0i | eB0i ]; thus in the common projective basis,
the matrix for camera i will become newPi := [ M0i | eB0i ] * T_common_1i ;
then for camera pair i,j the M matrix is found from newPi
and newPj (as the relative calibration)
*//////////////////////////////////////////////////////////////////////
void FindCameraMatricesInCommonProjectiveBasis_orig(
  rMatrix&			newPi,
  const Transform&		M0i,
  const rPoint3&		eB0i,
  const rMatrix&		T_common_1i		// 4x4 collineation
)
{
  rMatrix oldPi(3,4);
  FORi(3) FORj(3) oldPi[i][j]=M0i[i][j];
  oldPi[0][3]=eB0i.x;  oldPi[1][3]=eB0i.y;  oldPi[2][3]=eB0i.z;
  newPi=oldPi*T_common_1i;
}
#endif

#ifdef NOTDEFINED
// each camera C is a best-fit 3x4 matrix that maps world coords P to local
// image-plane coords p; i.e., p ~~ C*P; this version does not preserve fmats,
// so caller will need to recalculate fmats from camera matrices later on
///////////////////////////////////////////////////////////////////////
void FindCameraMatricesInCommonProjectiveBasis_tested_ver2(
  Array<rMatrix>&		CM,
  const Array<rPoint3>&		P,		// pts in world coords
  const Array2D<rPoint3>&	p		// pts in camera coords
)
{
  //::: find each camera matrix up to an unknown scale
  FORi(CM.size) {
    CM[i].SetSize(3,4);
    rMatrix Q(2*P.size,12);
    Q.FillWith(0);
    int r=0;
    FORj(P.size) {
      const real& x=P[j].x;  const real& y=P[j].y;  const real& z=P[j].z;
      const real w=1;
      const real& a=p[i][j].x;
      const real& b=p[i][j].y;
      const real& c=p[i][j].z;
      Q[r][0]=x*b;  Q[r][1]=y*b;  Q[r][2]=z*b;  Q[r][3]=w*b;
      Q[r][4]=-a*x; Q[r][5]=-a*y; Q[r][6]=-a*z; Q[r][7]=-a*w;
      ++r;
      Q[r][0]=x*c;  Q[r][1]=y*c;  Q[r][2]=z*c;  Q[r][3]=w*c;
      Q[r][8]=-a*x; Q[r][9]=-a*y; Q[r][10]=-a*z; Q[r][11]=-a*w;
      ++r;
    }
    rVector v;
    Q.FindNullEigenvector(v);
    int ind=0;
    FORj(3) FORk(4) CM[i][j][k]=v[ind++];
  }

/*
//CAN ONLY FIND A CAMERA MATRIX UP TO A SCALE ANYWAY!!!  
  //::: now determine relative scalings of camera matrices so pairwise epipoles are correct
  //::: use first two cameras to triangulate the positions of the remaining camera centers,
  //::: from the epipoles
  CM[0]=CM[0]/CM[0][0][0];
  
  rPoint3 cB(CM[1][0][3],CM[1][1][3],CM[1][2][3]);
  CM[1]=CM[1]/cB.Norm();
  
  Array<rPoint3> centers;
  MiscUtility::Triangulate(centers,CM[0],epis0,CM[1],epis1);
  
  FORi(CM.size) if (i>1) {
    Transform Twi;
    FORj(3) FORk(3) Twi[j][k]=CM[i][j][k];
    rPoint3 ci(CM[i][0][3],CM[i][1][3],CM[i][2][3]);
    rPoint3 cw=Twi.Inverse()*ci*(-1);
    real scale=center[i-2].x/cw.x;		//use least squares here...
    CM[i]=CM[i]*scale;
  }
*/
}
#endif

// this version works but assumes camera ONE is [I|0] and assures that original
// fmats (defined by M and eB) are preserved
///////////////////////////////////////////////////////////////////////
void FindCameraMatricesInCommonProjectiveBasis(
  Array<rMatrix>&		CM,
  const Array<rPoint3>&		P,		// pts in world coords
  const Array2D<rPoint3>&	p,		// pts in camera coords
  const Array2D<Transform>&	M,
  const Array2D<rPoint3>&	eB
)
{
  //::: CM = [ M + e*beta' | k*e ] = [ M | 0 ] + e*Beta'  <-- need only find Beta (in R^4)

  //::: find each camera matrix up to an unknown scale
  FORi(CM.size) {
    CM[i].SetSize(3,4);

    if (i==0) {
      CM[i].FillWith(0);
      CM[i][0][0]=CM[i][1][1]=CM[i][2][2]=1;
      continue;
    }

    rMatrix QQ(2*P.size,5);
    QQ.FillWith(0);
    rPoint3 E=(i==0) ? rPoint3(0,0,0) : eB[0][i];
    int r=0;
    FORj(P.size) {
      const real& x=P[j].x;  const real& y=P[j].y;  const real& z=P[j].z;
      const real w=1;
      const real& a=p[i][j].x;
      const real& b=p[i][j].y;
      const real& c=p[i][j].z;		// see 5-27-00
      rPoint3 Q=(i==0) ? Q=P[j] : Q=M[0][i]*P[j];
      QQ[r][0]=c*x*E.x-a*x*E.z;      QQ[r][1]=c*y*E.x-a*y*E.z;
      QQ[r][2]=c*z*E.x-a*z*E.z;      QQ[r][3]=c*w*E.x-a*w*E.z;
      QQ[r][4]=c*Q.x-a*Q.z;
      ++r;

      QQ[r][0]=c*x*E.y-b*x*E.z;      QQ[r][1]=c*y*E.y-b*y*E.z;
      QQ[r][2]=c*z*E.y-b*z*E.z;      QQ[r][3]=c*w*E.y-b*w*E.z;
      QQ[r][4]=c*Q.y-b*Q.z;
      ++r;
    }
    rVector v;
    QQ.FindNullEigenvector(v);
    rMatrix MM(3,4);
    MM.FillWith(0);
    FORj(3) FORk(3) MM[j][k]=M[0][i][j][k];
    rMatrix EE(3,1);
    EE[0][0]=E.x;    EE[1][0]=E.y;    EE[2][0]=E.z;
    rMatrix beta(1,4);
    beta[0][0]=v[0]/v[4];
    beta[0][1]=v[1]/v[4];
    beta[0][2]=v[2]/v[4];
    beta[0][3]=v[3]/v[4];
    CM[i]=MM+EE*beta;

//      LINE();
//      CM[i].Dump();
      rPoint2 tot(0,0);
      FORj(P.size) {
        rMatrix PP(4,1);
	PP[0][0]=P[j].x; PP[1][0]=P[j].y; PP[2][0]=P[j].z; PP[3][0]=1;
	rVector q=CM[i]*PP;
	q=q*(1.0/q[2]);
	rPoint3 qq(q[0],q[1],q[2]);
//	(qq-p[i][j]).Dump();
        tot.x+=abs(qq.x-p[i][j].x);
        tot.y+=abs(qq.y-p[i][j].y);
      }
      tot/=P.size;
//      tot.Dump();

  }
  
  Array<rPoint3> nul(0);  
  InventorFile::OutputPointsAndCameras("output.proj.world.iv",P,nul,CM);
}

#ifdef NOTDEFINED
///////////////////////////////////////////////////////////////////////
void main_simulated_data()
{
  srandom(8713);

  const int			CON_ncameras=10;
  const int			CON_nworldpts=200;
  const real			CON_noiseradius=0.001;
  
  MultiCameraWorkbench bench(CON_ncameras, CON_nworldpts, CON_noiseradius,0,1);

  //::: find camera matrices for common basis
  Array<rMatrix> commonCM(CON_ncameras);
  FindCameraMatricesInCommonProjectiveBasis_tested_ver2(commonCM,bench.R[0][1],bench.Kpts);
  
  //::: extract needed info from common-basis camera matrices
  Array<Transform> T0i_proj(CON_ncameras);	//first 3 columns
  Array<rPoint3> E(CON_ncameras);		//last column
  FORi(CON_ncameras) {
    FORj(3) FORk(3) T0i_proj[i][j][k]=commonCM[i][j][k];
    E[i]=rPoint3(commonCM[i][0][3],commonCM[i][1][3],commonCM[i][2][3]);
  }

  const AlphaCoord true_alpha=bench.alpha[0][1];

//  PlotPointsOnModulusConstraintManifold(T0i_proj,E,F,M,eL,true_alpha,4);
//  ManifoldBasedVoting(T0i_proj,E,F,M,eL,true_alpha,4);
//  VisualizePossible_h3s(F[0][1],M[0][1],eL[0][1],alpha[0][1]);

//  NearestFitAllVoxelsWithManifold(T0i_proj,E,F,M,eL,true_alpha,0,1);
//  VisualizePossible_h3s_ver3(F[0][1],M[0][1],eL[0][1],alpha[0][1],T0i_proj,E,0,1);

//  int one=0, two=1;
//  GeneticFillOutManifold(F[one][two],M[one][two],eL[one][two],true_alpha,T0i_proj,E,one,two);

  GeneticFillOutManifoldGroup* gfomg=
    new GeneticFillOutManifoldGroup(bench.F,bench.M,bench.eL,true_alpha,T0i_proj,bench.Tinorm,E,bench.CON_pairs,&bench.Kpts);
  gfomg->Run(bench.bPairOn);

//  ShowManifoldFixedTheta(F,M,eL,true_alpha,T0i_proj,E,0,1);
//  ShowManifold(F,M,eL,true_alpha,T0i_proj,E,0,2);

///  AlphaCoord target(1,1,1);
//  int one=0, two=2;
//  ManifoldPointStruct mtarg=ManifoldPointStruct(-13,1,bench.F[one][two],bench.M[one][two],
//    bench.eL[one][two],T0i_proj,E,one,two);
//  MoveAlongManifold2(mtarg.a+AlphaCoord(3,3,3),bench.F,bench.M,bench.eL,T0i_proj,E,one,two);
}
#endif

// use following main for experimental data runs
///////////////////////////////////////////////////////////////////////
void DoSimulatedTestRuns(int argc, char** argv)
{
  if (argc!=8) {
    printf("performing simulated test runs\n\n");
    printf("USAGE: %s computer_id ntrials ncameras n_manifolds_to_use noiseradius_lo noiseradius_hi min_retinal_obj_size\n\n",argv[0]);
    EXIT();
  }

  const int	computer_id			= atoi(argv[1]);
  const int	ntrials				= atoi(argv[2]);
  const int	CON_ncameras			= atoi(argv[3]);
  const int	CON_n_manifolds_to_use		= atoi(argv[4]);
  const real	CON_noiseradius_lo		= atof(argv[5]);
  const real	CON_noiseradius_hi		= atof(argv[6]);
  const real	CON_min_retinal_obj_size	= atof(argv[7]);

  const int one=0, two=1;
  const rPoint2	CON_image_size			= rPoint2(1000,1000);

//  const int seed=computer_id*17+912;
//  const int seed=computer_id*17+3912;
//  const int seed=computer_id*17+4912;
//  const int seed=computer_id*17+5912;
  const int seed=computer_id*17+7912;
  srandom(seed);
  Array<uint32> seed_per_trial(ntrials);
  Array<uint32> more_seeds(6*ntrials);
  FORi(ntrials) seed_per_trial[i]=uint32(random());
  FORi(more_seeds.size) more_seeds[i]=uint32(random());

//  String basedir="/p/vision/volumes/volume02/EXPERIMENTS/";
  String basedir="EXPERIMENTS/";
  String resultfilename=basedir+"results."+String(computer_id,"%02d");
  FILE* resultfile=fopen(resultfilename.Ptr(),"wb");
  if (!resultfile) EXIT("failed to open output file 1");

  String result2filename=basedir+"per.interation."+String(computer_id,"%02d");
  GLOBAL_testrun_data.file_per_iteration_data=fopen(result2filename.Ptr(),"wb");
  if (!GLOBAL_testrun_data.file_per_iteration_data) EXIT("failed to open output file 2");

  FORi(ntrials) {
    printf("(%2d) TRIAL %4d:\n",computer_id,i);   fflush(stdout);
 
    const int CON_max_redos=1;	// max 5
    FORj(CON_max_redos) {
      boole bWorked=FALSE;
      srandom(seed_per_trial[i]);
  
      real noiseradius=frandom(CON_noiseradius_lo,CON_noiseradius_hi);
  
      GLOBAL_testrun_data.Clear();
      GLOBAL_testrun_data.noise_radius=noiseradius;
      GLOBAL_testrun_data.n_views=CON_ncameras;
      GLOBAL_testrun_data.min_retinal_obj_size=CON_min_retinal_obj_size;

regenerate_random_scene:  
      //::: generate random scene
      MultiCameraWorkbench* bench=new MultiCameraWorkbench(CON_ncameras, 0, noiseradius,
        one, two, CON_min_retinal_obj_size, CON_image_size);
      AlphaCoord true_alpha=bench->alpha[one][two];

      /*
      // failed attempt to ensure alpha is nice
      AlphaCoord dum;
      dum.x=fabs(true_alpha.x);
      dum.y=fabs(true_alpha.y);
      dum.z=fabs(true_alpha.z);
      real max=max3(dum.x,dum.y,dum.z);
      dum/=max;
      if (1.0/dum.x>6) goto regenerate_random_scene;
      if (1.0/dum.y>6) goto regenerate_random_scene;
      if (1.0/dum.z>6) goto regenerate_random_scene;
      */
  
      GLOBAL_testrun_data.n_scene_points=bench->Kpts.cols;
  
      //::: find camera matrices for common basis
      Array<rMatrix> commonCM(bench->CON_ncameras);
      FindCameraMatricesInCommonProjectiveBasis(commonCM,bench->R[one][two],bench->Kpts,bench->M,bench->eL);
    
      //::: extract needed info from common-basis camera matrices; later transferred to pairinfo
      Array<Transform> T0i_proj(bench->CON_ncameras);		//first 3 columns
      Array<rPoint3> E(bench->CON_ncameras);			//last column
      FORl(E.size) {
        FORj(3) FORk(3) T0i_proj[l][j][k]=commonCM[l][j][k];
        E[l]=rPoint3(commonCM[l][0][3],commonCM[l][1][3],commonCM[l][2][3]);
      }
  
      //::: find/estimate true alpha from projective camera matrices if internal calibration
      //::: not known; this happens when perfect camera matrices are read from an external file;
      //::: done at this stage to account for Hartley-normalization of images
      if (bench->internal.Norm()!=0) {
        SCREENPUT bench->Hinf[one][two].DumpNormalized("Hinf");
        FindAlpha(true_alpha,bench->Hinf[one][two],T0i_proj[two],E[two]);
        SCREENPUT (bench->Hinf[one][two].NormalizationOf()-(T0i_proj[two]+Transform::Outer(E[two],true_alpha)).NormalizationOf()).Dump("0 if true_alpha correct");
        SCREENPUT true_alpha.Dump("true alpha");
      }

      //::: my intersection code assumes alpha coords in range [-100,100]
      if (fabs(true_alpha.x)>50) goto regenerate_random_scene;
      if (fabs(true_alpha.y)>50) goto regenerate_random_scene;
      if (fabs(true_alpha.z)>50) goto regenerate_random_scene;
  
      //::: put bench pairwise view info into more generic form for algorithm; this was
      //::: added to use the new "ViewPairInfo" data structure for improving computation
      Array<ViewPairInfo> pairinfo(bench->CON_pairs);
      int inx=0;
      FORk(bench->F.rows) FORl(bench->F.cols) if (k<l) {
        pairinfo[inx].Init(bench->F[k][l],bench->M[k][l],k,l);
        if (k==0) pairinfo[inx].SetProjectiveFrame(T0i_proj[l], E[l]);
        ++inx;
      }
  
  /*
  {
  //::: used to generate sample random scene image for CVPR01 paper
  Array<rPoint3> W;
  MiscUtility::Triangulate(W,bench->CM,bench->Kpts);
  InventorFile::Open((String("exper.scene.iv")+i).Ptr());
  InventorFile::OutputPoints(W,RGB3Byte_white,5);
  FORi(bench->CM.size) InventorFile::OutputCameraFulcrum(bench->CM[i],10,10);
  InventorFile::Close();
  }
  */
  
      srandom(more_seeds[i*5+j]);
  
      //::: convenience variables, for added clarity
      Array2D<rPoint3>* const final_dense=&bench->Kpts;
      Transform* const final_internal=(bench->internal.Norm()==0) ? NULL : &bench->internal;
  
#ifdef NOTDEFINED
      //::: find camera calibration using voting method
      GeneticFillOutManifoldGroup3Space* gfomg=
        new GeneticFillOutManifoldGroup3Space(pairinfo,
           T0i_proj,E,true_alpha,bench->normalization,final_dense,final_internal);
  
      try {
        GLOBAL_testrun_data.error_code=0;
        gfomg->Run(bench->bPairOn);
      }
      catch (GeneticFillOutManifoldGroup3SpaceError& errobj) {
        GLOBAL_testrun_data.error_code=errobj.error_code;
        if (errobj.error_code<0) {
          if (j+1==CON_max_redos) printf("(%2d) (ERROR) %s\n",computer_id,errobj.msg.Ptr());
        }
        else { printf("(%2d) %s\n",computer_id,errobj.msg.Ptr()); bWorked=TRUE; }
        fflush(stdout);
      }
  
      delete gfomg;
      delete bench;
  
      if (bWorked) break;
#endif

//#ifdef NOTDEFINED  
      //::: find camera calibration using triangle-brute-force method
      if (within(CON_n_manifolds_to_use,1,pairinfo.size-1))
        pairinfo.ReduceSize(CON_n_manifolds_to_use);
      IntersectionByTriangles* ibc=
        new IntersectionByTriangles(pairinfo,
              T0i_proj,E,true_alpha,bench->normalization,CON_image_size,final_dense,final_internal);
      try {
        GLOBAL_testrun_data.error_code=0;
//        ibc->Run("modc",bench->bPairOn);
        ibc->Run("tri",bench->bPairOn);
//        ibc->Run("mcmc",bench->bPairOn);
      }
      catch (IntersectionByTrianglesError& errobj) {
        GLOBAL_testrun_data.error_code=errobj.error_code;
        if (errobj.error_code<0) {
          if (j+1==CON_max_redos) printf("(%2d) (ERROR) %s\n",computer_id,errobj.msg.Ptr());
        }
        else { printf("(%2d) %s\n",computer_id,errobj.msg.Ptr()); bWorked=TRUE; }
        fflush(stdout);
      }
  
      delete ibc;
      delete bench;
  
      if (bWorked) break;
//#endif
  
    } // end redo loop
    GLOBAL_testrun_data.Write_TriangleIntersect_TR1428(resultfile);
//    GLOBAL_testrun_data.Write_TriangleIntersect_TR1428(stdout);
  } // end experiment loop
  
  fclose(resultfile); 
  fclose(GLOBAL_testrun_data.file_per_iteration_data);
  
  EXIT();
}

#define SIMULATED_DATA

///////////////////////////////////////////////////////////////////////
int main(int argc, char** argv)
{
  DoSimulatedTestRuns(argc,argv);
//  srandom(8713);
  srandom(36487);		// leads to an incorrect but legit mutual intersection point!
//  srandom(12864);		// with zoom-in:  beauty; manifolds have interesting twisty shapes after about 5 zoom-ins
//  srandom(28213);
//  srandom(19321);		// with 3 views, manifolds seem to intersect in at least 2 places
//  				// but works well with zoom-in variation!?
//  srandom(46883);		// not good for cyan
//  srandom(216674);		// beauty!  don't forget to remove "precomputed.bestalpha"
//  srandom(3471);

//  srandom(34713);
//  srandom(6245);		// would need another level of refinement of mesh
//  srandom(22887);
  srandom(1773);		// only single answer ?!
  const int one=0, two=1;
  rPoint2 image_size(10,10);

#ifdef SIMULATED_DATA
//main_simulated_data();
//return;
  const int			CON_ncameras=4;
  const int			CON_nworldpts=50;
  const real			CON_noiseradius=0.0;  //0.005

  image_size=rPoint2(1000,1000);
  
  MultiCameraWorkbench bench(CON_ncameras, CON_nworldpts, CON_noiseradius, one, two, 100, image_size);
  AlphaCoord true_alpha=bench.alpha[one][two];
//bench.Hinf[one][two].DumpNormalized("true Hinf");
//(bench.M[one][two]+Transform::Outer(bench.eL[one][two],bench.alpha[one][two])).DumpNormalized();
#else

  MultiCameraWorkbench bench("given.features","given.fmats",one,two);
  AlphaCoord true_alpha(0,0,0);
  if (bench.internal.Norm()!=0) {
    true_alpha=bench.alpha[one][two];
//    bench.Hinf[one][two].DumpNormalized("Hinf");
//    (bench.M[one][two]+Transform::Outer(bench.eL[one][two],true_alpha)).DumpNormalized();
//    return;
  }
#endif

  //::: find camera matrices for common basis
  Array<rMatrix> commonCM(bench.CON_ncameras);
  FindCameraMatricesInCommonProjectiveBasis(commonCM,bench.R[one][two],bench.Kpts,bench.M,bench.eL);
  
  //::: extract needed info from common-basis camera matrices; later transferred to pairinfo
  Array<Transform> T0i_proj(bench.CON_ncameras);	//first 3 columns
  Array<rPoint3> E(bench.CON_ncameras);			//last column
  FORi(E.size) {
    FORj(3) FORk(3) T0i_proj[i][j][k]=commonCM[i][j][k];
    E[i]=rPoint3(commonCM[i][0][3],commonCM[i][1][3],commonCM[i][2][3]);
  }

  //::: find/estimate true alpha from projective camera matrices if internal calibration
  //::: not known; this happens when perfect camera matrices are read from an external file;
  //::: done at this stage to account for Hartley-normalization of images
  if (bench.internal.Norm()!=0) {
    SCREENPUT bench.Hinf[one][two].DumpNormalized("Hinf");
    FindAlpha(true_alpha,bench.Hinf[one][two],T0i_proj[two],E[two]);
    SCREENPUT (bench.Hinf[one][two].NormalizationOf()-(T0i_proj[two]+Transform::Outer(E[two],true_alpha)).NormalizationOf()).Dump("0 if true_alpha correct");
    SCREENPUT true_alpha.Dump("true alpha");

/*
    //::: sanity check that true_alpha leads to correct world reconstruction
    Array2D<rPoint3> p(2,bench.Kpts.cols);
    p[0]=bench.Kpts[one];
    p[1]=bench.Kpts[two];
    Array<rMatrix> cm(2);
    cm[0]=bench.CM[one];
    cm[1]=bench.CM[two];
    Array<rPoint3> W;
    MiscUtility::Triangulate(W,cm,p);
    InventorFile::OutputPoints("initial.world.cm.test.iv",W,RGB3Byte_white);
    InventorFile::OutputPoints("original.world.iv",bench.W,RGB3Byte_white);
    
    MiscUtility::AffineTriangulate_WithInternalParams(W,T0i_proj[two]+Transform::Outer(E[two],
      true_alpha),bench.internal,E[two],bench.Kpts[one],bench.Kpts[two]);
    InventorFile::OutputPoints("initial.world.test.iv",W,RGB3Byte_white);
*/

/*
bench.internal.Dump("xx pInternal");
E[two].Dump("xx E[i]");
T0i_proj[two].Dump("xx T0i_proj");
(T0i_proj[two]+Transform::Outer(E[two],true_alpha)).Dump("xx calc Tab");
true_alpha.Dump("xx true_alpha");
*/
    
//    return;
  }

  //::: put bench pairwise view info into more generic form for algorithm; this was
  //::: added to use the new "ViewPairInfo" data structure for improving computation
  Array<ViewPairInfo> pairinfo(bench.CON_pairs);
  int inx=0;
  FORk(bench.F.rows) FORl(bench.F.cols) if (k<l) {
    pairinfo[inx].Init(bench.F[k][l],bench.M[k][l],k,l);
    if (k==0) pairinfo[inx].SetProjectiveFrame(T0i_proj[l], E[l]);
    ++inx;
  }
//  pairinfo[0].SetProjectiveFrame();
//return; 



//FORi(pairinfo.size) pairinfo[i].kmap.FreeMem();

/*
{
//::: find theta and kappa for true_alpha
real ka, th;
Transform Hinf=T0i_proj[two]+Transform::Outer(E[two],true_alpha);
FindParameterizationOfHinf(ka,th,Hinf,T0i_proj[two],E[two]);
printf("k=%e    th=%e\n",ka,th);
Transform calcHinf;
rPoint3 alpha;
//AffineCalibFrom2Params(calcHinf,alpha,ka,th,bench.F[one][two],T0i_proj[two],E[two]);
//AffineCalibFrom2Params(calcHinf,alpha,ka,th,pairinfo[0].F,pairinfo[0].Hinf_proj,pairinfo[0].eB_proj);
AffineCalibFrom2Params_efficient(calcHinf,alpha,ka,th,pairinfo[0]);
printf("%d  %d\n",pairinfo[0].i,pairinfo[0].j);
alpha.Dump();
true_alpha.Dump();
return;
}
*/



  //::: load point correspondences for reconstruction...
  //::: (*) first try dense correspondence file
  Array2D<rPoint3>* dense=NULL;
  FILE* file=fopen("input.correspondence","rb");
  if (file) {
    SCREENPUT printf("reading dense point correspondence...\n");
    dense=new Array2D<rPoint3>;
    int n_views, n_pts;
    fscanf(file,"%d\n",&n_views);
    fscanf(file,"%d\n",&n_pts);
    dense->SetSize(n_views,n_pts);
    FORi(n_views) {
      fscanf(file,"\n");
      FORj(n_pts) {
        fscanf(file,"%lf %lf\n",&((*dense)[i][j].x),&((*dense)[i][j].y));
        (*dense)[i][j].z=1;
      }
    }
    fclose(file);
  }
  else SCREENPUT printf("dense correspondence not available\n");

  if (dense==NULL) {
    //::: (*) look for klt feature table file
    file=fopen("input.klt.feature.table","rb");
    if (file) {
      fclose(file);
      SCREENPUT printf("reading klt feature table...\n");
      KltFeatureTable ftab;
      int r,c;
      ftab.ReadFromFile(r,c,"input.klt.feature.table");
      dense=new Array2D<rPoint3>;
      dense->SetSize(ftab.cols,ftab.rows);
      FORi(ftab.cols) FORj(ftab.rows) (*dense)[i][j]=rPoint3(ftab[j][i].x,ftab[j][i].y,1);
      }
    else SCREENPUT printf("klt feature table not available\n");
  }

  //::: further process dense correspondence if available
  if (dense!=NULL) {
    FORi(dense->rows) FORj(dense->cols) {
      (*dense)[i][j]=bench.normalization*(*dense)[i][j];
      (*dense)[i][j]/=(*dense)[i][j].z;
    }
    /*
    if (dense->rows<bench.Kpts.rows) {
      Array2D<rPoint3>* new_dense=new Array2D<rPoint3>;
      new_dense->SetSize(bench.Kpts.rows,dense->cols);
      FORi(dense->rows) (*new_dense)[i]=(*dense)[i];
      FORiIN(dense->rows,bench.Kpts.rows-1) (*new_dense)[i]=(*dense)[dense->rows-1];
      delete dense;
      dense=new_dense;
    }
    */
  }

  //::: convenience variables, for added clarity
  Array2D<rPoint3>* const final_dense=(dense!=NULL) ? dense : &bench.Kpts;
  Transform* const final_internal=(bench.internal.Norm()==0) ? NULL : &bench.internal;

/*
  //::: find camera calibration using voting method
  GeneticFillOutManifoldGroup3Space* gfomg=
    new GeneticFillOutManifoldGroup3Space(pairinfo,
       T0i_proj,E,true_alpha,bench.normalization,final_dense,final_internal);
  try {
    GLOBAL_testrun_data.error_code=0;
    gfomg->Run(bench.bPairOn);
  }
  catch (GeneticFillOutManifoldGroup3SpaceError& errobj) {
    GLOBAL_testrun_data.error_code=errobj.error_code;
    if (errobj.error_code<0) {
      printf("(ERROR) %s\n",errobj.msg.Ptr());
    }
    else printf("%s\n",errobj.msg.Ptr());
    fflush(stdout);
  }
*/

  //::: find camera calibration using triangle-brute-force method
  IntersectionByTriangles* ibc=
    new IntersectionByTriangles(pairinfo,
          T0i_proj,E,true_alpha,bench.normalization,image_size,final_dense,final_internal);
  try {
    GLOBAL_testrun_data.error_code=0;
    ibc->Run("tri",bench.bPairOn);
  }
  catch (IntersectionByTrianglesError& errobj) {
    GLOBAL_testrun_data.error_code=errobj.error_code;
    if (errobj.error_code<0) {
      printf("(ERROR) %s\n",errobj.msg.Ptr());
    }
    else printf("%s\n",errobj.msg.Ptr());
    fflush(stdout);
  }




  if (dense!=NULL) delete dense;	// not used in triangle intersection method yet
}

